/*
 * FX_CutSubs.c
 *
 * FX modules that cuts individual elements.
 *
 * Copyright 1999 - 2000, Massachusetts Institute of Technology.
 *
 */

#include <stdio.h>
#include <stdlib.h>


#include "FX_types.h"


/*
 *     Sets the 2D node into the intersection structure returns node #
 *       allows the non-duplication of 2D nodes
 */

static void
FX_2DNode(int n1, int n2, float d, int *nc,
          int *nedgeX, FXinter *edgeX, int *nodeTable)
{
  int   k1, k2, kw, kf;
  float dist;

  if (n1 > n2) {
    k1   = n2;
    k2   = n1;
    dist = 1.0 - d;
  } else {
    k1   = n1;
    k2   = n2;
    dist = d;
  }
  kw = nodeTable[k1-1];

  /* first time node has been hit */
  if (kw == 0) {
    nodeTable[k1-1] = *nedgeX + 1;
    goto newX;
  }

  /* old first (lower) node */
  do {
    if (edgeX[kw-1].n2 == k2) {
      *nc = kw;
      return;
    }
    kf = kw;
    kw = edgeX[kw-1].thread;
  } while (kw != 0);
  edgeX[kf-1].thread = *nedgeX + 1;

newX:
  edgeX[*nedgeX].n1     = k1;
  edgeX[*nedgeX].n2     = k2;
  edgeX[*nedgeX].interp = dist;
  edgeX[*nedgeX].thread = 0;
  *nedgeX += 1;
  *nc = *nedgeX;
}


/*
 *      This routine generates the surface created by intersecting
 *      the surface at cut into the tetrahedron with the vertices in cvertex:
 *
 *                         1
 *                        / \~- _ 4
 *                       /   \   /
 *                      /     \ /
 *                     2-------3
 *
 *      can add up to 4 edgeXs
 *                    2 tris
 */

void
FX_CutTet(int kc, float cut, float *cvertex, int *vertices, int *ntris, 
          FXfacet *tris, int *nedgeX, FXinter *edgeX, int *nodeTable)
{
  int    ninter, i, n1, n2, index, nc[4];
  float  dist;
  static int ncell[4]    = { 0, 0, 1, 2 };
  static int npcel[2][3] = { 1,2,3, 1,3,4 };
  static char npts[16]   = { 0, 3, 3, 4, 3, 4, 4, 3, 3, 4, 4, 3, 4, 3, 3, 0 };
  static char ecut[16][4][2] = {
                            0, 0, 0, 0, 0, 0, 0, 0,  1, 2, 1, 4, 1, 3, 0, 0,
                            2, 1, 2, 3, 2, 4, 0, 0,  1, 3, 2, 3, 2, 4, 1, 4,
                            3, 2, 3, 1, 3, 4, 0, 0,  1, 2, 1, 4, 3, 4, 3, 2,
                            2, 1, 3, 1, 3, 4, 2, 4,  3, 4, 2, 4, 1, 4, 0, 0,
                            4, 3, 4, 1, 4, 2, 0, 0,  1, 2, 4, 2, 4, 3, 1, 3,
                            2, 1, 2, 3, 4, 3, 4, 1,  1, 3, 2, 3, 4, 3, 0, 0,
                            3, 2, 3, 1, 4, 1, 4, 2,  1, 2, 4, 2, 3, 2, 0, 0,
                            2, 1, 3, 1, 4, 1, 0, 0,  0, 0, 0, 0, 0, 0, 0, 0 };

                         index  = 0;
  if (cvertex[0] >= cut) index += 1;
  if (cvertex[1] >= cut) index += 2;
  if (cvertex[2] >= cut) index += 4;
  if (cvertex[3] >= cut) index += 8;

  /* make the 2D nodes at edge intersections */

  ninter = npts[index];
  if (ninter == 0) return;

  for (i = 0; i < ninter; i++) {
    n1   = ecut[index][i][0]-1;
    n2   = ecut[index][i][1]-1;
    dist = (cut-cvertex[n1])/(cvertex[n2]-cvertex[n1]);
    FX_2DNode(vertices[n1], vertices[n2], dist, &nc[i],
              nedgeX, edgeX, nodeTable);
  }

  /* make the new triangles */

  for (i = 0; i < ncell[ninter-1]; i++) {
    tris[*ntris].nodes[0] = nc[npcel[i][0]-1];
    tris[*ntris].nodes[1] = nc[npcel[i][1]-1];
    tris[*ntris].nodes[2] = nc[npcel[i][2]-1];
    tris[*ntris].nodes[3] = kc;
    *ntris += 1;
  }
}


/*
 *      This routine generates the surface created by intersecting
 *      the surface at cut into the pyramid with the vertices in cvertex:
 *
 *                        .5`--._
 *                       4/-\- - -3
 *                       /   \   /
 *                      /     \ /
 *                     1-------2
 *
 *      can add up to 6 edgeXs
 *                    4 tris
 */

void
FX_CutPyr(int kc, float cut, float *cvertex, int *vertices, int *ntris, 
          FXfacet *tris, int *nedgeX, FXinter *edgeX, int *nodeTable)
{
  int    ninter, npoly, i, j, n1, n2, index, ist, nc[6], nnod, ni;
  float  dist;
  static int ncell[6][2] = { 0,0, 0,0, 1,1, 2,1, 3,1, 4,4 };
  static int npcel[7][3] = { 1,2,3, 1,3,4, 1,4,5, 1,2,3, 1,3,4, 1,4,6, 4,5,6 };
  static char npts[32] = { 0, 3, 3, 4, 3, 6, 4, 5, 3, 4, 6, 5,
                           4, 5, 5, 4, 4, 5, 5, 4, 5, 6, 4, 3,
                           5, 4, 6, 3, 4, 3, 3, 0 };
  static char nelem[32] = { 0, 1, 1, 1, 1, 2, 1, 1, 1, 1, 2, 1,
                            1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
                            1, 1, 1, 1, 1, 1, 1, 0 };
  static char npel[32][2] = { 0, 0, 3, 0, 3, 0, 4, 0, 3, 0, 3, 3,
                              4, 0, 5, 0, 3, 0, 4, 0, 3, 3, 5, 0,
                              4, 0, 5, 0, 5, 0, 4, 0, 4, 0, 5, 0,
                              5, 0, 4, 0, 5, 0, 6, 0, 4, 0, 3, 0,
                              5, 0, 4, 0, 6, 0, 3, 0, 4, 0, 3, 0,
                              3, 0, 0, 0 };
  static char ecut[32][6][2] = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                                 1, 2, 1, 5, 1, 4, 0, 0, 0, 0, 0, 0,
                                 2, 5, 2, 1, 2, 3, 0, 0, 0, 0, 0, 0,
                                 2, 5, 1, 5, 1, 4, 2, 3, 0, 0, 0, 0,
                                 3, 5, 3, 2, 3, 4, 0, 0, 0, 0, 0, 0,
                                 1, 2, 1, 5, 1, 4, 3, 5, 3, 2, 3, 4,
                                 2, 5, 2, 1, 3, 4, 3, 5, 0, 0, 0, 0,
                                 2, 5, 1, 5, 1, 4, 3, 4, 3, 5, 0, 0,
                                 4, 5, 4, 3, 4, 1, 0, 0, 0, 0, 0, 0,
                                 1, 2, 1, 5, 4, 5, 4, 3, 0, 0, 0, 0,
                                 2, 5, 2, 1, 2, 3, 4, 5, 4, 3, 4, 1,
                                 2, 5, 1, 5, 4, 5, 4, 3, 2, 3, 0, 0,
                                 3, 5, 3, 2, 4, 1, 4, 5, 0, 0, 0, 0,
                                 1, 2, 1, 5, 4, 5, 3, 5, 3, 2, 0, 0,
                                 2, 5, 2, 1, 4, 1, 4, 5, 3, 5, 0, 0,
                                 2, 5, 1, 5, 4, 5, 3, 5, 0, 0, 0, 0,
                                 5, 1, 5, 2, 5, 3, 5, 4, 0, 0, 0, 0,
                                 1, 2, 5, 2, 5, 3, 5, 4, 1, 4, 0, 0,
                                 5, 1, 2, 1, 2, 3, 5, 3, 5, 4, 0, 0,
                                 2, 3, 5, 3, 5, 4, 1, 4, 0, 0, 0, 0,
                                 5, 1, 5, 2, 3, 2, 3, 4, 5, 4, 0, 0,
                                 1, 2, 5, 2, 3, 2, 3, 4, 5, 4, 1, 4,
                                 5, 1, 2, 1, 3, 4, 5, 4, 0, 0, 0, 0,
                                 3, 4, 5, 4, 1, 4, 0, 0, 0, 0, 0, 0,
                                 5, 1, 5, 2, 5, 3, 4, 3, 4, 1, 0, 0,
                                 1, 2, 5, 2, 5, 3, 4, 3, 0, 0, 0, 0,
                                 5, 1, 2, 1, 2, 3, 5, 3, 4, 3, 4, 1,
                                 2, 3, 5, 3, 4, 3, 0, 0, 0, 0, 0, 0,
                                 5, 1, 5, 2, 3, 2, 4, 1, 0, 0, 0, 0,
                                 1, 2, 5, 2, 3, 2, 0, 0, 0, 0, 0, 0,
                                 5, 1, 2, 1, 4, 1, 0, 0, 0, 0, 0, 0,
                                 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
                                
                         index  =  0;
  if (cvertex[0] >= cut) index +=  1;
  if (cvertex[1] >= cut) index +=  2;
  if (cvertex[2] >= cut) index +=  4;
  if (cvertex[3] >= cut) index +=  8;
  if (cvertex[4] >= cut) index += 16;

  /* make the 2D nodes at edge intersections */

  ninter = npts[index];
  if (ninter == 0) return;

  for (i = 0; i < ninter; i++) {
    n1   = ecut[index][i][0]-1;
    n2   = ecut[index][i][1]-1;
    dist = (cut-cvertex[n1])/(cvertex[n2]-cvertex[n1]);
    FX_2DNode(vertices[n1], vertices[n2], dist, &nc[i], 
              nedgeX, edgeX, nodeTable);
  }

  /* make the new triangles */

  npoly = nelem[index];
  for (ni = i = 0; i < npoly; i++) {
    nnod = npel[index][i]-1;
    ist  = ncell[nnod][1]-1;
    for (j = 0; j < ncell[nnod][0]; j++, ist++) {
      tris[*ntris].nodes[0] = nc[ni+npcel[ist][0]-1];
      tris[*ntris].nodes[1] = nc[ni+npcel[ist][1]-1];
      tris[*ntris].nodes[2] = nc[ni+npcel[ist][2]-1];
      tris[*ntris].nodes[3] = kc;
      *ntris += 1;
    }
    ni += nnod+1;
  }
}


/*
 *      This routine generates the surface created by intersecting
 *      the surface at cut into the prism with the vertices in cvertex:
 *
 *                      6-------5
 *                      |\      |\
 *                      | 4- - -|-3
 *                     | .     | /
 *                     |.      |/
 *                     1-------2
 *
 *      can add up to 7 edgeXs
 *                    4 tris
 */

void
FX_CutPri(int kc, float cut, float *cvertex, int *vertices, int *ntris,
          FXfacet *tris, int *nedgeX, FXinter *edgeX, int *nodeTable)
{
  int    ninter, npoly, i, j, n1, n2, index, ist, nc[7], nnod, ni;
  float  dist;
  static int ncell[6][2] = { 0,0, 0,0, 1,1, 2,1, 3,1, 4,4 };
  static int npcel[7][3] = { 1,2,3, 1,3,4, 1,4,5, 1,2,3, 1,3,4, 1,4,6, 4,5,6 };
  static char npts[64] = { 0, 3, 3, 4, 3, 6, 4, 5, 3, 4, 6, 5, 4, 5,
                           5, 4, 3, 6, 4, 5, 4, 7, 3, 4, 6, 7, 7, 6,
                           5, 6, 4, 3, 3, 4, 6, 5, 6, 7, 7, 6, 4, 3,
                           7, 4, 5, 4, 6, 3, 4, 5, 5, 4, 5, 6, 4, 3,
                           5, 4, 6, 3, 4, 3, 3, 0 };
  static char nelem[64] = { 0, 1, 1, 1, 1, 2, 1, 1, 1, 1, 2, 1, 1, 1,
                            1, 1, 1, 2, 1, 1, 1, 2, 1, 1, 2, 2, 2, 1,
                            1, 1, 1, 1, 1, 1, 2, 1, 2, 2, 2, 1, 1, 1,
                            2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
                            1, 1, 1, 1, 1, 1, 1, 0 };
  static char npel[64][2] = { 0, 0, 3, 0, 3, 0, 4, 0, 3, 0, 3, 3, 4, 0,
                              5, 0, 3, 0, 4, 0, 3, 3, 5, 0, 4, 0, 5, 0,
                              5, 0, 4, 0, 3, 0, 3, 3, 4, 0, 5, 0, 4, 0,
                              3, 4, 3, 0, 4, 0, 3, 3, 4, 3, 4, 3, 6, 0,
                              5, 0, 6, 0, 4, 0, 3, 0, 3, 0, 4, 0, 3, 3,
                              5, 0, 3, 3, 4, 3, 3, 4, 6, 0, 4, 0, 3, 0,
                              4, 3, 4, 0, 5, 0, 4, 0, 6, 0, 3, 0, 4, 0,
                              5, 0, 5, 0, 4, 0, 5, 0, 6, 0, 4, 0, 3, 0,
                              5, 0, 4, 0, 6, 0, 3, 0, 4, 0, 3, 0, 3, 0,
                              0, 0 };
  static char ecut[64][7][2] = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                                 1, 2, 1, 6, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0,
                                 2, 5, 2, 1, 2, 3, 0, 0, 0, 0, 0, 0, 0, 0,
                                 2, 5, 1, 6, 1, 4, 2, 3, 0, 0, 0, 0, 0, 0,
                                 3, 5, 3, 2, 3, 4, 0, 0, 0, 0, 0, 0, 0, 0,
                                 1, 2, 1, 6, 1, 4, 3, 5, 3, 2, 3, 4, 0, 0,
                                 2, 5, 2, 1, 3, 4, 3, 5, 0, 0, 0, 0, 0, 0,
                                 2, 5, 1, 6, 1, 4, 3, 4, 3, 5, 0, 0, 0, 0,
                                 4, 6, 4, 3, 4, 1, 0, 0, 0, 0, 0, 0, 0, 0,
                                 1, 2, 1, 6, 4, 6, 4, 3, 0, 0, 0, 0, 0, 0,
                                 2, 5, 2, 1, 2, 3, 4, 6, 4, 3, 4, 1, 0, 0,
                                 2, 5, 1, 6, 4, 6, 4, 3, 2, 3, 0, 0, 0, 0,
                                 3, 5, 3, 2, 4, 1, 4, 6, 0, 0, 0, 0, 0, 0,
                                 1, 2, 1, 6, 4, 6, 3, 5, 3, 2, 0, 0, 0, 0,
                                 2, 5, 2, 1, 4, 1, 4, 6, 3, 5, 0, 0, 0, 0,
                                 2, 5, 1, 6, 4, 6, 3, 5, 0, 0, 0, 0, 0, 0,
                                 5, 6, 5, 2, 5, 3, 0, 0, 0, 0, 0, 0, 0, 0,
                                 1, 2, 1, 6, 1, 4, 5, 6, 5, 2, 5, 3, 0, 0,
                                 5, 6, 2, 1, 2, 3, 5, 3, 0, 0, 0, 0, 0, 0,
                                 5, 6, 1, 6, 1, 4, 2, 3, 5, 3, 0, 0, 0, 0,
                                 5, 6, 5, 2, 3, 2, 3, 4, 0, 0, 0, 0, 0, 0,
                                 1, 2, 1, 6, 1, 4, 5, 6, 5, 2, 3, 2, 3, 4,
                                 5, 6, 2, 1, 3, 4, 0, 0, 0, 0, 0, 0, 0, 0,
                                 5, 6, 1, 6, 1, 4, 3, 4, 0, 0, 0, 0, 0, 0,
                                 5, 6, 5, 2, 5, 3, 4, 6, 4, 3, 4, 1, 0, 0,
                                 1, 2, 1, 6, 4, 6, 4, 3, 5, 6, 5, 2, 5, 3,
                                 5, 6, 2, 1, 2, 3, 5, 3, 4, 6, 4, 3, 4, 1,
                                 5, 6, 1, 6, 4, 6, 4, 3, 2, 3, 5, 3, 0, 0,
                                 5, 6, 5, 2, 3, 2, 4, 1, 4, 6, 0, 0, 0, 0,
                                 1, 2, 1, 6, 4, 6, 5, 6, 5, 2, 3, 2, 0, 0,
                                 5, 6, 2, 1, 4, 1, 4, 6, 0, 0, 0, 0, 0, 0,
                                 5, 6, 1, 6, 4, 6, 0, 0, 0, 0, 0, 0, 0, 0,
                                 6, 1, 6, 5, 6, 4, 0, 0, 0, 0, 0, 0, 0, 0,
                                 1, 2, 6, 5, 6, 4, 1, 4, 0, 0, 0, 0, 0, 0,
                                 6, 1, 6, 5, 6, 4, 2, 5, 2, 1, 2, 3, 0, 0,
                                 2, 5, 6, 5, 6, 4, 1, 4, 2, 3, 0, 0, 0, 0,
                                 6, 1, 6, 5, 6, 4, 3, 5, 3, 2, 3, 4, 0, 0,
                                 1, 2, 6, 5, 6, 4, 1, 4, 3, 5, 3, 2, 3, 4,
                                 6, 1, 6, 5, 6, 4, 2, 5, 2, 1, 3, 4, 3, 5,
                                 2, 5, 6, 5, 6, 4, 1, 4, 3, 4, 3, 5, 0, 0,
                                 6, 1, 6, 5, 4, 3, 4, 1, 0, 0, 0, 0, 0, 0,
                                 1, 2, 6, 5, 4, 3, 0, 0, 0, 0, 0, 0, 0, 0,
                                 6, 1, 6, 5, 4, 3, 4, 1, 2, 5, 2, 1, 2, 3,
                                 2, 5, 6, 5, 4, 3, 2, 3, 0, 0, 0, 0, 0, 0,
                                 6, 1, 6, 5, 3, 5, 3, 2, 4, 1, 0, 0, 0, 0,
                                 1, 2, 6, 5, 3, 5, 3, 2, 0, 0, 0, 0, 0, 0,
                                 6, 1, 6, 5, 3, 5, 2, 5, 2, 1, 4, 1, 0, 0,
                                 2, 5, 6, 5, 3, 5, 0, 0, 0, 0, 0, 0, 0, 0,
                                 6, 1, 5, 2, 5, 3, 6, 4, 0, 0, 0, 0, 0, 0,
                                 1, 2, 5, 2, 5, 3, 6, 4, 1, 4, 0, 0, 0, 0,
                                 6, 1, 2, 1, 2, 3, 5, 3, 6, 4, 0, 0, 0, 0,
                                 2, 3, 5, 3, 6, 4, 1, 4, 0, 0, 0, 0, 0, 0,
                                 6, 1, 5, 2, 3, 2, 3, 4, 6, 4, 0, 0, 0, 0,
                                 1, 2, 5, 2, 3, 2, 3, 4, 6, 4, 1, 4, 0, 0,
                                 6, 1, 2, 1, 3, 4, 6, 4, 0, 0, 0, 0, 0, 0,
                                 3, 4, 6, 4, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0,
                                 6, 1, 5, 2, 5, 3, 4, 3, 4, 1, 0, 0, 0, 0,
                                 1, 2, 5, 2, 5, 3, 4, 3, 0, 0, 0, 0, 0, 0,
                                 6, 1, 2, 1, 2, 3, 5, 3, 4, 3, 4, 1, 0, 0,
                                 2, 3, 5, 3, 4, 3, 0, 0, 0, 0, 0, 0, 0, 0,
                                 6, 1, 5, 2, 3, 2, 4, 1, 0, 0, 0, 0, 0, 0,
                                 1, 2, 5, 2, 3, 2, 0, 0, 0, 0, 0, 0, 0, 0,
                                 6, 1, 2, 1, 4, 1, 0, 0, 0, 0, 0, 0, 0, 0,
                                 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
                               
                         index  =  0;
  if (cvertex[0] >= cut) index +=  1;
  if (cvertex[1] >= cut) index +=  2;
  if (cvertex[2] >= cut) index +=  4;
  if (cvertex[3] >= cut) index +=  8;
  if (cvertex[4] >= cut) index += 16;
  if (cvertex[5] >= cut) index += 32;

  /* make the 2D nodes at edge intersections */

  ninter = npts[index];
  if (ninter == 0) return;

  for (i = 0; i < ninter; i++) {
    n1   = ecut[index][i][0]-1;
    n2   = ecut[index][i][1]-1;
    dist = (cut-cvertex[n1])/(cvertex[n2]-cvertex[n1]);
    FX_2DNode(vertices[n1], vertices[n2], dist, &nc[i],
              nedgeX, edgeX, nodeTable);
  }

  /* make the new triangles */

  npoly = nelem[index];
  for (ni = i = 0; i < npoly; i++) {
    nnod = npel[index][i]-1;
    ist  = ncell[nnod][1]-1;
    for (j = 0; j < ncell[nnod][0]; j++, ist++) {
      tris[*ntris].nodes[0] = nc[ni+npcel[ist][0]-1];
      tris[*ntris].nodes[1] = nc[ni+npcel[ist][1]-1];
      tris[*ntris].nodes[2] = nc[ni+npcel[ist][2]-1];
      tris[*ntris].nodes[3] = kc;
      *ntris += 1;
    }
    ni += nnod+1;
  }
}


/*
 *      This routine generates the surface created by intersecting
 *      the surface at cut into the hexahedron with the vertices in cvertex:
 *
 *                       8-------7
 *                      /:      /|
 *                     / :     / |
 *                    5-------6  |
 *                    |  4- - | -3
 *                    | .     | /
 *                    |.      |/
 *                    1-------2
 *
 *      can add up to 12 edgeXs
 *                     5 tris
 */

void
FX_CutHex(int kc, float cut, float *cvertex, int *vertices, int *ntris, 
          FXfacet *tris, int *nedgeX, FXinter *edgeX, int *nodeTable)
{
  int    ninter, npoly, i, j, n1, n2, index, ist, nc[12], nnod, ni;
  float  dist;
  static int ncell[7][2]  = { 0,0, 0,0, 1,1, 2,1, 3,1, 4,4, 5,8 };
  static int npcel[12][3] = { 1,2,3, 1,3,4, 1,4,5, 1,2,3, 1,3,4, 1,4,6, 4,5,6,
                              1,2,3, 1,3,4, 1,4,7, 4,5,7, 5,6,7 };
  static char npts[256] = {
        0,  3,  3,  4,  3,  6,  4,  5,  3,  4,  6,  5,  4,  5,  5,  4,  3,  4,
        6,  5,  6,  7,  7,  6,  6,  5,  9,  6,  7,  6,  8,  5,  3,  6,  4,  5,
        6,  9,  5,  6,  6,  7,  7,  6,  7,  8,  6,  5,  4,  5,  5,  4,  7,  8,
        6,  5,  7,  6,  8,  5,  8,  7,  7,  4,  3,  6,  6,  7,  4,  7,  5,  6,
        6,  7,  9,  8,  5,  6,  6,  5,  6,  7,  9,  8,  7,  8,  8,  7,  9,  8,
       12,  9,  8,  7,  9,  6,  4,  7,  5,  6,  5,  8,  4,  5,  7,  8,  8,  7,
        6,  7,  5,  4,  5,  6,  6,  5,  6,  7,  5,  4,  8,  7,  9,  6,  7,  6,
        6,  3,  3,  6,  6,  7,  6,  9,  7,  8,  4,  5,  7,  6,  5,  6,  6,  5,
        4,  5,  7,  6,  7,  8,  8,  7,  5,  4,  8,  5,  6,  5,  7,  4,  6,  9,
        7,  8,  9, 12,  8,  9,  7,  8,  8,  7,  8,  9,  7,  6,  5,  6,  6,  5,
        8,  9,  7,  6,  6,  5,  7,  4,  7,  6,  6,  3,  4,  7,  7,  8,  5,  8,
        6,  7,  5,  6,  8,  7,  4,  5,  5,  4,  5,  6,  8,  7,  6,  7,  7,  6,
        6,  5,  9,  6,  5,  4,  6,  3,  5,  8,  6,  7,  6,  9,  5,  6,  6,  7,
        7,  6,  5,  6,  4,  3,  4,  5,  5,  4,  5,  6,  4,  3,  5,  4,  6,  3,
        4,  3,  3,  0 };
  static char nelem[256] = {
        0, 1, 1, 1, 1, 2, 1, 1, 1, 1, 2, 1, 1, 1, 1, 1, 1, 1, 2, 1, 2, 2, 2, 1,
        2, 1, 3, 1, 2, 1, 2, 1, 1, 2, 1, 1, 2, 3, 1, 1, 2, 2, 2, 1, 2, 2, 1, 1,
        1, 1, 1, 1, 2, 2, 1, 1, 2, 1, 2, 1, 2, 1, 1, 1, 1, 2, 2, 2, 1, 2, 1, 1,
        2, 2, 3, 2, 1, 1, 1, 1, 2, 2, 3, 2, 2, 2, 2, 1, 3, 2, 4, 2, 2, 1, 2, 1,
        1, 2, 1, 1, 1, 2, 1, 1, 2, 2, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
        2, 1, 2, 1, 1, 2, 1, 1, 1, 2, 2, 2, 2, 3, 2, 2, 1, 1, 2, 1, 1, 1, 1, 1,
        1, 1, 2, 1, 2, 2, 2, 1, 1, 1, 2, 1, 1, 1, 1, 1, 2, 3, 2, 2, 3, 4, 2, 2,
        2, 2, 2, 1, 2, 2, 1, 1, 1, 1, 1, 1, 2, 2, 1, 1, 1, 1, 1, 1, 1, 1, 2, 1,
        1, 2, 2, 2, 1, 2, 1, 1, 1, 1, 2, 1, 1, 1, 1, 1, 1, 1, 2, 1, 1, 1, 1, 2,
        1, 1, 2, 1, 1, 1, 1, 1, 1, 2, 1, 1, 1, 2, 1, 1, 1, 1, 1, 2, 1, 1, 1, 1,
        1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0 };
  static char npel[256][4] = {
        0, 0, 0, 0, 3, 0, 0, 0, 3, 0, 0, 0, 4, 0, 0, 0, 3, 0, 0, 0, 3, 3, 0, 0,
        4, 0, 0, 0, 5, 0, 0, 0, 3, 0, 0, 0, 4, 0, 0, 0, 3, 3, 0, 0, 5, 0, 0, 0,
        4, 0, 0, 0, 5, 0, 0, 0, 5, 0, 0, 0, 4, 0, 0, 0, 3, 0, 0, 0, 4, 0, 0, 0,
        3, 3, 0, 0, 5, 0, 0, 0, 3, 3, 0, 0, 4, 3, 0, 0, 3, 4, 0, 0, 6, 0, 0, 0,
        3, 3, 0, 0, 5, 0, 0, 0, 3, 3, 3, 0, 6, 0, 0, 0, 3, 4, 0, 0, 6, 0, 0, 0,
        3, 5, 0, 0, 5, 0, 0, 0, 3, 0, 0, 0, 3, 3, 0, 0, 4, 0, 0, 0, 5, 0, 0, 0,
        3, 3, 0, 0, 3, 3, 3, 0, 5, 0, 0, 0, 6, 0, 0, 0, 3, 3, 0, 0, 4, 3, 0, 0,
        4, 3, 0, 0, 6, 0, 0, 0, 3, 4, 0, 0, 5, 3, 0, 0, 6, 0, 0, 0, 5, 0, 0, 0,
        4, 0, 0, 0, 5, 0, 0, 0, 5, 0, 0, 0, 4, 0, 0, 0, 4, 3, 0, 0, 5, 3, 0, 0,
        6, 0, 0, 0, 5, 0, 0, 0, 4, 3, 0, 0, 6, 0, 0, 0, 5, 3, 0, 0, 5, 0, 0, 0,
        4, 4, 0, 0, 7, 0, 0, 0, 7, 0, 0, 0, 4, 0, 0, 0, 3, 0, 0, 0, 3, 3, 0, 0,
        3, 3, 0, 0, 4, 3, 0, 0, 4, 0, 0, 0, 3, 4, 0, 0, 5, 0, 0, 0, 6, 0, 0, 0,
        3, 3, 0, 0, 4, 3, 0, 0, 3, 3, 3, 0, 5, 3, 0, 0, 5, 0, 0, 0, 6, 0, 0, 0,
        6, 0, 0, 0, 5, 0, 0, 0, 3, 3, 0, 0, 4, 3, 0, 0, 3, 3, 3, 0, 5, 3, 0, 0,
        3, 4, 0, 0, 4, 4, 0, 0, 3, 5, 0, 0, 7, 0, 0, 0, 3, 3, 3, 0, 5, 3, 0, 0,
        3, 3, 3, 3, 6, 3, 0, 0, 3, 5, 0, 0, 7, 0, 0, 0, 3, 6, 0, 0, 6, 0, 0, 0,
        4, 0, 0, 0, 3, 4, 0, 0, 5, 0, 0, 0, 6, 0, 0, 0, 5, 0, 0, 0, 3, 5, 0, 0,
        4, 0, 0, 0, 5, 0, 0, 0, 4, 3, 0, 0, 4, 4, 0, 0, 5, 3, 0, 0, 7, 0, 0, 0,
        6, 0, 0, 0, 7, 0, 0, 0, 5, 0, 0, 0, 4, 0, 0, 0, 5, 0, 0, 0, 6, 0, 0, 0,
        6, 0, 0, 0, 5, 0, 0, 0, 6, 0, 0, 0, 7, 0, 0, 0, 5, 0, 0, 0, 4, 0, 0, 0,
        5, 3, 0, 0, 7, 0, 0, 0, 6, 3, 0, 0, 6, 0, 0, 0, 7, 0, 0, 0, 3, 3, 0, 0,
        6, 0, 0, 0, 3, 0, 0, 0, 3, 0, 0, 0, 3, 3, 0, 0, 3, 3, 0, 0, 4, 3, 0, 0,
        3, 3, 0, 0, 3, 3, 3, 0, 4, 3, 0, 0, 5, 3, 0, 0, 4, 0, 0, 0, 5, 0, 0, 0,
        3, 4, 0, 0, 6, 0, 0, 0, 5, 0, 0, 0, 6, 0, 0, 0, 6, 0, 0, 0, 5, 0, 0, 0,
        4, 0, 0, 0, 5, 0, 0, 0, 4, 3, 0, 0, 6, 0, 0, 0, 4, 3, 0, 0, 5, 3, 0, 0,
        4, 4, 0, 0, 7, 0, 0, 0, 5, 0, 0, 0, 4, 0, 0, 0, 5, 3, 0, 0, 5, 0, 0, 0,
        6, 0, 0, 0, 5, 0, 0, 0, 7, 0, 0, 0, 4, 0, 0, 0, 3, 3, 0, 0, 3, 3, 3, 0,
        4, 3, 0, 0, 5, 3, 0, 0, 3, 3, 3, 0, 3, 3, 3, 3, 5, 3, 0, 0, 6, 3, 0, 0,
        3, 4, 0, 0, 5, 3, 0, 0, 4, 4, 0, 0, 7, 0, 0, 0, 3, 5, 0, 0, 6, 3, 0, 0,
        7, 0, 0, 0, 6, 0, 0, 0, 5, 0, 0, 0, 6, 0, 0, 0, 6, 0, 0, 0, 5, 0, 0, 0,
        5, 3, 0, 0, 6, 3, 0, 0, 7, 0, 0, 0, 6, 0, 0, 0, 6, 0, 0, 0, 5, 0, 0, 0,
        7, 0, 0, 0, 4, 0, 0, 0, 7, 0, 0, 0, 6, 0, 0, 0, 3, 3, 0, 0, 3, 0, 0, 0,
        4, 0, 0, 0, 3, 4, 0, 0, 3, 4, 0, 0, 4, 4, 0, 0, 5, 0, 0, 0, 3, 5, 0, 0,
        6, 0, 0, 0, 7, 0, 0, 0, 5, 0, 0, 0, 6, 0, 0, 0, 3, 5, 0, 0, 7, 0, 0, 0,
        4, 0, 0, 0, 5, 0, 0, 0, 5, 0, 0, 0, 4, 0, 0, 0, 5, 0, 0, 0, 6, 0, 0, 0,
        5, 3, 0, 0, 7, 0, 0, 0, 6, 0, 0, 0, 7, 0, 0, 0, 7, 0, 0, 0, 3, 3, 0, 0,
        6, 0, 0, 0, 5, 0, 0, 0, 6, 3, 0, 0, 6, 0, 0, 0, 5, 0, 0, 0, 4, 0, 0, 0,
        6, 0, 0, 0, 3, 0, 0, 0, 5, 0, 0, 0, 3, 5, 0, 0, 6, 0, 0, 0, 7, 0, 0, 0,
        6, 0, 0, 0, 3, 6, 0, 0, 5, 0, 0, 0, 6, 0, 0, 0, 6, 0, 0, 0, 7, 0, 0, 0,
        7, 0, 0, 0, 3, 3, 0, 0, 5, 0, 0, 0, 6, 0, 0, 0, 4, 0, 0, 0, 3, 0, 0, 0,
        4, 0, 0, 0, 5, 0, 0, 0, 5, 0, 0, 0, 4, 0, 0, 0, 5, 0, 0, 0, 6, 0, 0, 0,
        4, 0, 0, 0, 3, 0, 0, 0, 5, 0, 0, 0, 4, 0, 0, 0, 6, 0, 0, 0, 3, 0, 0, 0,
        4, 0, 0, 0, 3, 0, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0 };
  static char ecut[256][12][2] = {
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        1, 2, 1, 5, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        2, 6, 2, 1, 2, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        2, 6, 1, 5, 1, 4, 2, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        3, 7, 3, 2, 3, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        1, 2, 1, 5, 1, 4, 3, 7, 3, 2, 3, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        2, 6, 2, 1, 3, 4, 3, 7, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        2, 6, 1, 5, 1, 4, 3, 4, 3, 7, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        4, 8, 4, 3, 4, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        1, 2, 1, 5, 4, 8, 4, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        2, 6, 2, 1, 2, 3, 4, 8, 4, 3, 4, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        2, 6, 1, 5, 4, 8, 4, 3, 2, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        3, 7, 3, 2, 4, 1, 4, 8, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        1, 2, 1, 5, 4, 8, 3, 7, 3, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        2, 6, 2, 1, 4, 1, 4, 8, 3, 7, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        2, 6, 1, 5, 4, 8, 3, 7, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        5, 1, 5, 6, 5, 8, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        1, 2, 5, 6, 5, 8, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        5, 1, 5, 6, 5, 8, 2, 6, 2, 1, 2, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        2, 6, 5, 6, 5, 8, 1, 4, 2, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        5, 1, 5, 6, 5, 8, 3, 7, 3, 2, 3, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        1, 2, 5, 6, 5, 8, 1, 4, 3, 7, 3, 2, 3, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        5, 1, 5, 6, 5, 8, 2, 6, 2, 1, 3, 4, 3, 7, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        2, 6, 5, 6, 5, 8, 1, 4, 3, 4, 3, 7, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        5, 1, 5, 6, 5, 8, 4, 8, 4, 3, 4, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        1, 2, 5, 6, 5, 8, 4, 8, 4, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        5, 1, 5, 6, 5, 8, 2, 6, 2, 1, 2, 3, 4, 8, 4, 3, 4, 1, 0, 0, 0, 0, 0, 0,
        2, 6, 5, 6, 5, 8, 4, 8, 4, 3, 2, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        5, 1, 5, 6, 5, 8, 3, 7, 3, 2, 4, 1, 4, 8, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        1, 2, 5, 6, 5, 8, 4, 8, 3, 7, 3, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        5, 1, 5, 6, 5, 8, 2, 6, 2, 1, 4, 1, 4, 8, 3, 7, 0, 0, 0, 0, 0, 0, 0, 0,
        2, 6, 5, 6, 5, 8, 4, 8, 3, 7, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        6, 5, 6, 2, 6, 7, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        1, 2, 1, 5, 1, 4, 6, 5, 6, 2, 6, 7, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        6, 5, 2, 1, 2, 3, 6, 7, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        6, 5, 1, 5, 1, 4, 2, 3, 6, 7, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        6, 5, 6, 2, 6, 7, 3, 7, 3, 2, 3, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        1, 2, 1, 5, 1, 4, 6, 5, 6, 2, 6, 7, 3, 7, 3, 2, 3, 4, 0, 0, 0, 0, 0, 0,
        6, 5, 2, 1, 3, 4, 3, 7, 6, 7, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        6, 5, 1, 5, 1, 4, 3, 4, 3, 7, 6, 7, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        6, 5, 6, 2, 6, 7, 4, 8, 4, 3, 4, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        1, 2, 1, 5, 4, 8, 4, 3, 6, 5, 6, 2, 6, 7, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        6, 5, 2, 1, 2, 3, 6, 7, 4, 8, 4, 3, 4, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        6, 5, 1, 5, 4, 8, 4, 3, 2, 3, 6, 7, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        6, 5, 6, 2, 6, 7, 3, 7, 3, 2, 4, 1, 4, 8, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        1, 2, 1, 5, 4, 8, 3, 7, 3, 2, 6, 5, 6, 2, 6, 7, 0, 0, 0, 0, 0, 0, 0, 0,
        6, 5, 2, 1, 4, 1, 4, 8, 3, 7, 6, 7, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        6, 5, 1, 5, 4, 8, 3, 7, 6, 7, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        5, 1, 6, 2, 6, 7, 5, 8, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        1, 2, 6, 2, 6, 7, 5, 8, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        5, 1, 2, 1, 2, 3, 6, 7, 5, 8, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        2, 3, 6, 7, 5, 8, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        5, 1, 6, 2, 6, 7, 5, 8, 3, 7, 3, 2, 3, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        1, 2, 6, 2, 6, 7, 5, 8, 1, 4, 3, 7, 3, 2, 3, 4, 0, 0, 0, 0, 0, 0, 0, 0,
        5, 1, 2, 1, 3, 4, 3, 7, 6, 7, 5, 8, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        3, 7, 6, 7, 5, 8, 1, 4, 3, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        5, 1, 6, 2, 6, 7, 5, 8, 4, 8, 4, 3, 4, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        1, 2, 6, 2, 6, 7, 5, 8, 4, 8, 4, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        5, 1, 2, 1, 2, 3, 6, 7, 5, 8, 4, 8, 4, 3, 4, 1, 0, 0, 0, 0, 0, 0, 0, 0,
        2, 3, 6, 7, 5, 8, 4, 8, 4, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        5, 1, 6, 2, 6, 7, 5, 8, 3, 7, 3, 2, 4, 1, 4, 8, 0, 0, 0, 0, 0, 0, 0, 0,
        1, 2, 6, 2, 6, 7, 5, 8, 4, 8, 3, 7, 3, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        5, 1, 2, 1, 4, 1, 4, 8, 3, 7, 6, 7, 5, 8, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        3, 7, 6, 7, 5, 8, 4, 8, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        7, 6, 7, 3, 7, 8, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        1, 2, 1, 5, 1, 4, 7, 6, 7, 3, 7, 8, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        2, 6, 2, 1, 2, 3, 7, 6, 7, 3, 7, 8, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        2, 6, 1, 5, 1, 4, 2, 3, 7, 6, 7, 3, 7, 8, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        7, 6, 3, 2, 3, 4, 7, 8, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        1, 2, 1, 5, 1, 4, 7, 6, 3, 2, 3, 4, 7, 8, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        2, 6, 2, 1, 3, 4, 7, 8, 7, 6, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        2, 6, 1, 5, 1, 4, 3, 4, 7, 8, 7, 6, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        7, 6, 7, 3, 7, 8, 4, 8, 4, 3, 4, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        1, 2, 1, 5, 4, 8, 4, 3, 7, 6, 7, 3, 7, 8, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        2, 6, 2, 1, 2, 3, 7, 6, 7, 3, 7, 8, 4, 8, 4, 3, 4, 1, 0, 0, 0, 0, 0, 0,
        2, 6, 1, 5, 4, 8, 4, 3, 2, 3, 7, 6, 7, 3, 7, 8, 0, 0, 0, 0, 0, 0, 0, 0,
        7, 6, 3, 2, 4, 1, 4, 8, 7, 8, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        1, 2, 1, 5, 4, 8, 7, 8, 7, 6, 3, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        2, 6, 2, 1, 4, 1, 4, 8, 7, 8, 7, 6, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        2, 6, 1, 5, 4, 8, 7, 8, 7, 6, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        5, 1, 5, 6, 5, 8, 7, 6, 7, 3, 7, 8, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        1, 2, 5, 6, 5, 8, 1, 4, 7, 6, 7, 3, 7, 8, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        5, 1, 5, 6, 5, 8, 2, 6, 2, 1, 2, 3, 7, 6, 7, 3, 7, 8, 0, 0, 0, 0, 0, 0,
        2, 6, 5, 6, 5, 8, 1, 4, 2, 3, 7, 6, 7, 3, 7, 8, 0, 0, 0, 0, 0, 0, 0, 0,
        5, 1, 5, 6, 5, 8, 7, 6, 3, 2, 3, 4, 7, 8, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        1, 2, 5, 6, 5, 8, 1, 4, 7, 6, 3, 2, 3, 4, 7, 8, 0, 0, 0, 0, 0, 0, 0, 0,
        5, 1, 5, 6, 5, 8, 2, 6, 2, 1, 3, 4, 7, 8, 7, 6, 0, 0, 0, 0, 0, 0, 0, 0,
        2, 6, 5, 6, 5, 8, 1, 4, 3, 4, 7, 8, 7, 6, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        5, 1, 5, 6, 5, 8, 7, 6, 7, 3, 7, 8, 4, 8, 4, 3, 4, 1, 0, 0, 0, 0, 0, 0,
        1, 2, 5, 6, 5, 8, 4, 8, 4, 3, 7, 6, 7, 3, 7, 8, 0, 0, 0, 0, 0, 0, 0, 0,
        5, 1, 5, 6, 5, 8, 2, 6, 2, 1, 2, 3, 7, 6, 7, 3, 7, 8, 4, 8, 4, 3, 4, 1,
        2, 6, 5, 6, 5, 8, 4, 8, 4, 3, 2, 3, 7, 6, 7, 3, 7, 8, 0, 0, 0, 0, 0, 0,
        5, 1, 5, 6, 5, 8, 7, 6, 3, 2, 4, 1, 4, 8, 7, 8, 0, 0, 0, 0, 0, 0, 0, 0,
        1, 2, 5, 6, 5, 8, 4, 8, 7, 8, 7, 6, 3, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        5, 1, 5, 6, 5, 8, 2, 6, 2, 1, 4, 1, 4, 8, 7, 8, 7, 6, 0, 0, 0, 0, 0, 0,
        2, 6, 5, 6, 5, 8, 4, 8, 7, 8, 7, 6, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        6, 5, 6, 2, 7, 3, 7, 8, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        1, 2, 1, 5, 1, 4, 6, 5, 6, 2, 7, 3, 7, 8, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        6, 5, 2, 1, 2, 3, 7, 3, 7, 8, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        6, 5, 1, 5, 1, 4, 2, 3, 7, 3, 7, 8, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        6, 5, 6, 2, 3, 2, 3, 4, 7, 8, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        1, 2, 1, 5, 1, 4, 6, 5, 6, 2, 3, 2, 3, 4, 7, 8, 0, 0, 0, 0, 0, 0, 0, 0,
        6, 5, 2, 1, 3, 4, 7, 8, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        6, 5, 1, 5, 1, 4, 3, 4, 7, 8, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        6, 5, 6, 2, 7, 3, 7, 8, 4, 8, 4, 3, 4, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        1, 2, 1, 5, 4, 8, 4, 3, 6, 5, 6, 2, 7, 3, 7, 8, 0, 0, 0, 0, 0, 0, 0, 0,
        6, 5, 2, 1, 2, 3, 7, 3, 7, 8, 4, 8, 4, 3, 4, 1, 0, 0, 0, 0, 0, 0, 0, 0,
        6, 5, 1, 5, 4, 8, 4, 3, 2, 3, 7, 3, 7, 8, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        6, 5, 6, 2, 3, 2, 4, 1, 4, 8, 7, 8, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        1, 2, 1, 5, 4, 8, 7, 8, 6, 5, 6, 2, 3, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        6, 5, 2, 1, 4, 1, 4, 8, 7, 8, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        6, 5, 1, 5, 4, 8, 7, 8, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        5, 1, 6, 2, 7, 3, 7, 8, 5, 8, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        1, 2, 6, 2, 7, 3, 7, 8, 5, 8, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        5, 1, 2, 1, 2, 3, 7, 3, 7, 8, 5, 8, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        2, 3, 7, 3, 7, 8, 5, 8, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        5, 1, 6, 2, 3, 2, 3, 4, 7, 8, 5, 8, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        1, 2, 6, 2, 3, 2, 3, 4, 7, 8, 5, 8, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        5, 1, 2, 1, 3, 4, 7, 8, 5, 8, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        3, 4, 7, 8, 5, 8, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        5, 1, 6, 2, 7, 3, 7, 8, 5, 8, 4, 8, 4, 3, 4, 1, 0, 0, 0, 0, 0, 0, 0, 0,
        1, 2, 6, 2, 7, 3, 7, 8, 5, 8, 4, 8, 4, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        5, 1, 2, 1, 2, 3, 7, 3, 7, 8, 5, 8, 4, 8, 4, 3, 4, 1, 0, 0, 0, 0, 0, 0,
        2, 3, 7, 3, 7, 8, 5, 8, 4, 8, 4, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        5, 1, 6, 2, 3, 2, 4, 1, 4, 8, 7, 8, 5, 8, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        1, 2, 6, 2, 3, 2, 4, 8, 7, 8, 5, 8, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        5, 1, 2, 1, 4, 1, 4, 8, 7, 8, 5, 8, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        4, 8, 7, 8, 5, 8, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        8, 7, 8, 4, 8, 5, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        1, 2, 1, 5, 1, 4, 8, 7, 8, 4, 8, 5, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        2, 6, 2, 1, 2, 3, 8, 7, 8, 4, 8, 5, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        2, 6, 1, 5, 1, 4, 2, 3, 8, 7, 8, 4, 8, 5, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        3, 7, 3, 2, 3, 4, 8, 7, 8, 4, 8, 5, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        1, 2, 1, 5, 1, 4, 3, 7, 3, 2, 3, 4, 8, 7, 8, 4, 8, 5, 0, 0, 0, 0, 0, 0,
        2, 6, 2, 1, 3, 4, 3, 7, 8, 7, 8, 4, 8, 5, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        2, 6, 1, 5, 1, 4, 3, 4, 3, 7, 8, 7, 8, 4, 8, 5, 0, 0, 0, 0, 0, 0, 0, 0,
        8, 7, 4, 3, 4, 1, 8, 5, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        1, 2, 1, 5, 8, 5, 8, 7, 4, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        2, 6, 2, 1, 2, 3, 8, 7, 4, 3, 4, 1, 8, 5, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        2, 6, 1, 5, 8, 5, 8, 7, 4, 3, 2, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        3, 7, 3, 2, 4, 1, 8, 5, 8, 7, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        1, 2, 1, 5, 8, 5, 8, 7, 3, 7, 3, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        2, 6, 2, 1, 4, 1, 8, 5, 8, 7, 3, 7, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        2, 6, 1, 5, 8, 5, 8, 7, 3, 7, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        5, 1, 5, 6, 8, 7, 8, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        1, 2, 5, 6, 8, 7, 8, 4, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        5, 1, 5, 6, 8, 7, 8, 4, 2, 6, 2, 1, 2, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        2, 6, 5, 6, 8, 7, 8, 4, 1, 4, 2, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        5, 1, 5, 6, 8, 7, 8, 4, 3, 7, 3, 2, 3, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        1, 2, 5, 6, 8, 7, 8, 4, 1, 4, 3, 7, 3, 2, 3, 4, 0, 0, 0, 0, 0, 0, 0, 0,
        5, 1, 5, 6, 8, 7, 8, 4, 2, 6, 2, 1, 3, 4, 3, 7, 0, 0, 0, 0, 0, 0, 0, 0,
        2, 6, 5, 6, 8, 7, 8, 4, 1, 4, 3, 4, 3, 7, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        5, 1, 5, 6, 8, 7, 4, 3, 4, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        1, 2, 5, 6, 8, 7, 4, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        5, 1, 5, 6, 8, 7, 4, 3, 4, 1, 2, 6, 2, 1, 2, 3, 0, 0, 0, 0, 0, 0, 0, 0,
        2, 6, 5, 6, 8, 7, 4, 3, 2, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        5, 1, 5, 6, 8, 7, 3, 7, 3, 2, 4, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        1, 2, 5, 6, 8, 7, 3, 7, 3, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        5, 1, 5, 6, 8, 7, 3, 7, 2, 6, 2, 1, 4, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        2, 6, 5, 6, 8, 7, 3, 7, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        6, 5, 6, 2, 6, 7, 8, 7, 8, 4, 8, 5, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        1, 2, 1, 5, 1, 4, 6, 5, 6, 2, 6, 7, 8, 7, 8, 4, 8, 5, 0, 0, 0, 0, 0, 0,
        6, 5, 2, 1, 2, 3, 6, 7, 8, 7, 8, 4, 8, 5, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        6, 5, 1, 5, 1, 4, 2, 3, 6, 7, 8, 7, 8, 4, 8, 5, 0, 0, 0, 0, 0, 0, 0, 0,
        6, 5, 6, 2, 6, 7, 3, 7, 3, 2, 3, 4, 8, 7, 8, 4, 8, 5, 0, 0, 0, 0, 0, 0,
        1, 2, 1, 5, 1, 4, 6, 5, 6, 2, 6, 7, 3, 7, 3, 2, 3, 4, 8, 7, 8, 4, 8, 5,
        6, 5, 2, 1, 3, 4, 3, 7, 6, 7, 8, 7, 8, 4, 8, 5, 0, 0, 0, 0, 0, 0, 0, 0,
        6, 5, 1, 5, 1, 4, 3, 4, 3, 7, 6, 7, 8, 7, 8, 4, 8, 5, 0, 0, 0, 0, 0, 0,
        6, 5, 6, 2, 6, 7, 8, 7, 4, 3, 4, 1, 8, 5, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        1, 2, 1, 5, 8, 5, 8, 7, 4, 3, 6, 5, 6, 2, 6, 7, 0, 0, 0, 0, 0, 0, 0, 0,
        6, 5, 2, 1, 2, 3, 6, 7, 8, 7, 4, 3, 4, 1, 8, 5, 0, 0, 0, 0, 0, 0, 0, 0,
        6, 5, 1, 5, 8, 5, 8, 7, 4, 3, 2, 3, 6, 7, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        6, 5, 6, 2, 6, 7, 3, 7, 3, 2, 4, 1, 8, 5, 8, 7, 0, 0, 0, 0, 0, 0, 0, 0,
        1, 2, 1, 5, 8, 5, 8, 7, 3, 7, 3, 2, 6, 5, 6, 2, 6, 7, 0, 0, 0, 0, 0, 0,
        6, 5, 2, 1, 4, 1, 8, 5, 8, 7, 3, 7, 6, 7, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        6, 5, 1, 5, 8, 5, 8, 7, 3, 7, 6, 7, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        5, 1, 6, 2, 6, 7, 8, 7, 8, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        1, 2, 6, 2, 6, 7, 8, 7, 8, 4, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        5, 1, 2, 1, 2, 3, 6, 7, 8, 7, 8, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        2, 3, 6, 7, 8, 7, 8, 4, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        5, 1, 6, 2, 6, 7, 8, 7, 8, 4, 3, 7, 3, 2, 3, 4, 0, 0, 0, 0, 0, 0, 0, 0,
        1, 2, 6, 2, 6, 7, 8, 7, 8, 4, 1, 4, 3, 7, 3, 2, 3, 4, 0, 0, 0, 0, 0, 0,
        5, 1, 2, 1, 3, 4, 3, 7, 6, 7, 8, 7, 8, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        3, 7, 6, 7, 8, 7, 8, 4, 1, 4, 3, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        5, 1, 6, 2, 6, 7, 8, 7, 4, 3, 4, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        1, 2, 6, 2, 6, 7, 8, 7, 4, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        5, 1, 2, 1, 2, 3, 6, 7, 8, 7, 4, 3, 4, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        2, 3, 6, 7, 8, 7, 4, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        5, 1, 6, 2, 6, 7, 8, 7, 3, 7, 3, 2, 4, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        1, 2, 6, 2, 6, 7, 8, 7, 3, 7, 3, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        5, 1, 2, 1, 4, 1, 3, 7, 6, 7, 8, 7, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        3, 7, 6, 7, 8, 7, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        7, 6, 7, 3, 8, 4, 8, 5, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        1, 2, 1, 5, 1, 4, 7, 6, 7, 3, 8, 4, 8, 5, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        2, 6, 2, 1, 2, 3, 7, 6, 7, 3, 8, 4, 8, 5, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        2, 6, 1, 5, 1, 4, 2, 3, 7, 6, 7, 3, 8, 4, 8, 5, 0, 0, 0, 0, 0, 0, 0, 0,
        7, 6, 3, 2, 3, 4, 8, 4, 8, 5, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        1, 2, 1, 5, 1, 4, 7, 6, 3, 2, 3, 4, 8, 4, 8, 5, 0, 0, 0, 0, 0, 0, 0, 0,
        2, 6, 2, 1, 3, 4, 8, 4, 8, 5, 7, 6, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        2, 6, 1, 5, 1, 4, 3, 4, 8, 4, 8, 5, 7, 6, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        7, 6, 7, 3, 4, 3, 4, 1, 8, 5, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        1, 2, 1, 5, 8, 5, 7, 6, 7, 3, 4, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        2, 6, 2, 1, 2, 3, 7, 6, 7, 3, 4, 3, 4, 1, 8, 5, 0, 0, 0, 0, 0, 0, 0, 0,
        2, 6, 1, 5, 8, 5, 7, 6, 7, 3, 4, 3, 2, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        7, 6, 3, 2, 4, 1, 8, 5, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        1, 2, 1, 5, 8, 5, 7, 6, 3, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        2, 6, 2, 1, 4, 1, 8, 5, 7, 6, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        2, 6, 1, 5, 8, 5, 7, 6, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        5, 1, 5, 6, 7, 6, 7, 3, 8, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        1, 2, 5, 6, 7, 6, 7, 3, 8, 4, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        5, 1, 5, 6, 7, 6, 7, 3, 8, 4, 2, 6, 2, 1, 2, 3, 0, 0, 0, 0, 0, 0, 0, 0,
        2, 6, 5, 6, 7, 6, 7, 3, 8, 4, 1, 4, 2, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        5, 1, 5, 6, 7, 6, 3, 2, 3, 4, 8, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        1, 2, 5, 6, 7, 6, 3, 2, 3, 4, 8, 4, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        5, 1, 5, 6, 7, 6, 2, 6, 2, 1, 3, 4, 8, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        2, 6, 5, 6, 7, 6, 3, 4, 8, 4, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        5, 1, 5, 6, 7, 6, 7, 3, 4, 3, 4, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        1, 2, 5, 6, 7, 6, 7, 3, 4, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        5, 1, 5, 6, 7, 6, 7, 3, 4, 3, 4, 1, 2, 6, 2, 1, 2, 3, 0, 0, 0, 0, 0, 0,
        2, 6, 5, 6, 7, 6, 7, 3, 4, 3, 2, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        5, 1, 5, 6, 7, 6, 3, 2, 4, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        1, 2, 5, 6, 7, 6, 3, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        5, 1, 5, 6, 7, 6, 2, 6, 2, 1, 4, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        2, 6, 5, 6, 7, 6, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        6, 5, 6, 2, 7, 3, 8, 4, 8, 5, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        1, 2, 1, 5, 1, 4, 6, 5, 6, 2, 7, 3, 8, 4, 8, 5, 0, 0, 0, 0, 0, 0, 0, 0,
        6, 5, 2, 1, 2, 3, 7, 3, 8, 4, 8, 5, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        6, 5, 1, 5, 1, 4, 2, 3, 7, 3, 8, 4, 8, 5, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        6, 5, 6, 2, 3, 2, 3, 4, 8, 4, 8, 5, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        1, 2, 1, 5, 1, 4, 6, 5, 6, 2, 3, 2, 3, 4, 8, 4, 8, 5, 0, 0, 0, 0, 0, 0,
        6, 5, 2, 1, 3, 4, 8, 4, 8, 5, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        6, 5, 1, 5, 1, 4, 3, 4, 8, 4, 8, 5, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        6, 5, 6, 2, 7, 3, 4, 3, 4, 1, 8, 5, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        1, 2, 1, 5, 8, 5, 6, 5, 6, 2, 7, 3, 4, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        6, 5, 2, 1, 2, 3, 7, 3, 4, 3, 4, 1, 8, 5, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        6, 5, 1, 5, 8, 5, 2, 3, 7, 3, 4, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        6, 5, 6, 2, 3, 2, 4, 1, 8, 5, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        1, 2, 1, 5, 8, 5, 6, 5, 6, 2, 3, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        6, 5, 2, 1, 4, 1, 8, 5, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        6, 5, 1, 5, 8, 5, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        5, 1, 6, 2, 7, 3, 8, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        1, 2, 6, 2, 7, 3, 8, 4, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        5, 1, 2, 1, 2, 3, 7, 3, 8, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        2, 3, 7, 3, 8, 4, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        5, 1, 6, 2, 3, 2, 3, 4, 8, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        1, 2, 6, 2, 3, 2, 3, 4, 8, 4, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        5, 1, 2, 1, 3, 4, 8, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        3, 4, 8, 4, 1, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        5, 1, 6, 2, 7, 3, 4, 3, 4, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        1, 2, 6, 2, 7, 3, 4, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        5, 1, 2, 1, 2, 3, 7, 3, 4, 3, 4, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        2, 3, 7, 3, 4, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        5, 1, 6, 2, 3, 2, 4, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        1, 2, 6, 2, 3, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        5, 1, 2, 1, 4, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};

                         index  =   0;
  if (cvertex[0] >= cut) index +=   1;
  if (cvertex[1] >= cut) index +=   2;
  if (cvertex[2] >= cut) index +=   4;
  if (cvertex[3] >= cut) index +=   8;
  if (cvertex[4] >= cut) index +=  16;
  if (cvertex[5] >= cut) index +=  32;
  if (cvertex[6] >= cut) index +=  64;
  if (cvertex[7] >= cut) index += 128;

  /* make the 2D nodes at edge intersections */

  ninter = npts[index];
  if (ninter == 0) return;

  for (i = 0; i < ninter; i++) {
    n1   = ecut[index][i][0]-1;
    n2   = ecut[index][i][1]-1;
    dist = (cut-cvertex[n1])/(cvertex[n2]-cvertex[n1]);
    FX_2DNode(vertices[n1], vertices[n2], dist, &nc[i], 
              nedgeX, edgeX, nodeTable);
  }

  /* make the new triangles */

  npoly = nelem[index];
  for (ni = i = 0; i < npoly; i++) {
    nnod = npel[index][i]-1;
    ist  = ncell[nnod][1]-1;
    for (j = 0; j < ncell[nnod][0]; j++, ist++) {
      tris[*ntris].nodes[0] = nc[ni+npcel[ist][0]-1];
      tris[*ntris].nodes[1] = nc[ni+npcel[ist][1]-1];
      tris[*ntris].nodes[2] = nc[ni+npcel[ist][2]-1];
      tris[*ntris].nodes[3] = kc;
      *ntris += 1;
    }
    ni += nnod+1;
  }
}
