/*
 * FX_EigenSubs.c
 *
 * FX modules that compute eigen systems.
 *
 * Copyright 1999 - 2000, Massachusetts Institute of Technology.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#define FUZZ     1.0e-14
#define MINTINY -1.0e-4
#define TINY     1.0e-10

#define COMPLEX_CONJUGATE -3
#define THREE_REAL         3

#define Zero_test2(result,term1,term2) \
  if (fabs(result) <= FUZZ*(fabs(term1)+fabs(term2))) \
    result = 0.0

#define Zero_test3(result,term1,term2,term3) \
  if (fabs(result) <= FUZZ*(fabs(term1)+fabs(term2)+fabs(term3))) \
    result = 0.0

#define Zero_test4(result,term1,term2,term3,term4) \
  if (fabs(result) <= FUZZ*(fabs(term1)+fabs(term2)+fabs(term3)+fabs(term4))) \
    result = 0.0



static double
FX_CubeRoot(double x)
{
  if (x > 0.0)      return (pow (x, 1.0/3.0));
  else if (x < 0.0) return (-pow (-x, 1.0/3.0));
  else              return 0.0;
}


static int
FX_SolveLinear(double a, double b, double *x)
{
  /* 
   * Solve ax + b = 0
   *
   * Return values: 0 - no solution 
   *                1 - one real solution, returned in *x
   *               -1 - infinite number of solutions
   */

  if (a == 0) {
    if (b == 0) {
      *x = 0;
      return -1;
    } else {
      return 0;
    }
  }
  *x = -b/a;
  return 1;
}


static int
FX_SolveQuadratic(double a, double b, double c, double *x1, double *x2)
{
  /* 
   *         2
   * Solve ax  + bx + c = 0
   *
   * Return values: 0 - no solution 
   *               -1 - infinite number of solutions
   *                1 - one real root, possibly of multiplicity 2.
   *                    returned in *x1
   *		2 - two distinct real roots of mutiplicity 1
   *                    returned in *x1 and *x2;
   *	         -2 - two complex roots (conjugates),
   *		    real part returned in *x1,
   *		    imaginary part returned in *x2
   */

  double d;

  if (a == 0) return FX_SolveLinear(b, c, x1);

  d = b*b - 4*a*c;
  Zero_test2(d, b*b, 4*a*c);

  if (d == 0) {
    *x1 = *x2 = -b/(2*a);
    return 1;
  }

  if (d > 0) {
    d = sqrt(d);
    *x1 = (-b + d)/(2*a);
    Zero_test2(*x1, -b/2*a, d/2*a);
    *x2 = (-b - d)/(2*a);
    Zero_test2(*x2, -b/2*a, -d/2*a);
    return 2;
  }

  if (d < 0) {
    d = sqrt(-d);
    *x1 = -b/(2*a);
    *x2 =  d/(2*a);
    return -2;
  }

  return 0;
}


static int
FX_SolveCubic(double a, double b, double c, double d, double *x1, double *x2,
              double *x3)
{
  /* 
   *         3     2
   * Solve ax  + bx  + cx + d = 0
   *
   * Return values: 0 - no solution 
   *               -1 - infinite number of solutions
   *                1 - only one distinct root, real, possibly of multiplicity
   *                     2 or 3.  returned in *x1
   *	            2 - only two distinct roots, both real, one possibly of 
   *                    multiplicity 2, returned in *x1 and *x2
   *	           -2 - only two distinct roots (complex conjugates),
   *                    real part returned in *x1,
   *                    imaginary part returned in *x2
   *                3 - three distinct real roots of multiplicity 1
   *                    returned in *x1, *x2, and *x3.
   *               -3 - one real root and one complex conjugate pair.
   *                    real root in *x1, real part of complex conjugate root
   *                    in *x2, imaginary part in *x3.
   *
   *              XXX - this whole scheme is wrong.  
   *                    Each root should be returned along with its 
   *                    multiplicity.
   */

  int    ret;
  double a3, d2, p, q, r, a2, b2, z1, z2, z3, des, t1, t2, b3;
  double temp_r, theta, cos_term, sin_term;
  
  *x1 = *x2 = *x3 = 0.0;

  if (a == 0.0) return FX_SolveQuadratic (b, c, d, x1, x2);

  p = b/a;			/* reduce to y^3 + py^2 + qy + r = 0 */
  q = c/a;			/* by dividing through by a */
  r = d/a;

  a2 = (3*q - p*p)/3;	        /* reduce to z^3 + a2*z + b2 = 0 */
  Zero_test2(a2, 3*q, p*p);

  b2 = (2*p*p*p - 9*p*q + 27*r)/27;	/* using y = z - p/3 */
  Zero_test3(b2, 2*p*p*p, 9*p*q, 27*r);

  des = (b2*b2/4 + a2*a2*a2/27);
  Zero_test2(des, b2*b2/4, a2*a2*a2/27);

  if (des == 0.0) {		/* three real roots, at least two equal */ 
    a3  = FX_CubeRoot (-b2/2);
    ret = 2;
    if (a3 == 0.0) {
      /* one distinct real root of multiplicity three */
      z1 = 0.0;
      *x1 = z1 - p/3;
      Zero_test2(*x1, z1, p/3);
      *x2 = 0.0;
      *x3 = 0.0;
      ret = 1;
    } else {
      /* one real root multiplicity one, another of multiplicity two */
      z1 = 2*a3;
      z2 = -a3;
      *x1 = z1 - p/3;
      Zero_test2(*x1, z1, p/3);
      *x2 = z2 - p/3;
      Zero_test2(*x2, z2, p/3);
      *x3 = 0.0;
    }
  } else if (des > 0.0) {	/* one real root, one complex conjugate pair */
    d2 = sqrt(des);
    t1 = -b2/2 + d2;
    t2 = -b2/2 - d2;
    Zero_test2(t1, b2/2, d2);
    Zero_test2(t2, b2/2, d2);
    a3 = FX_CubeRoot(t1);
    b3 = FX_CubeRoot(t2);
    z1 = a3 + b3;
    Zero_test2(z1, a3, b3);
    z2 = - z1/2;
    t1 = a3-b3;
    Zero_test2(t1, a3, b3);
    z3 = sqrt(3.0) * t1/2;
    *x1 = z1 - p/3;
    Zero_test2(*x1, z1, p/3);
    *x2 = z2 - p/3;
    Zero_test2(*x2, z2, p/3);
    *x3 = z3;
    ret = -3;
  } else {			/* three unequal real roots */
    t1 = b2*b2/4 - des;
    Zero_test2(t1, b2*b2/4, des);
    temp_r = FX_CubeRoot(sqrt(b2*b2/4 - des));
    theta  = atan2(sqrt(-des), (-b2/2));
    cos_term = temp_r * cos(theta/3);
    sin_term = temp_r * sin(theta/3) * sqrt(3.0);
    z1 = 2*cos_term;
    z2 = -cos_term - sin_term;
    Zero_test2(z2, cos_term, sin_term);
    z3 = -cos_term + sin_term;
    Zero_test2(z3, cos_term, sin_term);

    *x1 = z1 - p/3;
    Zero_test2(*x1, z1, p/3);
    *x2 = z2 - p/3;
    Zero_test2(*x2, z2, p/3);
    *x3 = z3 - p/3;
    Zero_test2(*x3, z3, p/3);
    ret = 3;
  }
  return ret;
}


static void
FX_NormalizeVector(double *v, double h)
{
  double mag = sqrt (v[0]*v[0] + v[1]*v[1] + v[2]*v[2]);
  double h_over_mag;

  if (mag == 0.0) mag = 1e-10;
  h_over_mag = h/mag;
  v[0] = v[0]*h_over_mag;
  v[1] = v[1]*h_over_mag;
  v[2] = v[2]*h_over_mag;
}


static void
FX_ZeroSmall(double *v)
{
  int i;

  for (i = 0; i < 3; i++) Zero_test3 (v[i], v[0], v[1], v[2]);
}


static int
FX_RealEigenvector (double a[3][3], double val, double vec[3])
{
    double det[3], m[3][3];
    int    i,j;
    
    for (i=0; i<3; i++)
      for (j=0; j<3; j++) m[i][j] = a[i][j];
    
    for (i=0; i<3; i++) {
      m[i][i] = m[i][i] - val;
      Zero_test2(m[i][i], m[i][i], val);
    }
    
    det[0] = m[1][1]*m[2][2] - m[1][2]*m[2][1];
    Zero_test4(det[0], m[1][1],m[2][2],m[1][2],m[2][1]);
    
    det[1] = m[0][0]*m[2][2] - m[0][2]*m[2][0];
    Zero_test4(det[1], m[0][0],m[2][2],m[0][2],m[2][0]);
    
    det[2] = m[0][0]*m[1][1] - m[0][1]*m[1][0];
    Zero_test4(det[2], m[0][0],m[1][1],m[0][1],m[1][0]);
  
    if ( (det[0] == 0.0) && (det[1] == 0.0) && (det[2] == 0.0) ) {
        vec[0] = 1.0; vec[1] = vec[2] = 0.0;
        return 0;
    } else if ( (fabs(det[0])>fabs(det[1])) && (fabs(det[0])>fabs(det[2])) ) {
        vec[0] = 1.0;
        vec[1] = (m[1][2]*m[2][0] - m[1][0]*m[2][2]) / det[0];
        vec[2] = (m[1][0]*m[2][1] - m[1][1]*m[2][0]) / det[0];
    } else if ( fabs(det[1]) > fabs(det[2]) ) {
        vec[0] = (m[0][2]*m[2][1] - m[0][1]*m[2][2]) / det[1];
        vec[1] = 1.0;
        vec[2] = (m[0][1]*m[2][0] - m[0][0]*m[2][1]) / det[1];
    } else {
        vec[0] = (m[0][1]*m[1][2] - m[0][2]*m[1][1]) / det[2];
        vec[1] = (m[0][2]*m[1][0] - m[0][0]*m[1][2]) / det[2];
        vec[2] = 1.0;
    }
    FX_ZeroSmall(vec);
    FX_NormalizeVector(vec, 1.0);
    return 1;
}


/*
 * given velocity gradient tensor (a 3x3) -- return axis of rotation
 *   type = 0 -- use vorticity
 *   type = 1 -- use eigenvector of real eigenvalue
 *
 *   returns = -3 OK
 *              0 Error
 *              other is the type of eigen system
 */

int
FX_RotAxis(int type, float *a, float *ax)
{
    float  mag;
    int    eigenvalue_type, ret;
    double A[3][3];
    double determinant, minors, trace, eigenvalue[3], eigenvec[3];

    ax[0] = 0.0;
    ax[1] = 0.0;
    ax[2] = 0.0;

    if (type != 0) {

      A[0][0] = a[0];
      A[0][1] = a[1];
      A[0][2] = a[2];
      A[1][0] = a[3];
      A[1][1] = a[4];
      A[1][2] = a[5];
      A[2][0] = a[6];
      A[2][1] = a[7];
      A[2][2] = a[8];
      determinant = -(A[0][0]*A[1][1]*A[2][2] - A[0][2]*A[1][1]*A[2][0]
                    + A[0][1]*A[1][2]*A[2][0] - A[0][1]*A[1][0]*A[2][2]
                    + A[0][2]*A[1][0]*A[2][1] - A[0][0]*A[1][2]*A[2][1]);

      minors = -(A[0][1]*A[1][0] - A[0][0]*A[1][1] + A[0][2]*A[2][0]
               + A[1][2]*A[2][1] - A[0][0]*A[2][2] - A[1][1]*A[2][2]);

      trace = -(A[0][0] + A[1][1] + A[2][2]);

      eigenvalue_type = FX_SolveCubic(1.0, trace, minors, determinant,
                                      &eigenvalue[0], &eigenvalue[1],
                                      &eigenvalue[2]);

      if (eigenvalue_type != COMPLEX_CONJUGATE) return eigenvalue_type;

      ret = FX_RealEigenvector(A, eigenvalue[0], eigenvec);
      if (ret == 0) return ret;

      ax[0] = 0.5*(a[5]-a[7]);
      ax[1] = 0.5*(a[6]-a[2]);
      ax[2] = 0.5*(a[1]-a[3]);
      ax[3] = sqrt(ax[0]*ax[0] + ax[1]*ax[1] + ax[2]*ax[2]);
      ax[0] = eigenvec[0];
      ax[1] = eigenvec[1];
      ax[2] = eigenvec[2];
      return eigenvalue_type;

    } else {

      ax[0] = 0.5*(a[5]-a[7]);
      ax[1] = 0.5*(a[6]-a[2]);
      ax[2] = 0.5*(a[1]-a[3]);
      mag = sqrt(ax[0]*ax[0] + ax[1]*ax[1] + ax[2]*ax[2]);
      ax[3] = mag;
      if (mag != 0.0) {
        ax[0] /= mag;
        ax[1] /= mag;
        ax[2] /= mag;
        return COMPLEX_CONJUGATE;
      } else {
        return 0;
      }
    }
}



/* 
 *  Solve the simultaneous equations: 
 *
 *      a1 x + b1 y = c1
 *      a2 x + b2 y = c2
 *	
 *  Function returns 1 if the lines intersect at exactly 1 point <x,y>.
 *  Otherwise, returns 0.
 *
 */

static int
FX_Solve2x2(double *x, double *y, double a1, double b1, double c1, 
            double a2, double b2, double c2 )
{
     double denom;
     
     denom = a1*b2 - a2*b1;
     
     /* return 0 if lines are parallel */
     if (fabs(denom) < TINY) return 0;

     *x = (c1*b2 - c2*b1) / denom;
     *y = (c2*a1 - c1*a2) / denom;
     
     return 1;
}


/* 
 *
 *   returns 1 if V1 and V2 are zero at ONE point within (including the 
 *   border of) the triangle.  Otherwise returns 0. 
 *
 */

static int
FX_ZeroOnTriangle(double V11, double V12, double V13, double V21, 
                  double V22, double V23, double *p0, double *q0)
{
     double a1, b1, c1, a2, b2, c2;

     if ((V11*V12) > 0.0 && (V12*V13) > 0.0) return 0;
     if ((V21*V22) > 0.0 && (V22*V23) > 0.0) return 0;

     /*======================================================================= 

	finding the interpolation function for V1:

	V11 = a1 + b1 p1 + c1 q1
	V12 = a2 + b2 p2 + c2 q2
	V13 = a3 + b3 p3 + c3 q3
	
	or

	[V1] = [ 1 0 0   [a1
	         1 1 0    b1
                 1 0 1 ]  c1]

	=> V1 = a1 + b1 p + c1 q

	similarly for V2:

	[V2] = [ 1 0 0   [a2
	         1 1 0    b2
                 1 0 1 ]  c2]
     
	=> V2 = a2 + b2 p + c2 q
     =================================================================*/

     a1 = V11;
     b1 = V12 - a1;
     c1 = V13 - a1;
     
     a2 = V21;
     b2 = V22 - a2;
     c2 = V23 - a2;

     if (! FX_Solve2x2 (p0, q0, b1, c1, -a1, b2, c2, -a2)) return 0;
     
     if (*p0 < MINTINY || *q0 < MINTINY || (1.0-*p0-*q0) < MINTINY) return 0;

     return 1;
}


/*
 *  given the triangle nodes -- return the intersection point with raxis
 */

int
FX_AxisXTri(float coor[][3], float v[][3], int kn1, int kn2, int kn3,
            float *raxis, float om0, float om1, float om2, 
            float *vort, float *point)
{
    int    i;
    double vel[3][3], xyz[3][3], projVx[3], projVy[3], Vx[3], Vy[3];
    double p, q;

    vel[0][0] = v[kn1-1][0];
    vel[0][1] = v[kn1-1][1];
    vel[0][2] = v[kn1-1][2];
    vel[1][0] = v[kn2-1][0];
    vel[1][1] = v[kn2-1][1];
    vel[1][2] = v[kn2-1][2];
    vel[2][0] = v[kn3-1][0];
    vel[2][1] = v[kn3-1][1];
    vel[2][2] = v[kn3-1][2];

    xyz[0][0] = coor[kn1-1][0];
    xyz[0][1] = coor[kn1-1][1];
    xyz[0][2] = coor[kn1-1][2];
    xyz[1][0] = coor[kn2-1][0];
    xyz[1][1] = coor[kn2-1][1];
    xyz[1][2] = coor[kn2-1][2];
    xyz[2][0] = coor[kn3-1][0];
    xyz[2][1] = coor[kn3-1][1];
    xyz[2][2] = coor[kn3-1][2];

    /* get the x' and y' projections */

    projVy[0] = 1.0;
    projVy[1] = 0.0;
    projVy[2] = 0.0;
    if ((raxis[0] > 0.65) || (raxis[0] < -0.65)) {
      projVy[0] = 0.0;
      projVy[1] = 1.0;
    }
    projVx[0] = projVy[1]*raxis[2] - raxis[1]*projVy[2];
    projVx[1] = projVy[2]*raxis[0] - raxis[2]*projVy[0];
    projVx[2] = projVy[0]*raxis[1] - raxis[0]*projVy[1];
    projVy[0] = projVx[1]*raxis[2] - raxis[1]*projVx[2];
    projVy[1] = projVx[2]*raxis[0] - raxis[2]*projVx[0];
    projVy[2] = projVx[0]*raxis[1] - raxis[0]*projVx[1];

    /* calculate projected velocities */

    Vx[0] = projVx[0]*vel[0][0] + projVx[1]*vel[0][1] + projVx[2]*vel[0][2];
    Vy[0] = projVy[0]*vel[0][0] + projVy[1]*vel[0][1] + projVy[2]*vel[0][2];
    Vx[1] = projVx[0]*vel[1][0] + projVx[1]*vel[1][1] + projVx[2]*vel[1][2];
    Vy[1] = projVy[0]*vel[1][0] + projVy[1]*vel[1][1] + projVy[2]*vel[1][2];
    Vx[2] = projVx[0]*vel[2][0] + projVx[1]*vel[2][1] + projVx[2]*vel[2][2];
    Vy[2] = projVy[0]*vel[2][0] + projVy[1]*vel[2][1] + projVy[2]*vel[2][2];

    if (FX_ZeroOnTriangle(Vx[0], Vx[1], Vx[2], Vy[0], Vy[1], Vy[2],
                          &p, &q) == 0) return 0;
     
    for (i = 0; i < 3; i++) 
      point[i] = xyz[0][i] + p*(xyz[1][i]-xyz[0][i]) + q*(xyz[2][i]-xyz[0][i]);

    *vort = om0 + p*(om1-om0) + q*(om2-om0);

    return 1;
}


int
FX_AxisXTriP3d(float *xcoor, float *ycoor, float *zcoor, float *vx, float *vy,
               float *vz, int kn1, int kn2, int kn3, float *raxis,
               float om0, float om1, float om2, float *vort, float *point)
{
    int    i;
    double vel[3][3], xyz[3][3], projVx[3], projVy[3], Vx[3], Vy[3];
    double p, q;

    vel[0][0] = vx[kn1-1];
    vel[0][1] = vy[kn1-1];
    vel[0][2] = vz[kn1-1];
    vel[1][0] = vx[kn2-1];
    vel[1][1] = vy[kn2-1];
    vel[1][2] = vz[kn2-1];
    vel[2][0] = vx[kn3-1];
    vel[2][1] = vy[kn3-1];
    vel[2][2] = vz[kn3-1];

    xyz[0][0] = xcoor[kn1-1];
    xyz[0][1] = ycoor[kn1-1];
    xyz[0][2] = zcoor[kn1-1];
    xyz[1][0] = xcoor[kn2-1];
    xyz[1][1] = ycoor[kn2-1];
    xyz[1][2] = zcoor[kn2-1];
    xyz[2][0] = xcoor[kn3-1];
    xyz[2][1] = ycoor[kn3-1];
    xyz[2][2] = zcoor[kn3-1];

    /* get the x' and y' projections */

    projVy[0] = 1.0;
    projVy[1] = 0.0;
    projVy[2] = 0.0;
    if ((raxis[0] > 0.65) || (raxis[0] < -0.65)) {
      projVy[0] = 0.0;
      projVy[1] = 1.0;
    }
    projVx[0] = projVy[1]*raxis[2] - raxis[1]*projVy[2];
    projVx[1] = projVy[2]*raxis[0] - raxis[2]*projVy[0];
    projVx[2] = projVy[0]*raxis[1] - raxis[0]*projVy[1];
    projVy[0] = projVx[1]*raxis[2] - raxis[1]*projVx[2];
    projVy[1] = projVx[2]*raxis[0] - raxis[2]*projVx[0];
    projVy[2] = projVx[0]*raxis[1] - raxis[0]*projVx[1];

    /* calculate projected velocities */

    Vx[0] = projVx[0]*vel[0][0] + projVx[1]*vel[0][1] + projVx[2]*vel[0][2];
    Vy[0] = projVy[0]*vel[0][0] + projVy[1]*vel[0][1] + projVy[2]*vel[0][2];
    Vx[1] = projVx[0]*vel[1][0] + projVx[1]*vel[1][1] + projVx[2]*vel[1][2];
    Vy[1] = projVy[0]*vel[1][0] + projVy[1]*vel[1][1] + projVy[2]*vel[1][2];
    Vx[2] = projVx[0]*vel[2][0] + projVx[1]*vel[2][1] + projVx[2]*vel[2][2];
    Vy[2] = projVy[0]*vel[2][0] + projVy[1]*vel[2][1] + projVy[2]*vel[2][2];

    if (FX_ZeroOnTriangle(Vx[0], Vx[1], Vx[2], Vy[0], Vy[1], Vy[2],
                          &p, &q) == 0) return 0;
     
    for (i = 0; i < 3; i++) 
      point[i] = xyz[0][i] + p*(xyz[1][i]-xyz[0][i]) + q*(xyz[2][i]-xyz[0][i]);

    *vort = om0 + p*(om1-om0) + q*(om2-om0);

    return 1;
}
