/*
 * FX_GradSubs.c
 *
 * FX modules that calculate gradients.
 *
 * Copyright 1999 - 2000, Massachusetts Institute of Technology.
 *
 */

#include <stdio.h>
#include <stdlib.h>


#include "FX_types.h"

#ifdef WIN32

#define PREFIX __stdcall

#else

#define PREFIX

#ifdef UNDERSCORE

#define FXCELLGET fxcellget_

#else

#define FXCELLGET fxcellget

#endif
#endif

  extern void PREFIX FXCELLGET (int *, int *, int *);



/*
 *  Computes and returns the gradient of src (dimension dim)
 */

static void
FX_TetGradPtr(int *node, FXcoor *coor, int dim, float *src, float *grad, 
              float *vol)
{
  int    i, i1, i2, i3, i4, j1, j2, j3, j4;
  double x1, y1, z1, x2, y2, z2, x3, y3, z3, x4, y4, z4, V, c[3][4];
  double u1, u2, u3, u4, dudx, dudy, dudz;

  i1 = node[0]-1;
  x1 = coor[i1].x;
  y1 = coor[i1].y;
  z1 = coor[i1].z;
  i2 = node[1]-1;
  x2 = coor[i2].x;
  y2 = coor[i2].y;
  z2 = coor[i2].z;
  i3 = node[2]-1;
  x3 = coor[i3].x;
  y3 = coor[i3].y;
  z3 = coor[i3].z;
  i4 = node[3]-1;
  x4 = coor[i4].x;
  y4 = coor[i4].y;
  z4 = coor[i4].z;

  V = (x2-x1)*( (y3-y1)*(z4-z1)-(z3-z1)*(y4-y1) ) -
      (x3-x1)*( (y2-y1)*(z4-z1)-(z2-z1)*(y4-y1) ) +
      (x4-x1)*( (y2-y1)*(z3-z1)-(z2-z1)*(y3-y1) );
  vol[i1] += V;
  vol[i2] += V;
  vol[i3] += V;
  vol[i4] += V;

  c[0][0] = (z2-z3)*(y3-y4)-(z3-z4)*(y2-y3);
  c[0][1] = (z4-z1)*(y3-y4)-(z3-z4)*(y4-y1);
  c[0][2] = (z4-z1)*(y1-y2)-(z1-z2)*(y4-y1);
  c[0][3] = (z2-z3)*(y1-y2)-(z1-z2)*(y2-y3);

  c[1][0] = (x2-x3)*(z3-z4)-(x3-x4)*(z2-z3);
  c[1][1] = (x4-x1)*(z3-z4)-(x3-x4)*(z4-z1);
  c[1][2] = (x4-x1)*(z1-z2)-(x1-x2)*(z4-z1);
  c[1][3] = (x2-x3)*(z1-z2)-(x1-x2)*(z2-z3);

  c[2][0] = (y2-y3)*(x3-x4)-(y3-y4)*(x2-x3);
  c[2][1] = (y4-y1)*(x3-x4)-(y3-y4)*(x4-x1);
  c[2][2] = (y4-y1)*(x1-x2)-(y1-y2)*(x4-x1);
  c[2][3] = (y2-y3)*(x1-x2)-(y1-y2)*(x2-x3);

  for (i = 0; i < dim; i++) {
    j1   = i1*dim+i;
    j2   = i2*dim+i;
    j3   = i3*dim+i;
    j4   = i4*dim+i;
    u1   = src[j1];
    u2   = src[j2];
    u3   = src[j3];
    u4   = src[j4];
    dudx = c[0][0]*u1 + c[0][1]*u2 + c[0][2]*u3 + c[0][3]*u4;
    dudy = c[1][0]*u1 + c[1][1]*u2 + c[1][2]*u3 + c[1][3]*u4;
    dudz = c[2][0]*u1 + c[2][1]*u2 + c[2][2]*u3 + c[2][3]*u4;

    grad[3*j1  ] += dudx;
    grad[3*j1+1] += dudy;
    grad[3*j1+2] += dudz;

    grad[3*j2  ] += dudx;
    grad[3*j2+1] += dudy;
    grad[3*j2+2] += dudz;

    grad[3*j3  ] += dudx;
    grad[3*j3+1] += dudy;
    grad[3*j3+2] += dudz;

    grad[3*j4  ] += dudx;
    grad[3*j4+1] += dudy;
    grad[3*j4+2] += dudz;
  }

}


static void
FX_TetGradP3d(int *node, float *xcoor, float *ycoor, float *zcoor, int dim,
              float *src, float *src1, float *src2, float *grad, float *vol)
{
  int    i, i1, i2, i3, i4, j1, j2, j3, j4;
  double x1, y1, z1, x2, y2, z2, x3, y3, z3, x4, y4, z4, V, c[3][4];
  double u1, u2, u3, u4, dudx, dudy, dudz;

  i1 = node[0]-1;
  x1 = xcoor[i1];
  y1 = ycoor[i1];
  z1 = zcoor[i1];
  i2 = node[1]-1;
  x2 = xcoor[i2];
  y2 = ycoor[i2];
  z2 = zcoor[i2];
  i3 = node[2]-1;
  x3 = xcoor[i3];
  y3 = ycoor[i3];
  z3 = zcoor[i3];
  i4 = node[3]-1;
  x4 = xcoor[i4];
  y4 = ycoor[i4];
  z4 = zcoor[i4];

  V = (x2-x1)*( (y3-y1)*(z4-z1)-(z3-z1)*(y4-y1) ) -
      (x3-x1)*( (y2-y1)*(z4-z1)-(z2-z1)*(y4-y1) ) +
      (x4-x1)*( (y2-y1)*(z3-z1)-(z2-z1)*(y3-y1) );
  vol[i1] += V;
  vol[i2] += V;
  vol[i3] += V;
  vol[i4] += V;

  c[0][0] = (z2-z3)*(y3-y4)-(z3-z4)*(y2-y3);
  c[0][1] = (z4-z1)*(y3-y4)-(z3-z4)*(y4-y1);
  c[0][2] = (z4-z1)*(y1-y2)-(z1-z2)*(y4-y1);
  c[0][3] = (z2-z3)*(y1-y2)-(z1-z2)*(y2-y3);

  c[1][0] = (x2-x3)*(z3-z4)-(x3-x4)*(z2-z3);
  c[1][1] = (x4-x1)*(z3-z4)-(x3-x4)*(z4-z1);
  c[1][2] = (x4-x1)*(z1-z2)-(x1-x2)*(z4-z1);
  c[1][3] = (x2-x3)*(z1-z2)-(x1-x2)*(z2-z3);

  c[2][0] = (y2-y3)*(x3-x4)-(y3-y4)*(x2-x3);
  c[2][1] = (y4-y1)*(x3-x4)-(y3-y4)*(x4-x1);
  c[2][2] = (y4-y1)*(x1-x2)-(y1-y2)*(x4-x1);
  c[2][3] = (y2-y3)*(x1-x2)-(y1-y2)*(x2-x3);

  if (dim == 1) {
    u1   = src[i1];
    u2   = src[i2];
    u3   = src[i3];
    u4   = src[i4];
    dudx = c[0][0]*u1 + c[0][1]*u2 + c[0][2]*u3 + c[0][3]*u4;
    dudy = c[1][0]*u1 + c[1][1]*u2 + c[1][2]*u3 + c[1][3]*u4;
    dudz = c[2][0]*u1 + c[2][1]*u2 + c[2][2]*u3 + c[2][3]*u4;

    grad[3*i1  ] += dudx;
    grad[3*i1+1] += dudy;
    grad[3*i1+2] += dudz;

    grad[3*i2  ] += dudx;
    grad[3*i2+1] += dudy;
    grad[3*i2+2] += dudz;

    grad[3*i3  ] += dudx;
    grad[3*i3+1] += dudy;
    grad[3*i3+2] += dudz;

    grad[3*i4  ] += dudx;
    grad[3*i4+1] += dudy;
    grad[3*i4+2] += dudz;
    return;
  }

  for (i = 0; i < dim; i++) {
    if (i == 0) {
      u1 = src[i1];
      u2 = src[i2];
      u3 = src[i3];
      u4 = src[i4];
    } else if (i == 1) {
      u1 = src1[i1];
      u2 = src1[i2];
      u3 = src1[i3];
      u4 = src1[i4];
    } else {
      u1 = src2[i1];
      u2 = src2[i2];
      u3 = src2[i3];
      u4 = src2[i4];
    }
    dudx = c[0][0]*u1 + c[0][1]*u2 + c[0][2]*u3 + c[0][3]*u4;
    dudy = c[1][0]*u1 + c[1][1]*u2 + c[1][2]*u3 + c[1][3]*u4;
    dudz = c[2][0]*u1 + c[2][1]*u2 + c[2][2]*u3 + c[2][3]*u4;

    j1 = i1*dim+i;
    j2 = i2*dim+i;
    j3 = i3*dim+i;
    j4 = i4*dim+i;

    grad[3*j1  ] += dudx;
    grad[3*j1+1] += dudy;
    grad[3*j1+2] += dudz;

    grad[3*j2  ] += dudx;
    grad[3*j2+1] += dudy;
    grad[3*j2+2] += dudz;

    grad[3*j3  ] += dudx;
    grad[3*j3+1] += dudy;
    grad[3*j3+2] += dudz;

    grad[3*j4  ] += dudx;
    grad[3*j4+1] += dudy;
    grad[3*j4+2] += dudz;
  }

}


static void
FX_TetGrad(FXglobal *globals, int *node, int dim, float *src, float *grad,
           float *vol)
{

  if ((globals->cmask & 64) == 0) {
    FX_TetGradPtr(node, globals->coor, dim, src, grad, vol);
  } else {
    if (dim != 3) {
      FX_TetGradP3d(node, globals->xcoor, globals->ycoor, globals->zcoor, dim,
                    src, NULL, NULL, grad, vol);
    } else {
      FX_TetGradP3d(node, globals->xcoor, globals->ycoor, globals->zcoor, dim,
                    globals->velx, globals->vely, globals->velz, grad, vol);
    }
  }

}


/*
 *  Computes and returns the gradient of src (dimension dim) with
 *  halo nodes
 */

static void
FX_HTetGradPtr(int *node, int n, FXcoor *coor, FXcoor *hcoor, int dim, 
               float *src, float *hsrc, float *grad, float *vol)
{
  int    i, i1, i2, i3, i4, j1, j2, j3, j4;
  double x1, y1, z1, x2, y2, z2, x3, y3, z3, x4, y4, z4, V, c[3][4];
  double u1, u2, u3, u4, dudx, dudy, dudz;

  i1 = node[0]-1;
  if (i1 < n) {
    x1 = coor[i1].x;
    y1 = coor[i1].y;
    z1 = coor[i1].z;
  } else {
    x1 = hcoor[i1-n].x;
    y1 = hcoor[i1-n].y;
    z1 = hcoor[i1-n].z;
  }
  i2 = node[1]-1;
  if (i2 < n) {
    x2 = coor[i2].x;
    y2 = coor[i2].y;
    z2 = coor[i2].z;
  } else {
    x2 = hcoor[i2-n].x;
    y2 = hcoor[i2-n].y;
    z2 = hcoor[i2-n].z;
  }
  i3 = node[2]-1;
  if (i3 < n) {
    x3 = coor[i3].x;
    y3 = coor[i3].y;
    z3 = coor[i3].z;
  } else {
    x3 = hcoor[i3-n].x;
    y3 = hcoor[i3-n].y;
    z3 = hcoor[i3-n].z;
  }
  i4 = node[3]-1;
  if (i4 < n) {
    x4 = coor[i4].x;
    y4 = coor[i4].y;
    z4 = coor[i4].z;
  } else {
    x4 = hcoor[i4-n].x;
    y4 = hcoor[i4-n].y;
    z4 = hcoor[i4-n].z;
  }

  V = (x2-x1)*( (y3-y1)*(z4-z1)-(z3-z1)*(y4-y1) ) -
      (x3-x1)*( (y2-y1)*(z4-z1)-(z2-z1)*(y4-y1) ) +
      (x4-x1)*( (y2-y1)*(z3-z1)-(z2-z1)*(y3-y1) );
  if (i1 < n) vol[i1] += V;
  if (i2 < n) vol[i2] += V;
  if (i3 < n) vol[i3] += V;
  if (i4 < n) vol[i4] += V;

  c[0][0] = (z2-z3)*(y3-y4)-(z3-z4)*(y2-y3);
  c[0][1] = (z4-z1)*(y3-y4)-(z3-z4)*(y4-y1);
  c[0][2] = (z4-z1)*(y1-y2)-(z1-z2)*(y4-y1);
  c[0][3] = (z2-z3)*(y1-y2)-(z1-z2)*(y2-y3);

  c[1][0] = (x2-x3)*(z3-z4)-(x3-x4)*(z2-z3);
  c[1][1] = (x4-x1)*(z3-z4)-(x3-x4)*(z4-z1);
  c[1][2] = (x4-x1)*(z1-z2)-(x1-x2)*(z4-z1);
  c[1][3] = (x2-x3)*(z1-z2)-(x1-x2)*(z2-z3);

  c[2][0] = (y2-y3)*(x3-x4)-(y3-y4)*(x2-x3);
  c[2][1] = (y4-y1)*(x3-x4)-(y3-y4)*(x4-x1);
  c[2][2] = (y4-y1)*(x1-x2)-(y1-y2)*(x4-x1);
  c[2][3] = (y2-y3)*(x1-x2)-(y1-y2)*(x2-x3);

  for (i = 0; i < dim; i++) {
    if (i1 < n) {
      j1 = i1*dim+i;
      u1 = src[j1];
    } else {
      j1 = (i1-n)*dim+i;
      u1 = hsrc[j1];
    }
    if (i2 < n) {
      j2 = i2*dim+i;
      u2 = src[j2];
    } else {
      j2 = (i2-n)*dim+i;
      u2 = hsrc[j2];
    }
    if (i3 < n) {
      j3 = i3*dim+i;
      u3 = src[j3];
    } else {
      j3 = (i3-n)*dim+i;
      u3 = hsrc[j3];
    }
    if (i4 < n) {
      j4 = i4*dim+i;
      u4 = src[j4];
    } else {
      j4 = (i4-n)*dim+i;
      u4 = hsrc[j4];
    }
    dudx = c[0][0]*u1 + c[0][1]*u2 + c[0][2]*u3 + c[0][3]*u4;
    dudy = c[1][0]*u1 + c[1][1]*u2 + c[1][2]*u3 + c[1][3]*u4;
    dudz = c[2][0]*u1 + c[2][1]*u2 + c[2][2]*u3 + c[2][3]*u4;

    if (i1 < n) {
      grad[3*j1  ] += dudx;
      grad[3*j1+1] += dudy;
      grad[3*j1+2] += dudz;
    }
    if (i2 < n) {
      grad[3*j2  ] += dudx;
      grad[3*j2+1] += dudy;
      grad[3*j2+2] += dudz;
    }
    if (i3 < n) {
      grad[3*j3  ] += dudx;
      grad[3*j3+1] += dudy;
      grad[3*j3+2] += dudz;
    }
    if (i4 < n) {
      grad[3*j4  ] += dudx;
      grad[3*j4+1] += dudy;
      grad[3*j4+2] += dudz;
    }
  }

}


static void
FX_HTetGradP3d(int *node, int n, float *xcoor, float *ycoor, float *zcoor,
               FXcoor *hcoor, int dim, float *src, float *src1, float *src2,
               float *hsrc, float *grad, float *vol)
{
  int    i, i1, i2, i3, i4, j1, j2, j3, j4;
  float  *source;
  double x1, y1, z1, x2, y2, z2, x3, y3, z3, x4, y4, z4, V, c[3][4];
  double u1, u2, u3, u4, dudx, dudy, dudz;

  i1 = node[0]-1;
  if (i1 < n) {
    x1 = xcoor[i1];
    y1 = ycoor[i1];
    z1 = zcoor[i1];
  } else {
    x1 = hcoor[i1-n].x;
    y1 = hcoor[i1-n].y;
    z1 = hcoor[i1-n].z;
  }
  i2 = node[1]-1;
  if (i2 < n) {
    x2 = xcoor[i2];
    y2 = ycoor[i2];
    z2 = zcoor[i2];
  } else {
    x2 = hcoor[i2-n].x;
    y2 = hcoor[i2-n].y;
    z2 = hcoor[i2-n].z;
  }
  i3 = node[2]-1;
  if (i3 < n) {
    x3 = xcoor[i3];
    y3 = ycoor[i3];
    z3 = zcoor[i3];
  } else {
    x3 = hcoor[i3-n].x;
    y3 = hcoor[i3-n].y;
    z3 = hcoor[i3-n].z;
  }
  i4 = node[3]-1;
  if (i4 < n) {
    x4 = xcoor[i4];
    y4 = ycoor[i4];
    z4 = zcoor[i4];
  } else {
    x4 = hcoor[i4-n].x;
    y4 = hcoor[i4-n].y;
    z4 = hcoor[i4-n].z;
  }

  V = (x2-x1)*( (y3-y1)*(z4-z1)-(z3-z1)*(y4-y1) ) -
      (x3-x1)*( (y2-y1)*(z4-z1)-(z2-z1)*(y4-y1) ) +
      (x4-x1)*( (y2-y1)*(z3-z1)-(z2-z1)*(y3-y1) );
  if (i1 < n) vol[i1] += V;
  if (i2 < n) vol[i2] += V;
  if (i3 < n) vol[i3] += V;
  if (i4 < n) vol[i4] += V;

  c[0][0] = (z2-z3)*(y3-y4)-(z3-z4)*(y2-y3);
  c[0][1] = (z4-z1)*(y3-y4)-(z3-z4)*(y4-y1);
  c[0][2] = (z4-z1)*(y1-y2)-(z1-z2)*(y4-y1);
  c[0][3] = (z2-z3)*(y1-y2)-(z1-z2)*(y2-y3);

  c[1][0] = (x2-x3)*(z3-z4)-(x3-x4)*(z2-z3);
  c[1][1] = (x4-x1)*(z3-z4)-(x3-x4)*(z4-z1);
  c[1][2] = (x4-x1)*(z1-z2)-(x1-x2)*(z4-z1);
  c[1][3] = (x2-x3)*(z1-z2)-(x1-x2)*(z2-z3);

  c[2][0] = (y2-y3)*(x3-x4)-(y3-y4)*(x2-x3);
  c[2][1] = (y4-y1)*(x3-x4)-(y3-y4)*(x4-x1);
  c[2][2] = (y4-y1)*(x1-x2)-(y1-y2)*(x4-x1);
  c[2][3] = (y2-y3)*(x1-x2)-(y1-y2)*(x2-x3);

  if (dim == 1) {
    if (i1 < n) {
      u1 = src[i1];
    } else {
      u1 = hsrc[i1-n];
    }
    if (i2 < n) {
      u2 = src[i2];
    } else {
      u2 = hsrc[i2-n];
    }
    if (i3 < n) {
      u3 = src[i3];
    } else {
      u3 = hsrc[i3-n];
    }
    if (i4 < n) {
      u4 = src[i4];
    } else {
      u4 = hsrc[i4-n];
    }
    dudx = c[0][0]*u1 + c[0][1]*u2 + c[0][2]*u3 + c[0][3]*u4;
    dudy = c[1][0]*u1 + c[1][1]*u2 + c[1][2]*u3 + c[1][3]*u4;
    dudz = c[2][0]*u1 + c[2][1]*u2 + c[2][2]*u3 + c[2][3]*u4;

    if (i1 < n) {
      grad[3*i1  ] += dudx;
      grad[3*i1+1] += dudy;
      grad[3*i1+2] += dudz;
    }
    if (i2 < n) {
      grad[3*i2  ] += dudx;
      grad[3*i2+1] += dudy;
      grad[3*i2+2] += dudz;
    }
    if (i3 < n) {
      grad[3*i3  ] += dudx;
      grad[3*i3+1] += dudy;
      grad[3*i3+2] += dudz;
    }
    if (i4 < n) {
      grad[3*i4  ] += dudx;
      grad[3*i4+1] += dudy;
      grad[3*i4+2] += dudz;
    }
    return;
  }

  for (i = 0; i < dim; i++) {
                source = src;
    if (i == 1) source = src1;
    if (i == 2) source = src2;

    if (i1 < n) {
      j1 = i1*dim+i;
      u1 = source[i1];
    } else {
      j1 = (i1-n)*dim+i;
      u1 = hsrc[j1];
    }
    if (i2 < n) {
      j2 = i2*dim+i;
      u2 = source[i2];
    } else {
      j2 = (i2-n)*dim+i;
      u2 = hsrc[j2];
    }
    if (i3 < n) {
      j3 = i3*dim+i;
      u3 = source[i3];
    } else {
      j3 = (i3-n)*dim+i;
      u3 = hsrc[j3];
    }
    if (i4 < n) {
      j4 = i4*dim+i;
      u4 = source[i4];
    } else {
      j4 = (i4-n)*dim+i;
      u4 = hsrc[j4];
    }
    dudx = c[0][0]*u1 + c[0][1]*u2 + c[0][2]*u3 + c[0][3]*u4;
    dudy = c[1][0]*u1 + c[1][1]*u2 + c[1][2]*u3 + c[1][3]*u4;
    dudz = c[2][0]*u1 + c[2][1]*u2 + c[2][2]*u3 + c[2][3]*u4;

    if (i1 < n) {
      grad[3*j1  ] += dudx;
      grad[3*j1+1] += dudy;
      grad[3*j1+2] += dudz;
    }
    if (i2 < n) {
      grad[3*j2  ] += dudx;
      grad[3*j2+1] += dudy;
      grad[3*j2+2] += dudz;
    }
    if (i3 < n) {
      grad[3*j3  ] += dudx;
      grad[3*j3+1] += dudy;
      grad[3*j3+2] += dudz;
    }
    if (i4 < n) {
      grad[3*j4  ] += dudx;
      grad[3*j4+1] += dudy;
      grad[3*j4+2] += dudz;
    }
  }

}


static void
FX_HTetGrad(FXglobal *globals, int *node, int dim, float *src, float *hsrc,
            float *grad, float *vol)
{

  if ((globals->cmask & 64) == 0) {
    FX_HTetGradPtr(node, globals->total_nodes, globals->coor, globals->hcoor,
                   dim, src, hsrc, grad, vol);
  } else {
    if (dim != 3) {
      FX_HTetGradP3d(node, globals->total_nodes, globals->xcoor, globals->ycoor,
                     globals->zcoor, globals->hcoor, dim, src, NULL, NULL,
                     hsrc, grad, vol);
    } else {
      FX_HTetGradP3d(node, globals->total_nodes, globals->xcoor, globals->ycoor,
                     globals->zcoor, globals->hcoor, dim, globals->velx, 
                     globals->vely, globals->velz, hsrc, grad, vol);
    }
  }

}


/*
 *  calculates the scalar gradient
 */

void
FX_Grad(FXglobal *globals, float *scalar, float *hscal, float **grad)
{
  int    kc, i, j, k, ib, ni, nj, nk, nc, kns, kz, bnode, type, node[8], tet[4];
  float  *vol, *sgrad;
  static int pyr2tet[2][4] = { 1,2,4,5, 2,3,4,5 };
  static int pri2tet[3][4] = { 1,2,4,5, 2,3,4,5, 1,4,5,6 };
  static int hex2tet[6][4] = { 2,4,5,6, 1,2,4,5, 2,3,4,6,
                               3,4,6,8, 4,5,6,8, 3,6,7,8 };

  *grad = NULL;
  vol = (float *) malloc(globals->total_nodes*sizeof(float));
  if (vol == NULL) {
    printf(" FX Error: MALLOC for Volume storage (FX_Grad)!\n");
    return;
  }
  sgrad = (float *) malloc(3*globals->total_nodes*sizeof(float));
  if (sgrad == NULL) {
    free(vol);
    printf(" FX Error: MALLOC for Gradient storage (FX_Grad)!\n");
    return;
  }

  /* initialize counters */
  for (i = 0; i < globals->total_nodes; i++)
    vol[i] = sgrad[3*i] = sgrad[3*i+1] = sgrad[3*i+2] = 0.0;

  /* structured blocks */

  if ((globals->nblock > 0) && (globals->iblank != NULL))
    bnode = globals->block[0].last_node - 
            globals->block[0].ni*globals->block[0].nj*globals->block[0].nk;
  for (ib = 0; ib < globals->nblock; ib++) {
    ni  = globals->block[ib].ni;
    nj  = globals->block[ib].nj;
    nk  = globals->block[ib].nk;
    nc  = (ni-1)*(nj-1)*(nk-1);
    kns = globals->block[ib].last_node - ni*nj*nk + 1;
    if (globals->iblank == NULL) {
      for (kc = 0; kc < nc; kc++) {
        i  =  kc%(ni-1);
        j  = (kc/(ni-1))%(nj-1);
        k  =  kc/((ni-1)*(nj-1));
        kz =  kns + i + j*ni + k*ni*nj;
        node[0] = kz;
        node[1] = kz + 1;
        node[2] = kz + 1 + ni;
        node[3] = kz     + ni;
        node[4] = kz          + ni*nj;
        node[5] = kz + 1      + ni*nj;
        node[6] = kz + 1 + ni + ni*nj;
        node[7] = kz     + ni + ni*nj;
        for (j = 0; j < 6; j++) {
          for (i = 0; i < 4; i++) {
            k = hex2tet[j][i]-1;
            tet[i] = node[k];
           }
           FX_TetGrad(globals, tet, 1, scalar, sgrad, vol);
        }
      }
    } else {
      for (kc = 0; kc < nc; kc++) {
        i  =  kc%(ni-1);
        j  = (kc/(ni-1))%(nj-1);
        k  =  kc/((ni-1)*(nj-1));
        kz =  kns + i + j*ni + k*ni*nj;
        node[0] = kz;
        node[1] = kz + 1;
        node[2] = kz + 1 + ni;
        node[3] = kz     + ni;
        node[4] = kz          + ni*nj;
        node[5] = kz + 1      + ni*nj;
        node[6] = kz + 1 + ni + ni*nj;
        node[7] = kz     + ni + ni*nj;
        for (j = i = 0; i < 8; i++)
          if (globals->iblank[node[i]-bnode-1] == 0) j++;
        if (j != 0) continue;

        for (j = 0; j < 6; j++) {
          for (i = 0; i < 4; i++) {
            k = hex2tet[j][i]-1;
            tet[i] = node[k];
           }
           FX_TetGrad(globals, tet, 1, scalar, sgrad, vol);
        }
      }
    }
  }

  /* disjoint and halo elements */

  if ((globals->cmask & 128) == 0) {

    /* tetrahedra */
    for (kc = 0; kc < globals->ntets; kc++)
      FX_TetGrad(globals, globals->tets[kc].nodes, 1, scalar, sgrad, vol);

    /* pyramids */
    for (kc = 0; kc < globals->npyra; kc++)
      for (j = 0; j < 2; j++) {
        for (i = 0; i < 4; i++) {
          k = pyr2tet[j][i]-1;
          node[i] = globals->pyra[kc].nodes[k];
        }
        FX_TetGrad(globals, node, 1, scalar, sgrad, vol);
      }

    /* prisms */
    for (kc = 0; kc < globals->nprism; kc++)
      for (j = 0; j < 3; j++) {
        for (i = 0; i < 4; i++) {
          k = pri2tet[j][i]-1;
          node[i] = globals->prism[kc].nodes[k];
        }
        FX_TetGrad(globals, node, 1, scalar, sgrad, vol);
      }

    /* hexahedra */
    for (kc = 0; kc < globals->nhexa; kc++)
      for (j = 0; j < 6; j++) {
        for (i = 0; i < 4; i++) {
          k = hex2tet[j][i]-1;
          node[i] = globals->hexa[kc].nodes[k];
        }
        FX_TetGrad(globals, node, 1, scalar, sgrad, vol);
      }

    /* halos */
    for (kc = 0; kc < globals->nhcell; kc++)
      switch (globals->hcell[kc].type) {

      case 1:
        /* tetrahedra */
        FX_HTetGrad(globals, globals->hcell[kc].nodes, 1, scalar, hscal, 
                    sgrad, vol);
        break;

      case 2:
        /* pyramids */
        for (j = 0; j < 2; j++) {
          for (kz = i = 0; i < 4; i++) {
            k = pyr2tet[j][i]-1;
            tet[i] = globals->hcell[kc].nodes[k];
            if (tet[i] <= globals->total_nodes) kz++;
          }
          if (kz != 0) FX_HTetGrad(globals, tet, 1, scalar, hscal, sgrad, vol);
        }
        break;

      case 3:
        /* prisms */
        for (j = 0; j < 3; j++) {
          for (kz = i = 0; i < 4; i++) {
            k = pri2tet[j][i]-1;
            tet[i] = globals->hcell[kc].nodes[k];
            if (tet[i] <= globals->total_nodes) kz++;
          }
          if (kz != 0) FX_HTetGrad(globals, tet, 1, scalar, hscal, sgrad, vol);
        }
        break;

      case 4:
        /* hexahedra */
        for (j = 0; j < 6; j++) {
          for (kz = i = 0; i < 4; i++) {
            k = hex2tet[j][i]-1;
            tet[i] = globals->hcell[kc].nodes[k];
            if (tet[i] <= globals->total_nodes) kz++;
          }
          if (kz != 0) FX_HTetGrad(globals, tet, 1, scalar, hscal, sgrad, vol);
        }
      }

  } else {

    /* construct elements on-the-fly */

    for (kc = 0; kc < globals->total_dis; kc++) {
      i = kc+1;
      FXCELLGET(&i, &type, node);

      switch (type) {

      case 1:
        /* tetrahedra */
        FX_TetGrad(globals, node, 1, scalar, sgrad, vol);
        break;

      case 2:
        /* pyramids */
        for (j = 0; j < 2; j++) {
          for (i = 0; i < 4; i++) {
            k = pyr2tet[j][i]-1;
            tet[i] = node[k];
          }
          FX_TetGrad(globals, tet, 1, scalar, sgrad, vol);
        }
        break;

      case 3:
        /* prisms */
        for (j = 0; j < 3; j++) {
          for (i = 0; i < 4; i++) {
            k = pri2tet[j][i]-1;
            tet[i] = node[k];
          }
          FX_TetGrad(globals, tet, 1, scalar, sgrad, vol);
        }
        break;

      case 4:
        /* hexahedra */
        for (j = 0; j < 6; j++) {
          for (i = 0; i < 4; i++) {
            k = hex2tet[j][i]-1;
            tet[i] = node[k];
          }
          FX_TetGrad(globals, tet, 1, scalar, sgrad, vol);
        }
        break;

      default:
        printf(" FX Warning: FXcellGet INDEX = %d -> TYPE = %d (FX_Grad)!\n",
               kc+1, type);
      }
    }

    /* halos */
    for (kc = 0; kc < globals->nhcell; kc++) {
      i = -(kc+1);
      FXCELLGET(&i, &type, node);

      switch (type) {

      case 1:
        /* tetrahedra */
        FX_HTetGrad(globals, node, 1, scalar, hscal, sgrad, vol);
        break;

      case 2:
        /* pyramids */
        for (j = 0; j < 2; j++) {
          for (kz = i = 0; i < 4; i++) {
            k = pyr2tet[j][i]-1;
            tet[i] = node[k];
            if (tet[i] <= globals->total_nodes) kz++;
          }
          if (kz != 0) FX_HTetGrad(globals, tet, 1, scalar, hscal, sgrad, vol);
        }
        break;

      case 3:
        /* prisms */
        for (j = 0; j < 3; j++) {
          for (kz = i = 0; i < 4; i++) {
            k = pri2tet[j][i]-1;
            tet[i] = node[k];
            if (tet[i] <= globals->total_nodes) kz++;
          }
          if (kz != 0) FX_HTetGrad(globals, tet, 1, scalar, hscal, sgrad, vol);
        }
        break;

      case 4:
        /* hexahedra */
        for (j = 0; j < 6; j++) {
          for (kz = i = 0; i < 4; i++) {
            k = hex2tet[j][i]-1;
            tet[i] = node[k];
            if (tet[i] <= globals->total_nodes) kz++;
          }
          if (kz != 0) FX_HTetGrad(globals, tet, 1, scalar, hscal, sgrad, vol);
        }
        break;

      default:
        printf(" FX Warning: FXcellGet INDEX = %d -> TYPE = %d (FX_Grad)!\n",
               -(kc+1), type);
      }
    }

  }

  /* finish vol weighting */

  for (i = 0; i < globals->total_nodes; i++) {
    if (vol[i] == 0.0) continue;
    sgrad[3*i  ] /= vol[i];
    sgrad[3*i+1] /= vol[i];
    sgrad[3*i+2] /= vol[i];
  }

  free(vol);
  *grad = sgrad;
}


/*
 *  calculates the velocity gradient
 */

void
FX_VelGrad(FXglobal *globals, float **grad)
{
  int    kc, i, j, k, ib, ni, nj, nk, nc, kns, kz, bnode, node[8], tet[4], type;
  float  *vol, *vgrad, *vel, *hvel;
  static int pyr2tet[2][4] = { 1,2,4,5, 2,3,4,5 };
  static int pri2tet[3][4] = { 1,2,4,5, 2,3,4,5, 1,4,5,6 };
  static int hex2tet[6][4] = { 2,4,5,6, 1,2,4,5, 2,3,4,6,
                               3,4,6,8, 4,5,6,8, 3,6,7,8 };

  *grad = NULL;
  vol = (float *) malloc(globals->total_nodes*sizeof(float));
  if (vol == NULL) {
    printf(" FX Error: MALLOC for Volume storage (FX_VelGrad)!\n");
    return;
  }
  vgrad = (float *) malloc(9*globals->total_nodes*sizeof(float));
  if (vgrad == NULL) {
    free(vol);
    printf(" FX Error: MALLOC for Gradient storage (FX_VelGrad)!\n");
    return;
  }
  vel  = (float *) globals->vel;
  hvel = (float *) globals->hvel;

  /* initialize counters */
  for (i = 0; i < globals->total_nodes; i++) {
    vol[i] = 0.0;
    for (j = 0; j < 9; j++) vgrad[9*i+j] = 0.0;
  }

  /* structured blocks */

  if ((globals->nblock > 0) && (globals->iblank != NULL))
    bnode = globals->block[0].last_node - 
            globals->block[0].ni*globals->block[0].nj*globals->block[0].nk;
  for (ib = 0; ib < globals->nblock; ib++) {
    ni  = globals->block[ib].ni;
    nj  = globals->block[ib].nj;
    nk  = globals->block[ib].nk;
    nc  = (ni-1)*(nj-1)*(nk-1);
    kns = globals->block[ib].last_node - ni*nj*nk + 1;
    if (globals->iblank == NULL) {
      for (kc = 0; kc < nc; kc++) {
        i  =  kc%(ni-1);
        j  = (kc/(ni-1))%(nj-1);
        k  =  kc/((ni-1)*(nj-1));
        kz =  kns + i + j*ni + k*ni*nj;
        node[0] = kz;
        node[1] = kz + 1;
        node[2] = kz + 1 + ni;
        node[3] = kz     + ni;
        node[4] = kz          + ni*nj;
        node[5] = kz + 1      + ni*nj;
        node[6] = kz + 1 + ni + ni*nj;
        node[7] = kz     + ni + ni*nj;
        for (j = 0; j < 6; j++) {
          for (i = 0; i < 4; i++) {
            k = hex2tet[j][i]-1;
            tet[i] = node[k];
           }
           FX_TetGrad(globals, tet, 3, vel, vgrad, vol);
        }
      }
    } else {
      for (kc = 0; kc < nc; kc++) {
        i  =  kc%(ni-1);
        j  = (kc/(ni-1))%(nj-1);
        k  =  kc/((ni-1)*(nj-1));
        kz =  kns + i + j*ni + k*ni*nj;
        node[0] = kz;
        node[1] = kz + 1;
        node[2] = kz + 1 + ni;
        node[3] = kz     + ni;
        node[4] = kz          + ni*nj;
        node[5] = kz + 1      + ni*nj;
        node[6] = kz + 1 + ni + ni*nj;
        node[7] = kz     + ni + ni*nj;
        for (j = i = 0; i < 8; i++)
          if (globals->iblank[node[i]-bnode-1] == 0) j++;
        if (j != 0) continue;

        for (j = 0; j < 6; j++) {
          for (i = 0; i < 4; i++) {
            k = hex2tet[j][i]-1;
            tet[i] = node[k];
           }
           FX_TetGrad(globals, tet, 3, vel, vgrad, vol);
        }
      }
    }
  }

  /* disjoint and halo elements */

  if ((globals->cmask & 128) == 0) {

    /* tetrahedra */
    for (kc = 0; kc < globals->ntets; kc++)
      FX_TetGrad(globals, globals->tets[kc].nodes, 3, vel, vgrad, vol);

    /* pyramids */
    for (kc = 0; kc < globals->npyra; kc++)
      for (j = 0; j < 2; j++) {
        for (i = 0; i < 4; i++) {
          k = pyr2tet[j][i]-1;
          node[i] = globals->pyra[kc].nodes[k];
        }
        FX_TetGrad(globals, node, 3, vel, vgrad, vol);
      }

    /* prisms */
    for (kc = 0; kc < globals->nprism; kc++)
      for (j = 0; j < 3; j++) {
        for (i = 0; i < 4; i++) {
          k = pri2tet[j][i]-1;
          node[i] = globals->prism[kc].nodes[k];
        }
        FX_TetGrad(globals, node, 3, vel, vgrad, vol);
      }

    /* hexahedra */
    for (kc = 0; kc < globals->nhexa; kc++)
      for (j = 0; j < 6; j++) {
        for (i = 0; i < 4; i++) {
          k = hex2tet[j][i]-1;
          node[i] = globals->hexa[kc].nodes[k];
        }
        FX_TetGrad(globals, node, 3, vel, vgrad, vol);
      }

    /* halos */
    for (kc = 0; kc < globals->nhcell; kc++)
      switch (globals->hcell[kc].type) {

      case 1:
        /* tetrahedra */
        FX_HTetGrad(globals, globals->hcell[kc].nodes, 3, vel, hvel,
                    vgrad, vol);
        break;

      case 2:
        /* pyramids */
        for (j = 0; j < 2; j++) {
          for (kz = i = 0; i < 4; i++) {
            k = pyr2tet[j][i]-1;
            tet[i] = globals->hcell[kc].nodes[k];
            if (tet[i] <= globals->total_nodes) kz++;
          }
          if (kz != 0) FX_HTetGrad(globals, tet, 3, vel, hvel, vgrad, vol);
        }
        break;

      case 3:
        /* prisms */
        for (j = 0; j < 3; j++) {
          for (kz = i = 0; i < 4; i++) {
            k = pri2tet[j][i]-1;
            tet[i] = globals->hcell[kc].nodes[k];
            if (tet[i] <= globals->total_nodes) kz++;
          }
          if (kz != 0) FX_HTetGrad(globals, tet, 3, vel, hvel, vgrad, vol);
        }
        break;

      case 4:
        /* hexahedra */
        for (j = 0; j < 6; j++) {
          for (kz = i = 0; i < 4; i++) {
            k = hex2tet[j][i]-1;
            tet[i] = globals->hcell[kc].nodes[k];
            if (tet[i] <= globals->total_nodes) kz++;
          }
          if (kz != 0) FX_HTetGrad(globals, tet, 3, vel, hvel, vgrad, vol);
        }
      }

  } else {

    /* construct elements on-the-fly */

    for (kc = 0; kc < globals->total_dis; kc++) {
      i = kc+1;
      FXCELLGET(&i, &type, node);

      switch (type) {

      case 1:
        /* tetrahedra */
        FX_TetGrad(globals, node, 3, vel, vgrad, vol);
        break;

      case 2:
        /* pyramids */
        for (j = 0; j < 2; j++) {
          for (i = 0; i < 4; i++) {
            k = pyr2tet[j][i]-1;
            tet[i] = node[k];
          }
          FX_TetGrad(globals, tet, 3, vel, vgrad, vol);
        }
        break;

      case 3:
        /* prisms */
        for (j = 0; j < 3; j++) {
          for (i = 0; i < 4; i++) {
            k = pri2tet[j][i]-1;
            tet[i] = node[k];
          }
          FX_TetGrad(globals, tet, 3, vel, vgrad, vol);
        }
        break;

      case 4:
        /* hexahedra */
        for (j = 0; j < 6; j++) {
          for (i = 0; i < 4; i++) {
            k = hex2tet[j][i]-1;
            tet[i] = node[k];
          }
          FX_TetGrad(globals, tet, 3, vel, vgrad, vol);
        }
        break;

      default:
        printf(" FX Warning: FXcellGet INDEX = %d -> TYPE = %d (FX_VelGrad)!\n",
               kc+1, type);
      }
    }

    /* halos */
    for (kc = 0; kc < globals->nhcell; kc++) {
      i = -(kc+1);
      FXCELLGET(&i, &type, node);

      switch (type) {

      case 1:
        /* tetrahedra */
        FX_HTetGrad(globals, node, 3, vel, hvel, vgrad, vol);
        break;

      case 2:
        /* pyramids */
        for (j = 0; j < 2; j++) {
          for (kz = i = 0; i < 4; i++) {
            k = pyr2tet[j][i]-1;
            tet[i] = node[k];
            if (tet[i] <= globals->total_nodes) kz++;
          }
          if (kz != 0) FX_HTetGrad(globals, tet, 3, vel, hvel, vgrad, vol);
        }
        break;

      case 3:
        /* prisms */
        for (j = 0; j < 3; j++) {
          for (kz = i = 0; i < 4; i++) {
            k = pri2tet[j][i]-1;
            tet[i] = node[k];
            if (tet[i] <= globals->total_nodes) kz++;
          }
          if (kz != 0) FX_HTetGrad(globals, tet, 3, vel, hvel, vgrad, vol);
        }
        break;

      case 4:
        /* hexahedra */
        for (j = 0; j < 6; j++) {
          for (kz = i = 0; i < 4; i++) {
            k = hex2tet[j][i]-1;
            tet[i] = node[k];
            if (tet[i] <= globals->total_nodes) kz++;
          }
          if (kz != 0) FX_HTetGrad(globals, tet, 3, vel, hvel, vgrad, vol);
        }
        break;

      default:
        printf(" FX Warning: FXcellGet INDEX = %d -> TYPE = %d (FX_VelGrad)!\n",
               -(kc+1), type);
      }
    }

  }

  /* finish vol weighting */

  for (i = 0; i < globals->total_nodes; i++) {
    if (vol[i] == 0.0) continue;
    for (j = 0; j < 9; j++) vgrad[9*i+j] /= vol[i];
  }

  free(vol);
  *grad = vgrad;
}
