/*
 * FX_Init.c
 *
 * FX module that Initializes the FX system.
 *
 * Copyright 1999 - 2000, Massachusetts Institute of Technology.
 *
 */

#include <stdio.h>
#include <stdlib.h>

#include "FX_types.h"

#define MAXERR 20		/* maximum # of error lines for FX_Check */


#ifdef WIN32

#define PREFIX __stdcall

#else

#define PREFIX

#ifdef UNDERSCORE

#define FXCELL       fxcell_
#define FXCELLPTR    fxcellptr_
#define FXCELLGET    fxcellget_
#define FXSURFACE    fxsurface_
#define FXSURFACEPTR fxsurfaceptr_
#define FXGRID       fxgrid_
#define FXGRIDPTR    fxgridptr_
#define FXGRIDP3D    fxgridp3d_
#define FXBLANK      fxblank_
#define FXBLANKPTR   fxblankptr_

#define FX_INIT      fx_init_
#define FX_RTCLEAN   fx_rtclean_
#define FX_CLOSE     fx_close_

#else

#define FXCELL       fxcell
#define FXCELLPTR    fxcellptr
#define FXCELLGET    fxcellget
#define FXSURFACE    fxsurface
#define FXSURFACEPTR fxsurfaceptr
#define FXGRID       fxgrid
#define FXGRIDPTR    fxgridptr
#define FXGRIDP3D    fxgridp3d
#define FXBLANK      fxblank
#define FXBLANKPTR   fxblankptr

#define FX_INIT      fx_init
#define FX_RTCLEAN   fx_rtclean
#define FX_CLOSE     fx_close

#endif
#endif


/* globals */

  FXglobal *FX_global = NULL;


/* externals */

  extern void PREFIX
         FX_RTCLEAN( void );

  extern void PREFIX
         FXCELL   (FXtet *,  FXpyra *,  FXprism *,  FXhexa *,  FXhalo *);
  extern void PREFIX
         FXCELLPTR(FXtet **, FXpyra **, FXprism **, FXhexa **, FXhalo *);
  extern void PREFIX
         FXCELLGET(int *, int *, int *);
  extern void PREFIX
         FXSURFACE   (FXbc *, int *,  FXfacet *);
  extern void PREFIX
         FXSURFACEPTR(FXbc *, int **, FXfacet **);
  extern void PREFIX
         FXGRID   (FXcoor *,  FXcoor *);
  extern void PREFIX
         FXGRIDPTR(FXcoor **, FXcoor *);
  extern void PREFIX
         FXGRIDP3D(float **, float **, float **, FXcoor *);
  extern void PREFIX
         FXBLANK   (int *);
  extern void PREFIX
         FXBLANKPTR(int **);


static void
FX_Clean(void)
{
  if (((FX_global->cmask & 1) == 0) && ((FX_global->cmask & 128) == 0)) {
    if (FX_global->tets  != NULL) free(FX_global->tets);
    if (FX_global->pyra  != NULL) free(FX_global->pyra);
    if (FX_global->prism != NULL) free(FX_global->prism);
    if (FX_global->hexa  != NULL) free(FX_global->hexa);
  }
  if (FX_global->block != NULL) free(FX_global->block);
  if (FX_global->bc    != NULL) free(FX_global->bc);
  if ((FX_global->cmask & 2) == 0) {
    if (FX_global->scel  != NULL) free(FX_global->scel);
    if (FX_global->facet != NULL) free(FX_global->facet);
  }
  if ((FX_global->cmask & 4) == 0)
    if (FX_global->coor != NULL) free(FX_global->coor);
  if (FX_global->scalar != NULL) free(FX_global->scalar);
  if ((FX_global->cmask & 8) == 0)
    if (FX_global->vel != NULL) free(FX_global->vel);
  if ((FX_global->cmask & 32) == 0)
    if (FX_global->iblank != NULL) free(FX_global->iblank);
  if (FX_global->hcoor   != NULL) free(FX_global->hcoor);
  if (FX_global->hscalar != NULL) free(FX_global->hscalar);
  if (FX_global->hvel    != NULL) free(FX_global->hvel);
  if (FX_global->hcell   != NULL) free(FX_global->hcell);

  free(FX_global);
  FX_global = NULL;
}


int
FX_Check(void)
{
  int kc, err, i, n, total, nb, ist, type, cell[8];

  total = FX_global->total_nodes + FX_global->nhalo;

  if ((FX_global->cmask & 128) == 0) {

    for (err = kc = 0; kc < FX_global->ntets; kc++) {
      for (i = 0; i < 4; i++)
        if ((FX_global->tets[kc].nodes[i] < 1) ||
            (FX_global->tets[kc].nodes[i] > FX_global->total_nodes)) {
          err++;
          printf(" FX_Check Index Error: tets[%d][%d] = %d (%d)\n", kc, i,
                 FX_global->tets[kc].nodes[i], FX_global->total_nodes);
        }
      if (err > MAXERR) break;
    }

    for (kc = 0; kc < FX_global->npyra; kc++) {
      for (i = 0; i < 5; i++)
        if ((FX_global->pyra[kc].nodes[i] < 1) ||
            (FX_global->pyra[kc].nodes[i] > FX_global->total_nodes)) {
          err++;
          printf(" FX_Check Index Error: pyra[%d][%d] = %d (%d)\n", kc, i,
                 FX_global->pyra[kc].nodes[i], FX_global->total_nodes);
        }
      if (err > MAXERR) break;
    }

    for (kc = 0; kc < FX_global->nprism; kc++) {
      for (i = 0; i < 6; i++)
        if ((FX_global->prism[kc].nodes[i] < 1) ||
            (FX_global->prism[kc].nodes[i] > FX_global->total_nodes)) {
          err++;
          printf(" FX_Check Index Error: prism[%d][%d] = %d (%d)\n", kc, i,
                 FX_global->prism[kc].nodes[i], FX_global->total_nodes);
        }
      if (err > MAXERR) break;
    }

    for (kc = 0; kc < FX_global->nhexa; kc++) {
      for (i = 0; i < 8; i++)
        if ((FX_global->hexa[kc].nodes[i] < 1) ||
            (FX_global->hexa[kc].nodes[i] > FX_global->total_nodes)) {
          err++;
          printf(" FX_Check Index Error: hexa[%d][%d] = %d (%d)\n", kc, i,
                 FX_global->hexa[kc].nodes[i], FX_global->total_nodes);
        }
      if (err > MAXERR) break;
    }

    for (kc = 0; kc < FX_global->nhcell; kc++) {
      n = 4;
      if (FX_global->hcell[kc].type == 2) n = 5;
      if (FX_global->hcell[kc].type == 3) n = 6;
      if (FX_global->hcell[kc].type == 4) n = 8;
      for (i = 0; i < n; i++)
        if ((FX_global->hcell[kc].nodes[i] < 1) ||
            (FX_global->hcell[kc].nodes[i] > total)) {
          err++;
          printf(" FX_Check Index Error: %d hcell[%d][%d] = %d (%d)\n",
                 n, kc, i, FX_global->hcell[kc].nodes[i], total);
        }
      if (err > MAXERR) break;
    }

  } else {

    for (err = kc = 0; kc < FX_global->total_dis; kc++) {
      i = kc+1;
      FXCELLGET(&i, &type, cell);

      n = 0;
      if (type == 1) n = 4;
      if (type == 2) n = 5;
      if (type == 3) n = 6;
      if (type == 4) n = 8;
      if (n == 0) {
        printf(" FX_Check Type Error: %d (1-4), cell #%d\n", i, type);
        err++;
      } else {
        for (i = 0; i < n; i++)
          if ((cell[i] < 1) || (cell[i] > FX_global->total_nodes)) {
            printf(" FX_Check Index Error: cell #%d  node[%d] = %d (1-%d)\n",
                   kc+1, i, cell[i], FX_global->total_nodes);
            err++;
          }
      }
      if (err > MAXERR) break;
    }

    for (kc = 0; kc < FX_global->nhcell; kc++) {
      i = -(kc+1);
      FXCELLGET(&i, &type, cell);

      n = 0;
      if (type == 1) n = 4;
      if (type == 2) n = 5;
      if (type == 3) n = 6;
      if (type == 4) n = 8;
      if (n == 0) {
        printf(" FX_Check Type Error: %d (1-4), cell #%d\n", i, type);
        err++;
      } else {
        for (i = 0; i < n; i++)
          if ((cell[i] < 1) || (cell[i] > total)) {
            printf(" FX_Check Index Error: cell #%d  node[%d] = %d (1-%d)\n",
                   -(kc+1), i, cell[i], total);
            err++;
          }
      }
      if (err > MAXERR) break;
    }
  }

  for (ist = nb = 0; nb < FX_global->nbc; nb++) {
    for (kc = ist; kc < FX_global->bc[nb].last_facet; kc++) {
      n = 4;
      if (FX_global->facet[kc].nodes[3] == 0) n = 3;
      for (i = 0; i < n; i++)
        if ((FX_global->facet[kc].nodes[i] < 1) ||
            (FX_global->facet[kc].nodes[i] > FX_global->total_nodes)) {
          err++;
          printf(" FX_Check Index Error: facet[%d][%d] = %d (%d)\n", kc, i,
                 FX_global->facet[kc].nodes[i], FX_global->total_nodes);
        }
      if ((FX_global->scel[kc] < 1) ||
          (FX_global->scel[kc] > FX_global->total_cells)) {
        err++;
        printf(" FX_Check Index Error: scel[%d] = %d (%d)\n", kc,
               FX_global->scel[kc], FX_global->total_cells);
      }
      if (err > MAXERR) break;

    }
    ist = FX_global->bc[nb].last_facet;
  }

  return err;
}


void PREFIX
FX_CLOSE()
{
  if (FX_global == NULL) {
    printf(" FX Error: Not Initialized (FX_Close)!\n");
  } else {
    if (FX_global->rt != 0) FX_RTCLEAN();
    FX_Clean();
  }
}


void PREFIX
FX_INIT(float *gam, int *iopt, int *knode, int *nhalo, int *ntets, int *npyra,
        int *nprism, int *nhexa, int *nblocks, int *blocks, int *nhcell,
        int *nsurf, int *nbc, int *istat)
{
  int i, ni, nj, nk, lnode, lcell;

  if (FX_global != NULL) {
    printf(" FX Error: Already Initialized (FX_Init)!\n");
    *istat = -1;
    return;
  }

  if ((*iopt < 0) || (*iopt > 3)) {
    printf(" FX Error: Illegal IOPT = %d (FX_Init)!\n", *iopt);
    *istat = -2;
    return;
  }

  FX_global = (FXglobal *) malloc(sizeof(FXglobal));
  if (FX_global == NULL) {
    printf(" FX Error: Allocation of Globals (FX_Init)!\n");
    *istat = -4;
    return;
  }

  /* initial fill for globals */

  FX_global->gamma   = *gam;
  FX_global->iopt    = *iopt;
  FX_global->cmask   = *istat;
  FX_global->time    = 0.0;
  FX_global->status  = 0;
  FX_global->rt      = 0;
  FX_global->nnode   = *knode;
  FX_global->ntets   = *ntets;
  FX_global->tets    = NULL;
  FX_global->npyra   = *npyra;
  FX_global->pyra    = NULL;
  FX_global->nprism  = *nprism;
  FX_global->prism   = NULL;
  FX_global->nhexa   = *nhexa;
  FX_global->hexa    = NULL;
  FX_global->mblock  = *nblocks;
  FX_global->nblock  = *nblocks;
  FX_global->block   = NULL;
  FX_global->nbc     = *nbc;
  FX_global->bc      = NULL;
  FX_global->nfacet  = *nsurf;
  FX_global->facet   = NULL;
  FX_global->scel    = NULL;
  FX_global->coor    = NULL;
  FX_global->xcoor   = NULL;
  FX_global->ycoor   = NULL;
  FX_global->zcoor   = NULL;
  FX_global->scalar  = NULL;
  FX_global->vel     = NULL;
  FX_global->velx    = NULL;
  FX_global->vely    = NULL;
  FX_global->velz    = NULL;
  FX_global->iblank  = NULL;

  FX_global->nhalo   = *nhalo;
  FX_global->hcoor   = NULL;
  FX_global->hscalar = NULL;
  FX_global->hvel    = NULL;
  FX_global->nhcell  = *nhcell;
  FX_global->hcell   = NULL;

  FX_global->sload   = 0;
  FX_global->vload   = 0;

  /* get total sizes */

  lnode = FX_global->nnode;
  lcell = FX_global->ntets + FX_global->npyra + FX_global->nprism +
          FX_global->nhexa;
  FX_global->total_dis = lcell;
  if (FX_global->nblock > 0) {
    FX_global->block = (FXblock *) malloc(FX_global->nblock*sizeof(FXblock));
    if (FX_global->block == NULL) {
      printf(" FX Error: Allocation of Structured Blocks (FX_Init)!\n");
      FX_Clean();
      *istat = -3;
      return;
    }
    for (i = 0; i < FX_global->nblock; i++) {
      ni = blocks[i*3  ];
      nj = blocks[i*3+1];
      nk = blocks[i*3+2];
      FX_global->block[i].ni = ni;
      FX_global->block[i].nj = nj;
      FX_global->block[i].nk = nk;
      lcell += (ni-1)*(nj-1)*(nk-1);
      lnode += ni*nj*nk;
      FX_global->block[i].last_cell = lcell;
      FX_global->block[i].last_node = lnode;
    }
  }
  FX_global->total_cells = lcell;
  FX_global->total_nodes = lnode;

  /* allocate memory for structure steady cases */

  if (FX_global->iopt != 3) {

    if ((FX_global->cmask & 128) == 0) {

      /* disjoint elements */

      if (FX_global->nhcell > 0) {
        FX_global->hcell = (FXhalo *) malloc(FX_global->nhcell*sizeof(FXhalo));
        if (FX_global->hcell == NULL) {
          printf(" FX Error: Allocation of Halo Cells (FX_Init)!\n");
          FX_Clean();
          *istat = -3;
          return;
        }
      }

      if ((FX_global->cmask & 1) == 0) {
        if (FX_global->ntets > 0) {
          FX_global->tets = (FXtet *) malloc(FX_global->ntets*sizeof(FXtet));
          if (FX_global->tets == NULL) {
            printf(" FX Error: Allocation of Tetrahedra (FX_Init)!\n");
            FX_Clean();
            *istat = -3;
            return;
          }
        }
        if (FX_global->npyra > 0) {
          FX_global->pyra = (FXpyra *) malloc(FX_global->npyra*sizeof(FXpyra));
          if (FX_global->pyra == NULL) {
            printf(" FX Error: Allocation of Pyramids (FX_Init)!\n");
            FX_Clean();
            *istat = -3;
            return;
          }
        }
        if (FX_global->nprism > 0) {
          FX_global->prism = (FXprism *) 
                             malloc(FX_global->nprism*sizeof(FXprism));
          if (FX_global->prism == NULL) {
            printf(" FX Error: Allocation of Prisms (FX_Init)!\n");
            FX_Clean();
            *istat = -3;
            return;
          }
        }
        if (FX_global->nhexa > 0) {
          FX_global->hexa = (FXhexa *) malloc(FX_global->nhexa*sizeof(FXhexa));
          if (FX_global->hexa == NULL) {
            printf(" FX Error: Allocation of Hexahedra (FX_Init)!\n");
            FX_Clean();
            *istat = -3;
            return;
          }
        }
        if (FX_global->total_dis + FX_global->nhcell > 0) 
          FXCELL   ( FX_global->tets,  FX_global->pyra,  FX_global->prism,
                     FX_global->hexa,  FX_global->hcell);
      } else {
        if (FX_global->total_dis + FX_global->nhcell > 0) 
          FXCELLPTR(&FX_global->tets, &FX_global->pyra, &FX_global->prism,
                    &FX_global->hexa,  FX_global->hcell);
      }
    }

    /* surface data */

    if (FX_global->nbc > 0) {
      FX_global->bc = (FXbc *) malloc(FX_global->nbc*sizeof(FXbc));
      if (FX_global->bc == NULL) {
        printf(" FX Error: Allocation of Boundary Groups (FX_Init)!\n");
        FX_Clean();
        *istat = -3;
        return;
      }
      if ((FX_global->cmask & 2) == 0) {
        FX_global->facet = (FXfacet *)malloc(FX_global->nfacet*sizeof(FXfacet));
        if (FX_global->facet == NULL) {
          printf(" FX Error: Allocation of Surface Facets (FX_Init)!\n");
          FX_Clean();
          *istat = -3;
          return;
        }
        FX_global->scel = (int *) malloc(FX_global->nfacet*sizeof(int));
        if (FX_global->scel == NULL) {
          printf(" FX Error: Allocation of Surface Cells (FX_Init)!\n");
          FX_Clean();
          *istat = -3;
          return;
        }
        FXSURFACE   (FX_global->bc,  FX_global->scel,  FX_global->facet);
      } else {
        FXSURFACEPTR(FX_global->bc, &FX_global->scel, &FX_global->facet);
      }
    }

    /* halo data */

    if (FX_global->nhalo > 0) {
      FX_global->hcoor = (FXcoor *) malloc(FX_global->nhalo*sizeof(FXcoor));
      if (FX_global->hcoor == NULL) {
        printf(" FX Error: Allocation of Halo Coordinates (FX_Init)!\n");
        FX_Clean();
        *istat = -3;
        return;
      }
      FX_global->hvel = (FXcoor *) malloc(FX_global->nhalo*sizeof(FXcoor));
      if (FX_global->hvel == NULL) {
        printf(" FX Error: Allocation of Halo Velocities (FX_Init)!\n");
        FX_Clean();
        *istat = -3;
        return;
      }
      FX_global->hscalar = (float *) malloc(FX_global->nhalo*sizeof(float));
      if (FX_global->hscalar == NULL) {
        printf(" FX Error: Allocation of Halo Scalars (FX_Init)!\n");
        FX_Clean();
        *istat = -3;
        return;
      }
    }

    /* node data */

    if ((FX_global->cmask & 4) == 0) {
      FX_global->coor = (FXcoor *)malloc(FX_global->total_nodes*sizeof(FXcoor));
      if (FX_global->coor == NULL) {
        printf(" FX Error: Allocation of Coordinates (FX_Init)!\n");
        FX_Clean();
        *istat = -3;
        return;
      }
      if (FX_global->iopt < 2) FXGRID(FX_global->coor, FX_global->hcoor);
    } else {
      if (FX_global->iopt < 2)
        if ((FX_global->cmask & 64) == 0) {
          FXGRIDPTR(&FX_global->coor, FX_global->hcoor);
        } else {
          FXGRIDP3D(&FX_global->xcoor, &FX_global->ycoor, &FX_global->zcoor,
                     FX_global->hcoor);
        }
    }
    if ((FX_global->cmask & 8) == 0) {
      FX_global->vel = (FXcoor *) malloc(FX_global->total_nodes*sizeof(FXcoor));
      if (FX_global->vel == NULL) {
        printf(" FX Error: Allocation of Velocity Field (FX_Init)!\n");
        FX_Clean();
        *istat = -3;
        return;
      }
    }
    FX_global->scalar = (float *) malloc(FX_global->total_nodes*sizeof(float));
    if (FX_global->scalar == NULL) {
      printf(" FX Error: Allocation of Scalar Field (FX_Init)!\n");
      FX_Clean();
      *istat = -3;
      return;
    }

    /* Iblanking */

    if (FX_global->nblock > 0)
      if ((FX_global->cmask & 16) != 0)
        if ((FX_global->cmask & 32) == 0) {
          lnode = FX_global->total_nodes - FX_global->nnode;
          FX_global->iblank = (int *) malloc(lnode*sizeof(int));
          if (FX_global->iblank == NULL) {
            printf(" FX Error: Allocation of IBlank (FX_Init)!\n");
            FX_Clean();
            *istat = -3;
            return;
          }
          if (FX_global->iopt < 2) FXBLANK   ( FX_global->iblank);
        } else {
          if (FX_global->iopt < 2) FXBLANKPTR(&FX_global->iblank);
        }

    /* Structure Check */

    if (FX_Check() != 0) {
      FX_Clean();
      *istat = -5;
      return;
    }
  }

  *istat = 0;
}
