/*
 * FX_RTSubs.c
 *
 * FX Residence Time modules.
 *
 * Copyright 1999 - 2000, Massachusetts Institute of Technology.
 *
 */

#include <stdlib.h>
#include <stdio.h>

#include "FX_types.h"


#ifdef WIN32

#define PREFIX __stdcall

#else

#define PREFIX

#ifdef UNDERSCORE

#define FX_RTSURFACE fx_rtsurface_
#define FX_RTCELL    fx_rtcell_
#define FX_STENCIL   fx_stencil_
#define FX_GETVEL    fx_getvel_
#define FX_GETSCAL   fx_getscal_
#define FX_GETXYZ    fx_getxyz_
#define FX_GETBC     fx_getbc_

#define FXSCAL       fxscal_
#define FXVEL        fxvel_
#define FXVELPTR     fxvelptr_
#define FXVELP3D     fxvelp3d_

#else

#define FX_RTSURFACE fx_rtsurface
#define FX_RTCELL    fx_rtcell
#define FX_STENCIL   fx_stencil
#define FX_GETVEL    fx_getvel
#define FX_GETSCAL   fx_getscal
#define FX_GETXYZ    fx_getxyz
#define FX_GETBC     fx_getbc

#define FXSCAL       fxscal
#define FXVEL        fxvel
#define FXVELPTR     fxvelptr
#define FXVELP3D     fxvelp3d

#endif
#endif


/* structures */

typedef struct {
  int node;		/* node number for neighbor */
  int thread;		/* thread to next neighbor */
} nodeList;


/* globals */

  extern FXglobal *FX_global;


/* externals */

  extern void PREFIX FXSCAL(int *, float *,  float *);
  extern void PREFIX FXVEL(FXcoor *,  FXcoor *);
  extern void PREFIX FXVELPTR(FXcoor **,  FXcoor *);
  extern void PREFIX FXVELP3D(float **, float **, float **, FXcoor *);

  extern void        FX_Cell2Nodes(FXglobal *, int, int *, int *);
  extern void        FX_GenSurf(FXglobal *globals, float val, float *scalar,
                                int *npts, FXcoor **xyz,
                                int *ntris, int **tris, int **cell);


void PREFIX
FX_RTSURFACE(float *rt, float *rtv, int *npts, FXcoor **xyz,
             int *ntris, int **tris, int **cell)
{
  FX_GenSurf(FX_global, *rtv, rt, npts, xyz, ntris, tris, cell);
}


int PREFIX
FX_RTCELL(int *kc, int *node)
{
  int n;

  FX_Cell2Nodes(FX_global, *kc, &n, node);

  if (n == 4) return 1;
  if (n == 5) return 2;
  if (n == 6) return 3;
  if (n == 8) return 4;

  printf(" FX_RTCell Error for cell # %d, n = %d\n", *kc, n);
  return 0;
}


/*
 * This routine examines all cells and constructs a nearest-neighbor
 * list in the form of the number of nodes and the index into the
 * list of nodes.  Pointers to the filled list and space for weights
 * are returned.
 */

void PREFIX
FX_STENCIL (int indsmo[][2], int **knsmo, float **wghsmo)
{
  int      listCnt, listLen, kn, kc, fkc, kt, ke, k1, k2, ie, pcell[8];
  int      totalc, totaln, *kni;
  nodeList *nl;

  static int nedges[4] = {6, 8, 9, 12};
  static int  edges[4][12][2] =
               { 1,2, 1,3, 1,4, 2,3, 2,4, 3,4, 0,0, 0,0, 0,0, 0,0, 0,0, 0,0,
                 1,2, 2,3, 3,4, 4,1, 1,5, 2,5, 3,5, 4,5, 0,0, 0,0, 0,0, 0,0,
                 1,2, 2,3, 3,4, 4,1, 5,6, 2,5, 3,5, 1,6, 4,6, 0,0, 0,0, 0,0,
                 1,2, 2,3, 3,4, 4,1, 5,6, 6,7, 7,8, 8,5, 1,5, 2,6, 3,7, 4,8 };

  totalc  = FX_global->total_cells + FX_global->nhcell;
  totaln  = FX_global->total_nodes + FX_global->nhalo;
  listLen = totaln*6;
  nl = (nodeList *) malloc(listLen*sizeof(nodeList));
  if (nl == NULL) {
    printf(" Allocation Error in FX_Stencil!\n");
    exit(1);
  }

  for (kn = 0; kn < totaln; kn++)  indsmo[kn][1] = -1;

  for (listCnt = kc = 0; kc < totalc; kc++) {
    fkc = kc+1;
    kt  = FX_RTCELL(&fkc, pcell)-1;
    for (ke = 0; ke < nedges[kt]; ke++) {
      k1 = pcell[edges[kt][ke][0]-1]-1;
      k2 = pcell[edges[kt][ke][1]-1]-1;

      /* link list with k1 */
      if (indsmo[k1][1] == -1) {

        /* virgin node */
        if (listCnt == listLen) {
          listLen *= 1.2;
          nl = (nodeList *) realloc(nl, listLen*sizeof(nodeList));
          if (nl == NULL) {
            printf(" ReAllocation Error in FX_Stencil!\n");
            exit(1);
          }
        }
        indsmo[k1][1]      = listCnt;
        nl[listCnt].node   = k2;
        nl[listCnt].thread = -1;
        listCnt++;
      } else {

        /* old node */
        ie = indsmo[k1][1];

again1:
        if (nl[ie].node != k2) {
          if (nl[ie].thread != -1) {
            ie = nl[ie].thread;
            goto again1;
          }
          if (listCnt == listLen) {
            listLen *= 1.2;
            nl = (nodeList *) realloc(nl, listLen*sizeof(nodeList));
            if (nl == NULL) {
              printf(" ReAllocation Error in FX_Stencil!\n");
              exit(1);
            }
          }
          nl[ie].thread      = listCnt;
          nl[listCnt].node   = k2;
          nl[listCnt].thread = -1;
          listCnt++;
        }
      }

      /* link list with k2 */
      if (indsmo[k2][1] == -1) {

        /* virgin node */
        if (listCnt == listLen) {
          listLen *= 1.2;
          nl = (nodeList *) realloc(nl, listLen*sizeof(nodeList));
          if (nl == NULL) {
            printf(" ReAllocation Error in FX_Stencil!\n");
            exit(1);
          }
        }
        indsmo[k2][1]      = listCnt;
        nl[listCnt].node   = k1;
        nl[listCnt].thread = -1;
        listCnt++;
      } else {

        /* old node */
        ie = indsmo[k2][1];

again2:
        if (nl[ie].node != k1) {
          if (nl[ie].thread != -1) {
            ie = nl[ie].thread;
            goto again2;
          }
          if (listCnt == listLen) {
            listLen *= 1.2;
            nl = (nodeList *) realloc(nl, listLen*sizeof(nodeList));
            if (nl == NULL) {
              printf(" ReAllocation Error in FX_Stencil!\n");
              exit(1);
            }
          }
          nl[ie].thread      = listCnt;
          nl[listCnt].node   = k1;
          nl[listCnt].thread = -1;
          listCnt++;
        }
      }
    }
  }

  /* count neighbors */

  for (kt = kn = 0; kn < FX_global->total_nodes; kn++)
    if (indsmo[kn][1] == -1) {
      indsmo[kn][0] = 0;
    } else {
      kc = 1;
      ie = indsmo[kn][1];
      while (nl[ie].thread != -1) {
        ie = nl[ie].thread;
        kc++;
      }
      indsmo[kn][0] = kc;
      kt += kc;
    }

  kni = (int *) malloc(kt*sizeof(int));
  if (kni == NULL) {
    printf(" Allocation Error in FX_Stencil (Indices)!\n");
    exit(1);
  }
  *knsmo  = kni;
  *wghsmo = (float *) malloc(kt*sizeof(float));
  if (*wghsmo == NULL) {
    printf(" Allocation Error in FX_Stencil (Weights)!\n");
    exit(1);
  }

  /* store node numbers */

  for (kt = kn = 0; kn < FX_global->total_nodes; kn++)
    if (indsmo[kn][1] != -1) {
      ie = indsmo[kn][1];
      indsmo[kn][1] = kt;
      kni[kt] = nl[ie].node+1;
      kt++;
      while (nl[ie].thread != -1) {
        ie = nl[ie].thread;
        kni[kt] = nl[ie].node+1;
        kt++;
      }
    }

  free(nl);
}


void PREFIX
FX_GETVEL(float **vel)
{
  int    i;
  FXcoor *coor;

  if (FX_global->vload == 0) {
    if ((FX_global->cmask & 8) == 0) {
      FXVEL    (FX_global->vel, FX_global->hvel);
    } else {
      if ((FX_global->cmask & 64) == 0) {
        FXVELPTR(&FX_global->vel, FX_global->hvel);
      } else {
        FXVELP3D(&FX_global->velx, &FX_global->vely, &FX_global->velz,
                  FX_global->hvel);
      }
    }
    FX_global->vload = 1;
  }

  if ((FX_global->nhalo == 0) && ((FX_global->cmask & 64) == 0)) {
    *vel = (float *) FX_global->vel;
  } else {
    coor = (FXcoor *) *vel;
    if ((FX_global->cmask & 64) == 0) {
      for (i = 0; i < FX_global->total_nodes; i++) coor[i] = FX_global->vel[i];
    } else {
      for (i = 0; i < FX_global->total_nodes; i++) {
        coor[i].x = FX_global->velx[i];
        coor[i].y = FX_global->vely[i];
        coor[i].z = FX_global->velz[i];
      }
    }
    for (i = 0; i < FX_global->nhalo; i++)
      coor[i+FX_global->total_nodes] = FX_global->hvel[i];
  }
}


void PREFIX
FX_GETSCAL(int *type, float *s)
{
  int i;

  if (FX_global->sload != *type) {
    FX_global->sload = *type;
    FXSCAL(&FX_global->sload, FX_global->scalar, FX_global->hscalar);
  }

  for (i = 0; i < FX_global->total_nodes; i++) s[i] = FX_global->scalar[i];
  for (i = 0; i < FX_global->nhalo; i++)
    s[i+FX_global->total_nodes] = FX_global->hscalar[i];
}


void PREFIX
FX_GETXYZ(float **xyz)
{
  int    i;
  FXcoor *coor;

  if ((FX_global->nhalo == 0) && ((FX_global->cmask & 64) == 0)) {
    *xyz = (float *) FX_global->coor;
  } else {
    coor = (FXcoor *) *xyz;
    if ((FX_global->cmask & 64) == 0) {
      for (i = 0; i < FX_global->total_nodes; i++) coor[i] = FX_global->coor[i];
    } else {
      for (i = 0; i < FX_global->total_nodes; i++) {
        coor[i].x = FX_global->xcoor[i];
        coor[i].y = FX_global->ycoor[i];
        coor[i].z = FX_global->zcoor[i];
      }
    }
    for (i = 0; i < FX_global->nhalo; i++)
      coor[i+FX_global->total_nodes] = FX_global->hcoor[i];
  }
}


void PREFIX
FX_GETBC(int *nbc, int **nsurf, int **scel, int **surf)
{
  *nbc   =         FX_global->nbc;
  *nsurf = (int *) FX_global->bc;
  *scel  =         FX_global->scel;
  *surf  = (int *) FX_global->facet;
}
