/*
 * FX_SepnLine.c
 *
 * FX module that computes Separation and Attachment Lines.
 *
 * Copyright 1999 - 2000, Massachusetts Institute of Technology.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "FX_types.h"


/***********************************************************************
 *  Fortran and C function definitions                                 *
 ***********************************************************************/

#ifdef WIN32

  #define PREFIX __stdcall

#else

  #define PREFIX

  #ifdef UNDERSCORE

    #define FXVEL       fxvel_
    #define FXVELPTR    fxvelptr_
    #define FXVELP3D    fxvelp3d_
    #define FX_SEPNLINE fx_sepnline_

  #else

    #define FXVEL       fxvel
    #define FXVELPTR    fxvelptr
    #define FXVELP3D    fxvelp3d
    #define FX_SEPNLINE fx_sepnline

  #endif

#endif


/***********************************************************************
 *  Parameters                                                         *
 ***********************************************************************/
 
    #define TOL                       1.0e-9
    #define FALSE	              0
    #define TRUE		      1

    #define TWO_REAL_ROOTS	      1
    #define COMPLEX_CONJUGATE_ROOTS   0
    #define COMPLEX_CONJUGATE        -3

    #define UNKNOWN		     -1
    #define NODE	              0
    #define ATTRACTING_NODE           0
    #define REPELLING_NODE            1
    #define PROPER_NODE               1
    #define SADDLE	              2
    #define CENTER	              3
    #define SPIRAL	              4



/***********************************************************************
 *  Global Variables                                                   *
 ***********************************************************************/

    extern FXglobal *FX_global;



/***********************************************************************
 *  External Function Prototypes                                       *
 ***********************************************************************/
 
  extern void PREFIX FXVEL(FXcoor *,  FXcoor *);
  extern void PREFIX FXVELPTR(FXcoor **,  FXcoor *);
  extern void PREFIX FXVELP3D(float **, float **, float **,  FXcoor *);
  extern void        FX_VelGrad(FXglobal *, float **);

  static void extract_separation_lines(float *, float **, int *, float **, 
                                       int *);


void PREFIX
FX_SEPNLINE(int *sepind, float **sepxyz, int *attind, float **attxyz)

/***********************************************************************
*                                                                      *
*     TITLE:    SEPARATION AND ATTACHMENT LINE DETECTION               *
*                                                                      *
*     AUTHOR:   David N. Kenwright                                     *
*                                                                      *
*     BUGS:     Report problems to: kenwright@mit.edu                  *
*                                                                      *
*     RELEASE:  July 2000 in FX Library                                *
*                                                                      *
*     NOTES:  This code is based on an algorithm by D. Kenwright       *
*	      "Automatic Detection of Open and Closed Separation       *
*             and Attachment Lines", IEEE Visualization '98,           *
*	      IEEE Computer Society Press, October 1998.               *
*                                                                      *
************************************************************************/

{
    int   i;
    float *vel_grad;

    *sepxyz = *attxyz = NULL;
    for (i = 0; i < FX_global->nbc; i++) sepind[i] = attind[i] = 0;

    /* check that FX has been initialized */

    if (FX_global == NULL) {
      printf(" FX Error: NOT Initialized (FX_SepnLine)!\n");
      return;
    }
    if (FX_global->status != 0) {
      printf(" FX Error: Initialization Status = %d (FX_SepnLine)!\n",
           FX_global->status);
      return;
    }

    /* get the velocity in order to compute the velocity gradient tensor */

    if (FX_global->vload == 0) {
      if ((FX_global->cmask & 8) == 0) {
        FXVEL(FX_global->vel, FX_global->hvel);
      } else {
        if ((FX_global->cmask & 64) == 0) {
          FXVELPTR(&FX_global->vel, FX_global->hvel);
        } else {
          FXVELP3D(&FX_global->velx, &FX_global->vely, &FX_global->velz,
                   FX_global->hvel);
        }
      }
      FX_global->vload = 1;
    }

    /* Compute the velocity gradient tensor */

    FX_VelGrad(FX_global, &vel_grad);
    if (vel_grad == NULL) return;

    /* extract the separation and attachment lines */

    extract_separation_lines(vel_grad, sepxyz, sepind, attxyz, attind);

    /* clean up */
    free(vel_grad);
}


/***********************************************************************
 *  Eigenvalue and Eigenvector Functions                               *
 ***********************************************************************/


static int
check_phase_portrait(double *p, double *discriminant, complex *lambda1,
                     complex *lambda2)
/*
 * 
 *  Determine the phase portrait of the system. It is assumed that the real 
 *  eigenvalues have already been sorted so that lambda1->re > lambda2->re.
 *  The following tests may otherwise give an incorrect result.
 * 
 */
{
    int classification;

    if (*discriminant >= 0.0) {
        if( (lambda1->re < 0.0) && (lambda2->re < 0.0) ) {
          classification = ATTRACTING_NODE;
        } else if( (lambda1->re > 0.0) && (lambda2->re > 0.0) ) {
          classification = REPELLING_NODE;
	} else if( lambda2->re == lambda1->re ) {
      	  classification = PROPER_NODE;
	} else if( (lambda2->re < 0.0) && (lambda1->re > 0.0) ) {
       	  classification = SADDLE;
	} else {
          classification = UNKNOWN;
        }
    } else {
       if (*p == 0.0) {
          classification = CENTER;
       } else {
          classification = SPIRAL;
       }
    }

    return classification;
}



static int
eigenvalues2x2(double b1, double c1, double b2, double c2, int *root_type,
               complex *lambda_1, complex *lambda_2)
/*
 * 
 *  Determine the eigenvalues of a 2x2 matrix
 * 
 */
{
    double p, q, discriminant;
    double sqrt_discriminant;
    int classification;

    /* Coefficients of the characteristic equation */

    p = b1+c2;
    q = b1*c2-c1*b2;
    discriminant = p*p - 4.0*q;
    
    if( fabs(discriminant) < 1.0e-16 ) discriminant = 0.0;

    /* Calculate eigenvalues */

    if (discriminant >= 0.0) {

       /*  two real eigenvalues */
       sqrt_discriminant = sqrt (discriminant);
       lambda_1->re = (p + sqrt_discriminant)*0.5;
       lambda_2->re = (p - sqrt_discriminant)*0.5;
       lambda_1->im = lambda_2->im = 0.0;
       *root_type = TWO_REAL_ROOTS;
    } else {

       /* complex conjugate or degenerate eigenvalues */
       sqrt_discriminant = sqrt (discriminant * -1.0);
       lambda_1->re = lambda_2->re = p*0.5;
       lambda_1->im = +sqrt_discriminant*0.5;
       lambda_2->im = -sqrt_discriminant*0.5;
       *root_type = COMPLEX_CONJUGATE_ROOTS;
    }

    /* Arrange real eigenvalues in decreasing order */

    if (discriminant > 0.0)
       if (lambda_1->re < lambda_2->re) {
          double tmp   = lambda_1->re;
          lambda_1->re = lambda_2->re;
          lambda_2->re = tmp;
       }
    
    /* Determine the phase portrait  */

    classification = check_phase_portrait(&p, &discriminant,
                                          lambda_1, lambda_2);

    /* Check for a zero descriminant -- case of repeated roots */

    if (fabs(discriminant) < TOL) classification = UNKNOWN;

    return classification;
}



static void
real_eigenvectors2x2(double a11, double a12, double a21, double a22, 
                     complex *lambda1, complex *lambda2, complex *x1, 
                     complex *y1, complex *x2, complex *y2)
/*
 * 
 *  Determine the eigenvectors of a 2x2 matrix
 * 
 */
{
    double b11, b12, b21, b22;
    
    /*  test for the most degenerate case  */
    
    if (a11==0.0 && a12==0.0 && a21==0.0 && a22==0.0) {
       x1->re = 1.0;
       y1->re = 0.0;
       x2->re = 0.0;
       y2->re = 1.0;
       return;
    }

    /*  find the eigenvector coresponding to the first eigenvalue  */

    b11 = a11 - lambda1->re;
    b12 = a12;
    b21 = a21;
    b22 = a22 - lambda1->re;
    if (fabs(b11) > TOL) {
       y1->re = 1.0;
       x1->re = -b12/b11;
    } else if (fabs(b21) > TOL) {
       y1->re = 1.0;
       x1->re = -b22/b21;
    } else {
       y1->re = 0.0;
       x1->re = 1.0;
    }

    /*  find the eigenvector coresponding to the second eigenvalue  */

    b11 = a11 - lambda2->re;
    b12 = a12;
    b21 = a21;
    b22 = a22 - lambda2->re;
    if (fabs(b11) > TOL) {
       y2->re = 1.0;
       x2->re = -b12/b11;
    } else if (fabs(b21) > TOL) {
       y2->re = 1.0;
       x2->re = -b22/b21;
    } else {
       y2->re = 0.0;
       x2->re = 1.0;
    }
}



/***********************************************************************
 *  Matrix and Vector Functions                                        *
 ***********************************************************************/


static int matrix_inv_2(double a[2][2], double b[2][2])
/*
 * 
 *  Invert a 2x2 matrix
 * 
 */
{
    double inv_determinant;
    double determinant;

    determinant = a[0][0]*a[1][1] - a[0][1]*a[1][0];
    if (fabs(determinant) > TOL) {
      inv_determinant = 1.0/determinant;
    } else {
      return 0;
    }    

    b[0][0] =  a[1][1]*inv_determinant;
    b[0][1] = -a[0][1]*inv_determinant;
    b[1][0] = -a[1][0]*inv_determinant;
    b[1][1] =  a[0][0]*inv_determinant;
    return 1;
}


static int matrix_inv_3(double a[3][3], double a_inv[3][3])
/*
 * Invert a 3x3 matrix
 *
 */
{
    double d01d12md11d02;
    double d01d22md21d02;
    double d11d22md21d12;
    double det, inv_det;
    
    d01d12md11d02 = a[0][1] * a[1][2] - a[1][1] * a[0][2];
    d01d22md21d02 = a[0][1] * a[2][2] - a[2][1] * a[0][2];
    d11d22md21d12 = a[1][1] * a[2][2] - a[2][1] * a[1][2];
    det = a[0][0]*d11d22md21d12 - a[1][0]*d01d22md21d02 + a[2][0]*d01d12md11d02;
    if (fabs(det) > TOL) {
       inv_det = 1.0/det;
    } else {
       return 0;
    }    

    a_inv[0][0] =  d11d22md21d12*inv_det;
    a_inv[0][1] = -d01d22md21d02*inv_det;
    a_inv[0][2] =  d01d12md11d02*inv_det;
    a_inv[1][0] =  (a[2][0] * a[1][2] - a[1][0] * a[2][2])*inv_det;
    a_inv[1][1] =  (a[0][0] * a[2][2] - a[2][0] * a[0][2])*inv_det;
    a_inv[1][2] =  (a[1][0] * a[0][2] - a[0][0] * a[1][2])*inv_det;
    a_inv[2][0] =  (a[1][0] * a[2][1] - a[2][0] * a[1][1])*inv_det;
    a_inv[2][1] =  (a[2][0] * a[0][1] - a[0][0] * a[2][1])*inv_det;
    a_inv[2][2] =  (a[0][0] * a[1][1] - a[1][0] * a[0][1])*inv_det;
    return 1;
}


static void matrix_vec_3(double a[3][3], double b[3], double c[3])
/*
 * 
 *  Multiply a 3x3 matrix by a 3x1 vector
 * 
 */
{
    c[0] = a[0][0]*b[0] + a[0][1]*b[1] + a[0][2]*b[2];
    c[1] = a[1][0]*b[0] + a[1][1]*b[1] + a[1][2]*b[2];
    c[2] = a[2][0]*b[0] + a[2][1]*b[1] + a[2][2]*b[2];
}


static void matrix_mult_3(double a[3][3], double b[3][3], double c[3][3])
/*
 * 
 *  Multiply two 3x3 matrices
 * 
 */
{
    c[0][0] = a[0][0]*b[0][0] + a[0][1]*b[1][0] + a[0][2]*b[2][0];
    c[0][1] = a[0][0]*b[0][1] + a[0][1]*b[1][1] + a[0][2]*b[2][1];
    c[0][2] = a[0][0]*b[0][2] + a[0][1]*b[1][2] + a[0][2]*b[2][2];

    c[1][0] = a[1][0]*b[0][0] + a[1][1]*b[1][0] + a[1][2]*b[2][0];
    c[1][1] = a[1][0]*b[0][1] + a[1][1]*b[1][1] + a[1][2]*b[2][1];
    c[1][2] = a[1][0]*b[0][2] + a[1][1]*b[1][2] + a[1][2]*b[2][2];

    c[2][0] = a[2][0]*b[0][0] + a[2][1]*b[1][0] + a[2][2]*b[2][0];
    c[2][1] = a[2][0]*b[0][1] + a[2][1]*b[1][1] + a[2][2]*b[2][1];
    c[2][2] = a[2][0]*b[0][2] + a[2][1]*b[1][2] + a[2][2]*b[2][2];
}


static void cross_product_3(double a[3], double b[3], double c[3])
/*
 * 
 *  Evaluate the cross product of two 3D vectors
 * 
 */
{
    c[0] = a[1]*b[2] - a[2]*b[1];
    c[1] = a[2]*b[0] - a[0]*b[2];
    c[2] = a[0]*b[1] - a[1]*b[0];
}


static void normalize_vector_3( double v[3], double vn[3] )
/*
 * 
 *  Normalize a 3D vector
 * 
 */
{
  double mag = sqrt (v[0]*v[0] + v[1]*v[1] + v[2]*v[2]);
  if (mag==0.0) mag = 1e-9;
  vn[0] = v[0]/mag;
  vn[1] = v[1]/mag;
  vn[2] = v[2]/mag;
}



static double magnitude_3(double a[3])
/*
 *
 *  Evaluate the vector magnitude
 *
 */
{
  return sqrt(a[0]*a[0] + a[1]*a[1] + a[2]*a[2]);
}



/***********************************************************************
 *  Vector and Tensor Transformation  Functions                        *
 ***********************************************************************/


static void 
convert_to_canonical(double b[2], double c_inv[2][2], double x, double y,
                     double *xc, double *yc)
/*
 * 
 *  Change the coordinates from (x,y) to (xc,yc)
 * 
 */
{
    double d[2];

    d[0] = x-b[0];
    d[1] = y-b[1];
    *xc = c_inv[0][0]*d[0] + c_inv[0][1]*d[1];
    *yc = c_inv[1][0]*d[0] + c_inv[1][1]*d[1];
}




static void
convert_to_2D_basis(double b[2], double c[2][2], double xc, double yc,
                    double *x, double *y)
/*
 * 
 *  Change the coordinates from (xc,yc) to (x,y)
 * 
 */
{
    *x = c[0][0]*xc + c[0][1]*yc + b[0];
    *y = c[1][0]*xc + c[1][1]*yc + b[1];
}




static void
transform_tensor(double q[3][3], double qinv[3][3], double tnsr[3][3],
                 double tnsrRot[3][3])
/*
 * 
 * Covariant Orthonormal Tensor Transformation for change of coordinate basis
 * 
 */
{
    double temp[3][3];

    matrix_mult_3(q, tnsr, temp);
    matrix_mult_3(temp, qinv, tnsrRot);
}



static void Change_Basis(facet *triptr)
/*
 *
 *   Orthonormal rotation of Cartesian vectors and Tensors
 *
 */
{
      int i;
      double vec1[3],  vec2[3],  vec3[3], mag;
      double base1[3], base2[3], base3[3];

      /* establish a new coordinate basis in the plane of the triangle */

      for (i = 0; i < 3; i++) {
        vec1[i] = triptr->xyz[1][i]-triptr->xyz[0][i];
	vec2[i] = triptr->xyz[2][i]-triptr->xyz[0][i];
      }

      cross_product_3(vec1, vec2, vec3);
      cross_product_3(vec1, vec3, vec2);

      normalize_vector_3(vec1, base1);
      normalize_vector_3(vec2, base2);
      normalize_vector_3(vec3, base3);

      /* build the coordinate transformation matrix from the basis vectors */

      for (i = 0; i < 3; i++) {
        triptr->q[0][i] = base1[i];
        triptr->q[1][i] = base2[i];
        triptr->q[2][i] = base3[i];
      }

      /* transform the triangle coordinates (x' = q x) */

      matrix_vec_3(triptr->q, triptr->xyz[0], triptr->xyzRot[0]);
      matrix_vec_3(triptr->q, triptr->xyz[1], triptr->xyzRot[1]);
      matrix_vec_3(triptr->q, triptr->xyz[2], triptr->xyzRot[2]);

      /* test for a no-slip boundary condition */

      mag = 0.0;
      for (i = 0; i < 3; i++) mag += magnitude_3(triptr->uvw[i]);
      if (mag > TOL) {

        /*
         *      transform the triangle velocity vectors (v' = q v) and
         *                assign the tangential velocity vectors
         */
        matrix_vec_3(triptr->q, triptr->uvw[0], triptr->uvwRot[0]);
        matrix_vec_3(triptr->q, triptr->uvw[1], triptr->uvwRot[1]);
        matrix_vec_3(triptr->q, triptr->uvw[2], triptr->uvwRot[2]);

      } else {

         /*
          * transform the velocity gradient tensor (T' = q T qinv) and
          *           assign the wall shear stress vectors
          */
         matrix_inv_3(triptr->q, triptr->qinv);

         for (i = 0; i < 3; i++) {
           transform_tensor(triptr->q, triptr->qinv, triptr->tnsr[i],
                            triptr->tnsrRot[i]);
           triptr->uvwRot[i][0] = triptr->tnsrRot[i][0][2];
           triptr->uvwRot[i][1] = triptr->tnsrRot[i][1][2];
           triptr->uvwRot[i][2] = triptr->tnsrRot[i][2][2];
         }
      }
}




static void convert_to_physical(facet *triptr, double comp[2], float *phys)
/*
 *
 *    Convert a 3D computational coordinate into a 3D physical coordinate
 *
 */
{
  double vec[3], pos[3]; 
  double qinv[3][3];

  matrix_inv_3(triptr->q, qinv);

  vec[0] = comp[0];
  vec[1] = comp[1];
  vec[2] = triptr->xyzRot[0][2];

  matrix_vec_3(qinv, vec, pos);

  phys[0] = (float) pos[0];
  phys[1] = (float) pos[1];
  phys[2] = (float) pos[2];
}




/***********************************************************************
 *  Primary algorithm functions                                        *
 ***********************************************************************/


static int
Create_Triangle_List(int nb, facet **triangle, int *ntri, float *vel_grad)
/*
 *
 *    Load the physical coordinates and velocity vectors for each triangle
 *
 */
{
      int ntris, first_facet, last_facet, num_facets, kc, kz, offset;
      int index, row, col, i, j;
      static int node[2][3] = { 0,1,2, 0,2,3 };
      facet *triList;
      facet *triptr;

      /* get the first and last facet for this block */

      if (nb == 0) {
        first_facet = 0;
      } else {
        first_facet = FX_global->bc[nb-1].last_facet;
      }
      last_facet = FX_global->bc[nb].last_facet;
      num_facets = last_facet-first_facet;

      /* allocate memory for the the facets */

      triList = (facet *) malloc(4*num_facets*sizeof(facet));
      if (triList == NULL) {
        printf(" FX Error: MALLOC for facet storage (FX_SepnLine)!\n");
	return 1;
      }

      /* loop over all the triangles an quadrilaterals in the block */

      index = 0;
      for (kc = first_facet; kc < FX_global->bc[nb].last_facet; kc++) {

          /* check whether the facet is a quadrilateral or a triangle */
          ntris = 1;
          if (FX_global->facet[kc].nodes[3] != 0) ntris = 2;

          /* load the triangle data into the facet structure */
	  for (j = 0; j < ntris; j++) {
            triptr = &triList[index];
            for (i = 0; i < 3; i++) {
              kz = FX_global->facet[kc].nodes[node[j][i]]-1;

              /* get the vertex coordinates and velocities */
              triptr->xyz[i][0] = FX_global->coor[kz].x;
              triptr->xyz[i][1] = FX_global->coor[kz].y;
              triptr->xyz[i][2] = FX_global->coor[kz].z;
              triptr->uvw[i][0] = FX_global->vel[kz].x;
              triptr->uvw[i][1] = FX_global->vel[kz].y;
              triptr->uvw[i][2] = FX_global->vel[kz].z;

              /* get the velocity gradient tensor */
              offset = 9*kz;
              for (row = 0; row < 3; row++)
                for (col = 0; col < 3; col++)
                  triptr->tnsr[i][row][col] = vel_grad[offset++];
	    }

            /* transform the coordinates and tensor into a new basis */
            Change_Basis(triptr);
            index++;
          }
      }
      *ntri = index;
      *triangle = triList;
      return 0;
}



static int
calculate_intercepts(double x_int[2], double y_int[2], double x[4], double y[4])
/*
 * 
 *  Determine if the y=0 axis crosses the triangle
 * 
 */
{
    int    i, num_points;
    double denom, frac;

    num_points = 0;
    for (i = 0; i < 3; i++) {
        if ( ((x[i]<0.0) && (x[i+1]>0.0)) || ((x[i]>0.0) && (x[i+1]<0.0)) ) {
	  x_int[num_points] = 0.0;
	  denom = x[i+1]-x[i];
	  frac  = -x[i]/denom;
	  if (fabs(denom) < TOL) {
	      y_int[num_points] = 0.0;
	  } else {
	      y_int[num_points] = y[i] + frac*(y[i+1]-y[i]);
          }
	  num_points++;
        }
    }

    if (num_points >= 2) return TRUE;
    return FALSE;
}



static void
phase_plane_analysis(int classification, double x[3], double y[3], double b[2],
                     double c[2][2], double c_inv[2][2], facet *triptr, 
                     int *sflag, float sxyz[6], int *aflag, float axyz[6])
/*
 * 
 *  Test for separation and attachment lines in the phase plane
 * 
 */
{
    double basis[2]; 
    double x_int[2], y_int[2];
    double x_can[4], y_can[4];

    *sflag = FALSE;
    *aflag = FALSE;
/*
 *  Convert the 2D physical coordinates into canonical coordinates
 */
    convert_to_canonical (b, c_inv, x[0], y[0], &x_can[0], &y_can[0]);    
    convert_to_canonical (b, c_inv, x[1], y[1], &x_can[1], &y_can[1]);
    convert_to_canonical (b, c_inv, x[2], y[2], &x_can[2], &y_can[2]);
    convert_to_canonical (b, c_inv, x[0], y[0], &x_can[3], &y_can[3]);
/*  
 *  Test for an attachment line (y=0 in the phase plane)
 */
    if ( (classification==REPELLING_NODE) ||  (classification==SADDLE)) {
       if ( calculate_intercepts(x_int, y_int, x_can, y_can) == TRUE ) {
          convert_to_2D_basis(b, c, x_int[0], y_int[0], &basis[0], &basis[1]);
          convert_to_physical(triptr, &basis[0], &sxyz[0]);

	  convert_to_2D_basis(b, c, x_int[1], y_int[1], &basis[0], &basis[1]);
          convert_to_physical(triptr, &basis[0], &sxyz[3]);

          *sflag = TRUE;
       }
    }
/*  
 *  Test for a separation line (x=0 in the phase plane)
 */
    if ( (classification==ATTRACTING_NODE) ||  (classification==SADDLE)) {
       if ( calculate_intercepts(y_int, x_int, y_can, x_can) == TRUE ) {
          convert_to_2D_basis(b, c, x_int[0], y_int[0], &basis[0], &basis[1]);
          convert_to_physical(triptr, &basis[0], &axyz[0]);
	      
	  convert_to_2D_basis(b, c, x_int[1], y_int[1], &basis[0], &basis[1]);
          convert_to_physical(triptr, &basis[0], &axyz[3]);

	  *aflag = TRUE;
       }
    }
}



static void
locate_separation_lines(double x[3], double y[3], double a1, double b1,
                        double c1, double a2, double b2, double c2, 
                        int root_type, int classification, complex *x1, 
                        complex *y1, complex *x2, complex *y2, facet *triptr,
                        int *sflag, float sxyz[6], int *aflag, float axyz[6])
/*
 * 
 *  Construct the eigenmatrix and perform an eigensystem analysis
 * 
 */
{
    double vel_determinant;
    double b[2], c[2][2], c_inv[2][2];
/*  
 *  Calculate the determinant of the velocity tensor
 */   
    vel_determinant = b1*c2-b2*c1;
    if (fabs(vel_determinant) > TOL) {
       b[0] = (a2*c1-a1*c2)/vel_determinant;
       b[1] = (a1*b2-a2*b1)/vel_determinant;
    } else {
       return;
    }
/*
 *  Create the eigenmatrix 
 */
    if (root_type == TWO_REAL_ROOTS) {
       c[0][0] = x1->re;
       c[0][1] = x2->re;
       c[1][0] = y1->re;
       c[1][1] = y2->re;
    }
/*  
 *  Invert the eigenvector matrix
 */
    if (matrix_inv_2(c, c_inv) == 0) return;
/*
 *  locate separation and attachment lines using phase plane analysis
 */
    phase_plane_analysis(classification, x, y, b, c, c_inv, triptr, sflag, sxyz,
                         aflag, axyz);
}




static int
allocate_line_storage(int *maxSep, float **sepcoor,
                      int *maxAtt, float **attcoor)
/*
 * 
 *  Allocate storage for the coordinates of the separation and attachment lines
 * 
 */
{
    int   nbc, maxFacets, estNumPts;
    float *sepLines, *attLines;
/*
 *  Check that the domain surface variables have been defined
 */
    nbc = FX_global->nbc;
    if (nbc <= 0) {
      printf(" FX Error: No domain surfaces are defined (FX_SepnLine)!\n");
      return 1;
    }
    maxFacets = 2*FX_global->bc[nbc-1].last_facet;
    if (maxFacets <= 0) {
      printf(" FX Error: last_facet has not been assigned for domain surfaces (FX_SepnLine)!\n");
      return 1;
    }
/*
 *  Estimate the number of coordintes needed to store the separation lines
 */      
    estNumPts = 0.1*maxFacets;
    if (estNumPts < 2) estNumPts = maxFacets;
/*
 *  Allocate the arrays
 */      
    sepLines  = (float *) malloc(3*estNumPts*sizeof(float));
    attLines  = (float *) malloc(3*estNumPts*sizeof(float));
/*
 *  Verify that the pointers are valid
 */      
    if ((sepLines == NULL) || (attLines  == NULL)) 
    {
      printf(" FX Error: MALLOC for Point storage (FX_SepnLine)!\n");
      if (sepLines != NULL) free(sepLines);
      if (attLines != NULL) free(attLines);
      return 1;
    }
/*
 *  Reassign pointers
 */      
    *maxSep  = estNumPts;
    *maxAtt  = estNumPts;
    *sepcoor = sepLines;
    *attcoor = attLines;
    return 0;
}



static void
extract_separation_lines(float *vel_grad, float **sepxyz, int *sepblk,
                         float **attxyz, int *attblk)
/*
 * 
 *  Extract separation and attachment lines on all user defined surfaces
 * 
 */
{
    complex lambda_1, lambda_2, x1, x2, y1, y2;
    double a1, b1, c1, a2, b2, c2, dudx, dudy, dvdx, dvdy;
    double determinant;
    double jac[3][3], inv_jac[3][3];
    double u[3], v[3], x[3], y[3], a_vec[3], b_vec[3];
    int classification = UNKNOWN;
    int root_type      = UNKNOWN;
    int i, ib, maxSep, maxAtt;
    int nsxyz = 0;
    int naxyz = 0;
    int ntri  = 0;
    int sflag = 0;
    int aflag = 0;
    int index = 0;
    float sxyz[6], axyz[6], *tmps, *tmpa, *sepcoor, *attcoor;
    facet *triptr, *triangle;

/* 
 *  allocate storage for the line segments
 */
    if (allocate_line_storage(&maxSep, &sepcoor, &maxAtt, &attcoor) != 0)
       goto err;
/*
 *  Iterate over all surface blocks
 */
    for (ib = 0; ib < FX_global->nbc; ib++) {
/*
 *     Create a list of triangles and transform all vectors and tensors
 *     into a 2D orthonormal coordinate system 
 */
       if (Create_Triangle_List(ib, &triangle, &ntri, vel_grad) != 0) goto err;
/*
 *     Loop over every triangle and test for separation and attachment lines
 */    
       for( index = 0; index < ntri; index++) {
          triptr = &triangle[index];
/*
 *	  Load the 2-D coordinates and velocities for one triangle
 */
          for (i=0; i<3; i++)
          {
            x[i] = triptr->xyzRot[i][0];
            y[i] = triptr->xyzRot[i][1];
            u[i] = triptr->uvwRot[i][0];
            v[i] = triptr->uvwRot[i][1];
	  }
/*
 *	  Calculate the Jacobian and inverse Jacobian matrices 
 */
	  jac[0][0] = 1.0;   jac[0][1] = x[0];   jac[0][2] = y[0];
	  jac[1][0] = 1.0;   jac[1][1] = x[1];   jac[1][2] = y[1];
	  jac[2][0] = 1.0;   jac[2][1] = x[2];   jac[2][2] = y[2];

	  matrix_inv_3 (jac, inv_jac);
/*
 *	  Evaluate the coefficients of the linear 
 */
	  matrix_vec_3(inv_jac, u, a_vec);
	  matrix_vec_3(inv_jac, v, b_vec);
	      
	  a1        = a_vec[0];
	  b1 = dudx = a_vec[1];
	  c1 = dudy = a_vec[2];
	      
	  a2        = b_vec[0];
          b2 = dvdx = b_vec[1];
          c2 = dvdy = b_vec[2];
/*
 *	  Check that the velocity gradient tensor is not degenerate  
 */
          determinant = dudx*dvdy - dvdx*dudy;

          if (fabs(determinant) < TOL) continue;
/*
 *	  Compute the eigenvalues of the velocity gradient tensor
 */
	  classification = eigenvalues2x2(b1, c1, b2, c2, &root_type,
                                          &lambda_1, &lambda_2);
/*
 *        Compute the eigenvectors if the tensor has two real eigenvalues
 */      
	  if (root_type == TWO_REAL_ROOTS) {
	     real_eigenvectors2x2(b1, c1, b2, c2, &lambda_1, &lambda_2,
                                  &x1, &y1, &x2, &y2);
/*
 *           Test for the presence of separation and attachment lines
 */
	     locate_separation_lines(x, y, a1, b1, c1, a2, b2, c2, root_type,
                                     classification, &x1, &y1, &x2, &y2, 
                                     triptr, &sflag, sxyz, &aflag, axyz);
/*
 *           Add a separation line segment to the list
 */
	     if (sflag == TRUE) {
	        if (nsxyz >= (maxSep-6) ) {
                    maxSep *= 1.5;
                    tmps = (float *) 
                           realloc( (void *) sepcoor, 3*maxSep*sizeof(float));
                    if (tmps == NULL) goto err;
		    sepcoor = tmps;
		}
                for (i = 0; i < 6; i++) sepcoor[nsxyz*3+i] = sxyz[i];
       	        nsxyz += 2;
	     }
/*
 *           Add an attachment line segment to the list
 */
	     if (aflag == TRUE) {
 	        if (naxyz >= (maxAtt-6) ) {
                    maxAtt *= 1.5;
                    tmpa = (float *) 
                           realloc( (void *) attcoor, 3*maxAtt*sizeof(float));
                    if (tmpa == NULL) goto err;
		    attcoor = tmpa;
		}
                for (i = 0; i < 6; i++) attcoor[naxyz*3+i] = axyz[i];
       		naxyz += 2;
	     }
       	  }
       }
       sepblk[ib] = nsxyz;
       attblk[ib] = naxyz;
       free(triangle);
    }
/*
 *  reassign pointers 
 */
    *sepxyz = sepcoor;
    *attxyz = attcoor;
    return;

err:
    free(sepcoor);
    free(attcoor);
    for (i = 0; i < FX_global->nbc; i++) sepblk[i] = attblk[i] = 0;
}
