/*
 * FX_ShockSubs.c
 *
 * FX module that computes the Shock functions.
 *
 * Copyright 1999 - 2000, Massachusetts Institute of Technology.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "FX_types.h"


#ifdef WIN32

#define PREFIX __stdcall

#else

#define PREFIX

#ifdef UNDERSCORE

#define FXSCAL          fxscal_
#define FXVEL           fxvel_
#define FXVELPTR        fxvelptr_
#define FXVELP3D        fxvelp3d_

#define FX_SHOCKFIND    fx_shockfind_
#define FX_SHOCKSURFACE fx_shocksurface_
#define FX_SHOCKVOLUMES fx_shockvolumes_
#define FX_SHOCKANGLES  fx_shockangles_

#else

#define FXSCAL          fxscal
#define FXVEL           fxvel
#define FXVELPTR        fxvelptr
#define FXVELP3D        fxvelp3d

#define FX_SHOCKFIND    fx_shockfind
#define FX_SHOCKSURFACE fx_shocksurface
#define FX_SHOCKVOLUMES fx_shockvolumes
#define FX_SHOCKANGLES  fx_shockangles

#endif
#endif


/* globals */

  extern FXglobal *FX_global;


/* externals */

  extern void PREFIX FXSCAL(int *, float *,  float *);
  extern void PREFIX FXVEL(FXcoor *,  FXcoor *);
  extern void PREFIX FXVELPTR(FXcoor **,  FXcoor *);
  extern void PREFIX FXVELP3D(float **, float **, float **, FXcoor *);

  extern void        FX_Grad(FXglobal *, float *, float *, float **);
  extern void        FX_Div (FXglobal *, float *, float *, float **);
  extern void        FX_GenSurf(FXglobal *globals, float val, float *scalar,
                                int *npts, FXcoor **xyz, 
                                int *ntris, int **tris, int **cell);
  extern void        FX_OrderCutVol(FXglobal *globals, float val, int abel,
                                    float *scal, int *nclust, int **clust, 
                                    int **cells, int **cn, int **cnds);
  extern void        FX_ICutVol(FXglobal *globals, float val, float *scalar,
                                int ncell, int *cells, float *above, 
                                float *below);



void PREFIX
FX_SHOCKFIND(float **test)
{
  int    i;
  float  dist, mag, coef, *scal, *grad, *sum, *dens, *hdens, *tmp;
  float  *vec, *hvec;

  *test = NULL;

  if (FX_global == NULL) {
    printf(" FX Error: NOT Initialized (FX_ShockFind)!\n");
    return;
  }

  if (FX_global->status != 0) {
    printf(" FX Error: Initialization Status = %d (FX_ShockFind)!\n",
           FX_global->status);
    return;
  }

  scal = (float *) malloc(FX_global->total_nodes*sizeof(float));
  if (scal == NULL) {
    printf(" FX Error: MALLOC for Volume storage (FX_ShockFind)!\n");
    return;
  }

  /* put pressure gradient in grad */

  i = 2;
  FXSCAL(&i, scal, FX_global->hscalar);
  FX_Grad(FX_global, scal, FX_global->hscalar, &grad);
  if (grad == NULL) {
    free(scal);
    return;
  }

  if (FX_global->vload == 0) {
    if ((FX_global->cmask & 8) == 0) {
      FXVEL    (FX_global->vel, FX_global->hvel);
    } else {
      if ((FX_global->cmask & 64) == 0) {
        FXVELPTR(&FX_global->vel, FX_global->hvel);
      } else {
        FXVELP3D(&FX_global->velx, &FX_global->vely, &FX_global->velz,
                  FX_global->hvel);
      }
    }
    FX_global->vload = 1;
  }

  if (FX_global->sload != 3) {
    FX_global->sload = 3;
    FXSCAL(&FX_global->sload, FX_global->scalar, FX_global->hscalar);
  }
  for (i = 0; i < FX_global->total_nodes; i++) {
    scal[i] = 0.0;
    dist = sqrt( grad[3*i  ]*grad[3*i  ] + grad[3*i+1]*grad[3*i+1] + 
                 grad[3*i+2]*grad[3*i+2] );
    if (dist == 0.0) continue;
    if (((FX_global->cmask & 8) == 0) || ((FX_global->cmask & 64) == 0)) {
      mag = sqrt( FX_global->vel[i].x*FX_global->vel[i].x +
                  FX_global->vel[i].y*FX_global->vel[i].y +
                  FX_global->vel[i].z*FX_global->vel[i].z );
      if (mag == 0.0) continue;
      mag = FX_global->scalar[i] / mag;
      scal[i] = mag * (FX_global->vel[i].x*grad[3*i  ] +
                       FX_global->vel[i].y*grad[3*i+1] +
                       FX_global->vel[i].z*grad[3*i+2]) / dist;
    } else {
      mag = sqrt( FX_global->velx[i]*FX_global->velx[i] +
                  FX_global->vely[i]*FX_global->vely[i] +
                  FX_global->velz[i]*FX_global->velz[i] );
      if (mag == 0.0) continue;
      mag = FX_global->scalar[i] / mag;
      scal[i] = mag * (FX_global->velx[i]*grad[3*i  ] +
                       FX_global->vely[i]*grad[3*i+1] +
                       FX_global->velz[i]*grad[3*i+2]) / dist;
    }
  }

  /* compute the transient test function */

  if (FX_global->iopt != 0) {
    vec = (float *) malloc(3*FX_global->total_nodes*sizeof(float));
    if (vec == NULL) {
      printf(" FX Error: MALLOC for Vector storage (FX_ShockFind)!\n");
      free(grad);
      free(scal);
      return;
    }
    hvec = NULL;
    if (FX_global->nhalo != 0) {
      hvec = (float *) malloc(3*FX_global->nhalo*sizeof(float));
      if (hvec == NULL) {
        printf(" FX Error: MALLOC for hVector storage (FX_ShockFind)!\n");
        free(vec);
        free(grad);
        free(scal);
        return;
      }
    }
    dens = (float *) malloc(FX_global->total_nodes*sizeof(float));
    if (dens == NULL) {
      printf(" FX Error: MALLOC for Density storage (FX_ShockFind)!\n");
      if (FX_global->nhalo != 0) free(hvec);
      free(vec);
      free(grad);
      free(scal);
      return;
    }
    hdens = NULL;
    if (FX_global->nhalo != 0) {
      hdens = (float *) malloc(FX_global->nhalo*sizeof(float));
      if (hdens == NULL) {
        printf(" FX Error: MALLOC for hDensity storage (FX_ShockFind)!\n");
        free(dens);
        if (FX_global->nhalo != 0) free(hvec);
        free(vec);
        free(grad);
        free(scal);
        return;
      }
    }
    FX_global->sload = 2;
    FXSCAL(&FX_global->sload, FX_global->scalar, FX_global->hscalar);

/* anisentropic correction -----
    i = 1;
    FXSCAL(&i, dens, hdens);
    if (((FX_global->cmask & 8) == 0) || ((FX_global->cmask & 64) == 0)) {
      for (i = 0; i < FX_global->total_nodes; i++) {
        vec[3*i  ] = dens[i]*FX_global->vel[i].x;
        vec[3*i+1] = dens[i]*FX_global->vel[i].y;
        vec[3*i+2] = dens[i]*FX_global->vel[i].z;
      }
    } else {
      for (i = 0; i < FX_global->total_nodes; i++) {
        vec[3*i  ] = dens[i]*FX_global->velx[i];
        vec[3*i+1] = dens[i]*FX_global->vely[i];
        vec[3*i+2] = dens[i]*FX_global->velz[i];
      }
    }
    for (i = 0; i < FX_global->nhalo; i++) {
      hvec[3*i  ] = hdens[i]*FX_global->hvel[i].x;
      hvec[3*i+1] = hdens[i]*FX_global->hvel[i].y;
      hvec[3*i+2] = hdens[i]*FX_global->hvel[i].z;
    }
    FX_Div(FX_global, vec, hvec, &sum);
    if (sum == NULL) {
      if (FX_global->nhalo != 0) free(hdens);
      free(dens);
      if (FX_global->nhalo != 0) free(hvec);
      free(vec);
      free(grad);
      free(scal);
      return;
    }

    for (i = 0; i < FX_global->total_nodes; i++) {
      dist = sqrt( grad[3*i  ]*grad[3*i  ] + grad[3*i+1]*grad[3*i+1] + 
                   grad[3*i+2]*grad[3*i+2] );
      if (dist == 0.0) continue;
      coef = sqrt(FX_global->gamma*FX_global->scalar[i]/dens[i]);
      scal[i] -= coef*sum[i]/dist;
    }
  ----- end of anisentropic correction */

    /* full correction */

    i = 5;
    FXSCAL(&i, dens, hdens);
    if (((FX_global->cmask & 8) == 0) || ((FX_global->cmask & 64) == 0)) {
      for (i = 0; i < FX_global->total_nodes; i++) {
        vec[3*i  ] = dens[i]*FX_global->vel[i].x;
        vec[3*i+1] = dens[i]*FX_global->vel[i].y;
        vec[3*i+2] = dens[i]*FX_global->vel[i].z;
      }
    } else {
      for (i = 0; i < FX_global->total_nodes; i++) {
        vec[3*i  ] = dens[i]*FX_global->velx[i];
        vec[3*i+1] = dens[i]*FX_global->vely[i];
        vec[3*i+2] = dens[i]*FX_global->velz[i];
      }
    }
    for (i = 0; i < FX_global->nhalo; i++) {
      hvec[3*i  ] = hdens[i]*FX_global->hvel[i].x;
      hvec[3*i+1] = hdens[i]*FX_global->hvel[i].y;
      hvec[3*i+2] = hdens[i]*FX_global->hvel[i].z;
    }
    i = 1;
    FXSCAL(&i, dens, hdens);
    for (i = 0; i < FX_global->total_nodes; i++) {
      vec[3*i  ] *= dens[i];
      vec[3*i+1] *= dens[i];
      vec[3*i+2] *= dens[i];
    }
    for (i = 0; i < FX_global->nhalo; i++) {
      hvec[3*i  ] *= hdens[i];
      hvec[3*i+1] *= hdens[i];
      hvec[3*i+2] *= hdens[i];
    }
    FX_Div(FX_global, vec, hvec, &sum);
    if (sum == NULL) {
      if (FX_global->nhalo != 0) free(hdens);
      free(dens);
      if (FX_global->nhalo != 0) free(hvec);
      free(vec);
      free(grad);
      free(scal);
      return;
    }

    FX_Grad(FX_global, dens, hdens, &tmp);
    if (tmp == NULL) {
      free(sum);
      if (FX_global->nhalo != 0) free(hdens);
      free(dens);
      if (FX_global->nhalo != 0) free(hvec);
      free(vec);
      free(grad);
      free(scal);
      return;
    }
    if (((FX_global->cmask & 8) == 0) || ((FX_global->cmask & 64) == 0)) {
      for (i = 0; i < FX_global->total_nodes; i++) 
        sum[i] = -sum[i] +
                 FX_global->vel[i].x*(grad[3*i  ] +
                      tmp[3*i  ]*FX_global->vel[i].x*FX_global->vel[i].x) +
                 FX_global->vel[i].y*(grad[3*i+1] +
                      tmp[3*i+1]*FX_global->vel[i].y*FX_global->vel[i].y) +
                 FX_global->vel[i].z*(grad[3*i+2] +
                      tmp[3*i+2]*FX_global->vel[i].z*FX_global->vel[i].z);
    } else {
      for (i = 0; i < FX_global->total_nodes; i++)
        sum[i] = -sum[i] +
                 FX_global->velx[i]*(grad[3*i  ] +
                      tmp[3*i  ]*FX_global->velx[i]*FX_global->velx[i]) +
                 FX_global->vely[i]*(grad[3*i+1] +
                      tmp[3*i+1]*FX_global->vely[i]*FX_global->vely[i]) +
                 FX_global->velz[i]*(grad[3*i+2] +
                      tmp[3*i+2]*FX_global->velz[i]*FX_global->velz[i]);
    }
    free(tmp);

    if (((FX_global->cmask & 8) == 0) || ((FX_global->cmask & 64) == 0)) {
      for (i = 0; i < FX_global->total_nodes; i++) {
        vec[3*i  ] = dens[i]*FX_global->vel[i].x;
        vec[3*i+1] = dens[i]*FX_global->vel[i].y;
        vec[3*i+2] = dens[i]*FX_global->vel[i].z;
      }
    } else {
      for (i = 0; i < FX_global->total_nodes; i++) {
        vec[3*i  ] = dens[i]*FX_global->velx[i];
        vec[3*i+1] = dens[i]*FX_global->vely[i];
        vec[3*i+2] = dens[i]*FX_global->velz[i];
      }
    }
    for (i = 0; i < FX_global->nhalo; i++) {
      hvec[3*i  ] = hdens[i]*FX_global->hvel[i].x;
      hvec[3*i+1] = hdens[i]*FX_global->hvel[i].y;
      hvec[3*i+2] = hdens[i]*FX_global->hvel[i].z;
    }
    FX_Div(FX_global, vec, hvec, &tmp);
    if (tmp == NULL) {
      free(sum);
      if (FX_global->nhalo != 0) free(hdens);
      free(dens);
      if (FX_global->nhalo != 0) free(hvec);
      free(vec);
      free(grad);
      free(scal);
      return;
    }
    if (((FX_global->cmask & 8) == 0) || ((FX_global->cmask & 64) == 0)) {
      for (i = 0; i < FX_global->total_nodes; i++)
        sum[i] -= 0.5*( FX_global->vel[i].x*FX_global->vel[i].x +
                        FX_global->vel[i].y*FX_global->vel[i].y +
                        FX_global->vel[i].z*FX_global->vel[i].z )*tmp[i];
    } else {
      for (i = 0; i < FX_global->total_nodes; i++)
        sum[i] -= 0.5*( FX_global->velx[i]*FX_global->velx[i] +
                        FX_global->vely[i]*FX_global->vely[i] +
                        FX_global->velz[i]*FX_global->velz[i] )*tmp[i];
    }
    free(tmp);

    for (i = 0; i < FX_global->total_nodes; i++) {
      dist = sqrt( grad[3*i  ]*grad[3*i  ] + grad[3*i+1]*grad[3*i+1] +
                   grad[3*i+2]*grad[3*i+2] );
      if (dist == 0.0) continue;
/*
      coef = (FX_global->gamma - 1.0)*
              sqrt(FX_global->gamma*FX_global->scalar[i]/dens[i]);
*/
      coef = (FX_global->gamma - 1.0) / (FX_global->gamma *
              sqrt(FX_global->gamma*FX_global->scalar[i]/dens[i]));
      scal[i] -= coef*sum[i]/dist;
    }
    
    free(sum);
    if (FX_global->nhalo != 0) free(hdens);
    free(dens);
    if (FX_global->nhalo != 0) free(hvec);
    free(vec);
  }

  free(grad);
  *test = scal;
}


void PREFIX
FX_SHOCKANGLES(float *test, float **angles)
{
  int    i;
  float  *scal, *grad;
  double dist;

  *angles = NULL;

  if (FX_global == NULL) {
    printf(" FX Error: NOT Initialized (FX_ShockAngles)!\n");
    return;
  }

  if (FX_global->status != 0) {
    printf(" FX Error: Initialization Status = %d (FX_ShockAngles)!\n",
           FX_global->status);
    return;
  }

  scal = (float *) malloc(FX_global->total_nodes*sizeof(float));
  if (scal == NULL) {
    printf(" FX Error: MALLOC for Temp storage (FX_ShockAngles)!\n");
    return;
  }

  /* put pressure gradient in grad */

  i = 2;
  FXSCAL(&i, scal, FX_global->hscalar);
  FX_Grad(FX_global, scal, FX_global->hscalar, &grad);
  if (grad == NULL) {
    free(scal);
    return;
  }

  if (FX_global->vload == 0) {
    if ((FX_global->cmask & 8) == 0) {
      FXVEL    (FX_global->vel, FX_global->hvel);
    } else {
      if ((FX_global->cmask & 64) == 0) {
        FXVELPTR(&FX_global->vel, FX_global->hvel);
      } else {
        FXVELP3D(&FX_global->velx, &FX_global->vely, &FX_global->velz,
                  FX_global->hvel);
      }
    }
    FX_global->vload = 1;
  }

  for (i = 0; i < FX_global->total_nodes; i++) {
    scal[i] = -2.0;
    if (test[i] >= 1.0) {
      scal[i] = 0.0;
      dist = sqrt( grad[3*i  ]*grad[3*i  ] + grad[3*i+1]*grad[3*i+1] +
                   grad[3*i+2]*grad[3*i+2] );
      if (dist == 0.0) continue;
      if (((FX_global->cmask & 8) == 0) || ((FX_global->cmask & 64) == 0)) {
        dist *= sqrt( FX_global->vel[i].x*FX_global->vel[i].x +
                      FX_global->vel[i].y*FX_global->vel[i].y +
                      FX_global->vel[i].z*FX_global->vel[i].z );
        if (dist == 0.0) continue;
        scal[i] = ( grad[3*i  ]*FX_global->vel[i].x +
                    grad[3*i+1]*FX_global->vel[i].y +
                    grad[3*i+2]*FX_global->vel[i].z ) / dist;
      } else {
        dist *= sqrt( FX_global->velx[i]*FX_global->velx[i] +
                      FX_global->vely[i]*FX_global->vely[i] +
                      FX_global->velz[i]*FX_global->velz[i] );
        if (dist == 0.0) continue;
        scal[i] = ( grad[3*i  ]*FX_global->velx[i] +
                    grad[3*i+1]*FX_global->vely[i] +
                    grad[3*i+2]*FX_global->velz[i] ) / dist;
      }
    }
  }

  free(grad);
  *angles = scal;
}


void PREFIX
FX_SHOCKSURFACE(float *test, int *npts, FXcoor **xyz,
                int *ntris, int **tris, int **cell)
{
  FX_GenSurf(FX_global, 1.0, test, npts, xyz, ntris, tris, cell);
}


void PREFIX
FX_SHOCKVOLUMES(float *test, int *nvol, float **vols, int **pcells, int **cells,
                int **cnode, int**cnodes)
{
  int   i, j, len, *cell, *clust;
  float *vol, below, above;
  
  *vols = NULL;
  FX_OrderCutVol(FX_global, 1.0, 1, test, nvol, pcells, cells, cnode, cnodes);
  if (*nvol == 0) return;

  vol = (float *) malloc(*nvol*sizeof(float));
  if (vol == NULL) {
    printf(" FX Error: MALLOC for Volume storage (FX_ShockVolumes)!\n");
    return;
  }

  cell  = *cells;
  clust = *pcells;
  for (j = i = 0; i < *nvol; i++) {
    len = clust[i] - j;
    FX_ICutVol(FX_global, 1.0, test, len, &cell[j], &above, &below);
    vol[i] = above;
    j = clust[i] + 1;
  }

  *vols = vol;
}
