/*
 * FX_VolSubs.c
 *
 * FX modules that calculate Volumes.
 *
 * Copyright 1999 - 2000, Massachusetts Institute of Technology.
 *
 */

#include <stdio.h>
#include <stdlib.h>


#include "FX_types.h"


#ifdef WIN32

#define PREFIX __stdcall

#else

#define PREFIX

#ifdef UNDERSCORE

#define FXCELLGET fxcellget_

#else

#define FXCELLGET fxcellget

#endif
#endif

  extern void PREFIX FXCELLGET (int *, int *, int *);


#define TREECHUNK 500


typedef struct CellList {
  int kcell;			/* cell number */
  struct CellList *next;	/* next entry */
} CellList;


typedef struct {
  int      root;		/* root of branch */
  CellList *list;		/* cell list for the link */
  CellList *last;		/* last entry in list */
} Tree;


typedef struct {
  int      mtree;		/* tree size */
  int      ntree;		/* current tree length */
  int      ltree;		/* last tree for cell */
  Tree     *tree;		/* the tree */
  int      *nodeTable;		/* current node allocation */
  CellList *cellTable;		/* the complete cell list */
} TheTree;


/*
 *  Given the cell number return the nodes
 */

void
FX_Cell2Nodes(FXglobal *globals, int kc, int *nodes, int *node)
{
  int kcr, i, j, k, ib, ni, nj, kns, kcs, kz, flg, nlo, nhi;

  if ((kc > globals->total_dis) && (kc <= globals->total_cells)) {

    /* structured blocks */
    *nodes = 8;
    kcs = globals->total_dis;
    kns = globals->nnode;
    /* find the block */
    ib  = 1;
    nlo = flg = 0;
    nhi = globals->nblock+1;
    if (globals->nblock > 1) flg = 1;
    while (flg != 0) {
      ib = (nlo+nhi)/2;
      if (kc > globals->block[ib-1].last_cell) {
        nlo = ib;
        if (nhi-nlo <= 1) {
          flg = 0;
          ib++;
        }
      } else {
        nhi = ib;
        if (nhi-nlo <= 1) flg = 0;
      }
    }
    ib--;

    if (ib > 0) {
      kcs = globals->block[ib-1].last_cell;
      kns = globals->block[ib-1].last_node;
    }
    ni  = globals->block[ib].ni;
    nj  = globals->block[ib].nj;
    kcr = kc - kcs - 1;
    i   =  kcr%(ni-1);
    j   = (kcr/(ni-1))%(nj-1);
    k   =  kcr/((ni-1)*(nj-1));
    kz  =  kns + i + j*ni + k*ni*nj + 1;
    node[0] = kz;
    node[1] = kz + 1;
    node[2] = kz + 1 + ni;
    node[3] = kz     + ni;
    node[4] = kz          + ni*nj;
    node[5] = kz + 1      + ni*nj;
    node[6] = kz + 1 + ni + ni*nj;
    node[7] = kz     + ni + ni*nj;
    return;
  }

  if ((globals->cmask & 128) != 0) {
    kcr = kc;
    if (kc > globals->total_cells) kcr = -(kc-globals->total_cells);
    FXCELLGET(&kcr, &k, node);
                *nodes = 4;
    if (k == 2) *nodes = 5;
    if (k == 3) *nodes = 6;
    if (k == 4) *nodes = 8;
    return;
  }

  if (kc <= globals->ntets) {

    /* tetrahedra */
    *nodes = 4;
    for (i = 0; i < 4; i++) node[i] = globals->tets[kc-1].nodes[i];

  } else if (kc <= globals->ntets + globals->npyra) {

    /* pyramids */
    *nodes = 5;
    kcr = kc - globals->ntets - 1;
    for (i = 0; i < 5; i++) node[i] = globals->pyra[kcr].nodes[i];

  } else if (kc <= globals->ntets + globals->npyra + globals->nprism) {

    /* prisms */
    *nodes = 6;
    kcr = kc - globals->ntets - globals->npyra - 1;
    for (i = 0; i < 6; i++) node[i] = globals->prism[kcr].nodes[i];

  } else if (kc <= globals->total_dis) {

    /* hexahedra */
    *nodes = 8;
    kcr = kc - globals->ntets - globals->npyra - globals->nprism - 1;
    for (i = 0; i < 8; i++) node[i] = globals->hexa[kcr].nodes[i];

  } else {

    /* halo cells */
    kcr = kc - globals->total_cells - 1;
                                       *nodes = 4;
    if (globals->hcell[kcr].type == 2) *nodes = 5;
    if (globals->hcell[kcr].type == 3) *nodes = 6;
    if (globals->hcell[kcr].type == 4) *nodes = 8;
    for (i = 0; i < *nodes; i++) node[i] = globals->hcell[kcr].nodes[i];

  }
}


/*
 *  Computes and returns the volume of the tetrahedron
 */

static double
FX_TetVol(FXcoor coor1, FXcoor coor2, FXcoor coor3, FXcoor coor4)
{
  double x1, y1, z1, x2, y2, z2, x3, y3, z3, x4, y4, z4, V;

  x1 = coor1.x;
  y1 = coor1.y;
  z1 = coor1.z;
  x2 = coor2.x;
  y2 = coor2.y;
  z2 = coor2.z;
  x3 = coor3.x;
  y3 = coor3.y;
  z3 = coor3.z;
  x4 = coor4.x;
  y4 = coor4.y;
  z4 = coor4.z;

  V = (x2-x1)*( (y3-y1)*(z4-z1)-(z3-z1)*(y4-y1) ) -
      (x3-x1)*( (y2-y1)*(z4-z1)-(z2-z1)*(y4-y1) ) +
      (x4-x1)*( (y2-y1)*(z3-z1)-(z2-z1)*(y3-y1) );
  if (V < 0.0) V = -V;

  return V/6.0;
}


/*
 *  Calaculates and returns the volume below and above the cut value
 */

static void
FX_CutVolTet(float cut, float *cvertex, FXcoor *coors,
             double *above, double *below)
{
  int    ninter, i, n1, n2, index;
  float  d;
  double vol, sliced;
  FXcoor pts[4];
  static int  slice[16][2] = { 0, 0,  1, 0,  2, 0,  2, 1,  3, 0,  4, 2,  3, 2,
                              -4, 0,  4, 0,  4, 1,  3, 1, -3, 0,  1, 2, -2, 0,
                              -1, 0,  0, 0 };
  static char npts[16]   = { 0, 3, 3, 4, 3, 4, 4, 3, 3, 4, 4, 3, 4, 3, 3, 0 };
  static char ecut[16][4][2] = {
                            0, 0, 0, 0, 0, 0, 0, 0,  1, 2, 1, 4, 1, 3, 0, 0,
                            2, 1, 2, 3, 2, 4, 0, 0,  1, 3, 2, 3, 2, 4, 1, 4,
                            3, 2, 3, 1, 3, 4, 0, 0,  1, 2, 1, 4, 3, 4, 3, 2,
                            2, 1, 3, 1, 3, 4, 2, 4,  3, 4, 2, 4, 1, 4, 0, 0,
                            4, 3, 4, 1, 4, 2, 0, 0,  1, 2, 4, 2, 4, 3, 1, 3,
                            2, 1, 2, 3, 4, 3, 4, 1,  1, 3, 2, 3, 4, 3, 0, 0,
                            3, 2, 3, 1, 4, 1, 4, 2,  1, 2, 4, 2, 3, 2, 0, 0,
                            2, 1, 3, 1, 4, 1, 0, 0,  0, 0, 0, 0, 0, 0, 0, 0 };

  vol = FX_TetVol(coors[0], coors[1], coors[2], coors[3]);

  /* get the cut topology */

                         index  = 0;
  if (cvertex[0] >= cut) index += 1;
  if (cvertex[1] >= cut) index += 2;
  if (cvertex[2] >= cut) index += 4;
  if (cvertex[3] >= cut) index += 8;

  *above = *below = 0.0;
  if (index == 0) {
    *below = vol;
    return;
  }
  if (index == 15) {
    *above = vol;
    return;
  }

  /* make the edge intersections */

  ninter = npts[index];
  for (i = 0; i < ninter; i++) {
    n1 = ecut[index][i][0]-1;
    n2 = ecut[index][i][1]-1;
    d  = (cut-cvertex[n1])/(cvertex[n2]-cvertex[n1]);
    pts[i].x = coors[n1].x + d*(coors[n2].x-coors[n1].x);
    pts[i].y = coors[n1].y + d*(coors[n2].y-coors[n1].y);
    pts[i].z = coors[n1].z + d*(coors[n2].z-coors[n1].z);
  }

  /* slice the volume */

  if (ninter == 3) {
    n1 = slice[index][0];
    if (n1 < 0) n1 = -n1;
    sliced = FX_TetVol(coors[n1-1], pts[0], pts[1], pts[2]);
    if (slice[index][0] > 0) {
      *above = sliced;
      *below = vol - sliced;
    } else {
      *below = sliced;
      *above = vol - sliced;
    }

  } else {

    n1 = slice[index][0]-1;
    n2 = slice[index][1]-1;
    sliced = FX_TetVol(pts[0], pts[1], pts[3],    coors[n1]) +
             FX_TetVol(pts[1], pts[2], pts[3],    coors[n1]) +
             FX_TetVol(pts[0], pts[3], coors[n1], coors[n2]);
    if (cvertex[n1] >= cut) {
      *above = sliced;
      *below = vol - sliced;
    } else {
      *below = sliced;
      *above = vol - sliced;
    }
  }
}


/*
 *  calculates the volumes above and below an iso-surface value
 */

void
FX_CutVol(FXglobal *globals, float val, float *scalar,
          float *above, float *below)
{
  int    kc, i, j, k, ib, ni, nj, nk, nc, kns, kz, bnode, node[8], type;
  float  scal[4];
  FXcoor coor[4];
  double up, down, abov, belo;
  static int pyr2tet[2][4] = { 1,2,4,5, 2,3,4,5 };
  static int pri2tet[3][4] = { 1,2,4,5, 2,3,4,5, 1,4,5,6 };
  static int hex2tet[6][4] = { 2,4,5,6, 1,2,4,5, 2,3,4,6,
                               3,4,6,8, 4,5,6,8, 3,6,7,8 };
/*
  static int hex2tet[6][4] = { 2,4,5,6, 1,2,4,5, 2,3,6,8,
                               2,3,4,8, 4,5,6,8, 3,6,7,8 };
*/

  up = down = 0.0;

  /* structured blocks */

  if ((globals->nblock > 0) && (globals->iblank != NULL))
    bnode = globals->block[0].last_node - 
            globals->block[0].ni*globals->block[0].nj*globals->block[0].nk;
  for (ib = 0; ib < globals->nblock; ib++) {
    ni  = globals->block[ib].ni;
    nj  = globals->block[ib].nj;
    nk  = globals->block[ib].nk;
    nc  = (ni-1)*(nj-1)*(nk-1);
    kns = globals->block[ib].last_node - ni*nj*nk + 1;
    if (globals->iblank == NULL) {
      for (kc = 0; kc < nc; kc++) {
        i  =  kc%(ni-1);
        j  = (kc/(ni-1))%(nj-1);
        k  =  kc/((ni-1)*(nj-1));
        kz =  kns + i + j*ni + k*ni*nj;
        node[0] = kz;
        node[1] = kz + 1;
        node[2] = kz + 1 + ni;
        node[3] = kz     + ni;
        node[4] = kz          + ni*nj;
        node[5] = kz + 1      + ni*nj;
        node[6] = kz + 1 + ni + ni*nj;
        node[7] = kz     + ni + ni*nj;
        for (j = 0; j < 6; j++) {
          for (i = 0; i < 4; i++) {
            k  = hex2tet[j][i]-1;
            kz = node[k]-1;
            if ((globals->cmask & 64) == 0) {
              coor[i] = globals->coor[kz];
            } else {
              coor[i].x = globals->xcoor[kz];
              coor[i].y = globals->ycoor[kz];
              coor[i].z = globals->zcoor[kz];
            }
            scal[i] = scalar[kz];
           }
           FX_CutVolTet(val, scal, coor, &abov, &belo);
           up   += abov;
           down += belo;
        }
      }
    } else {
      for (kc = 0; kc < nc; kc++) {
        i  =  kc%(ni-1);
        j  = (kc/(ni-1))%(nj-1);
        k  =  kc/((ni-1)*(nj-1));
        kz =  kns + i + j*ni + k*ni*nj;
        node[0] = kz;
        node[1] = kz + 1;
        node[2] = kz + 1 + ni;
        node[3] = kz     + ni;
        node[4] = kz          + ni*nj;
        node[5] = kz + 1      + ni*nj;
        node[6] = kz + 1 + ni + ni*nj;
        node[7] = kz     + ni + ni*nj;
        for (j = i = 0; i < 8; i++)
          if (globals->iblank[node[i]-bnode-1] == 0) j++;
        if (j != 0) continue;

        for (j = 0; j < 6; j++) {
          for (i = 0; i < 4; i++) {
            k  = hex2tet[j][i]-1;
            kz = node[k]-1;
            if ((globals->cmask & 64) == 0) {
              coor[i] = globals->coor[kz];
            } else {
              coor[i].x = globals->xcoor[kz];
              coor[i].y = globals->ycoor[kz];
              coor[i].z = globals->zcoor[kz];
            }
            scal[i] = scalar[kz];
           }
           FX_CutVolTet(val, scal, coor, &abov, &belo);
           up   += abov;
           down += belo;
        }
      }
    }
  }

  if ((globals->cmask & 128) == 0) {

    /* tetrahedra */
    for (kc = 0; kc < globals->ntets; kc++) {
      for (i = 0; i < 4; i++) {
        kz = globals->tets[kc].nodes[i]-1;
        if ((globals->cmask & 64) == 0) {
          coor[i] = globals->coor[kz];
        } else {
          coor[i].x = globals->xcoor[kz];
          coor[i].y = globals->ycoor[kz];
          coor[i].z = globals->zcoor[kz];
        }
        scal[i] = scalar[kz];
      }
      FX_CutVolTet(val, scal, coor, &abov, &belo);
      up   += abov;
      down += belo;
    }

    /* pyramids */
    for (kc = 0; kc < globals->npyra; kc++)
      for (j = 0; j < 2; j++) {
        for (i = 0; i < 4; i++) {
          k  = pyr2tet[j][i]-1;
          kz = globals->pyra[kc].nodes[k]-1;
          if ((globals->cmask & 64) == 0) {
            coor[i] = globals->coor[kz];
          } else {
            coor[i].x = globals->xcoor[kz];
            coor[i].y = globals->ycoor[kz];
            coor[i].z = globals->zcoor[kz];
          }
          scal[i] = scalar[kz];
        }
        FX_CutVolTet(val, scal, coor, &abov, &belo);
        up   += abov;
        down += belo;
      }

    /* prisms */
    for (kc = 0; kc < globals->nprism; kc++)
      for (j = 0; j < 3; j++) {
        for (i = 0; i < 4; i++) {
          k  = pri2tet[j][i]-1;
          kz = globals->prism[kc].nodes[k]-1;
          if ((globals->cmask & 64) == 0) {
            coor[i] = globals->coor[kz];
          } else {
            coor[i].x = globals->xcoor[kz];
            coor[i].y = globals->ycoor[kz];
            coor[i].z = globals->zcoor[kz];
          }
          scal[i] = scalar[kz];
        }
        FX_CutVolTet(val, scal, coor, &abov, &belo);
        up   += abov;
        down += belo;
      }

    /* hexahedra */
    for (kc = 0; kc < globals->nhexa; kc++)
      for (j = 0; j < 6; j++) {
        for (i = 0; i < 4; i++) {
          k  = hex2tet[j][i]-1;
          kz = globals->hexa[kc].nodes[k]-1;
          if ((globals->cmask & 64) == 0) {
            coor[i] = globals->coor[kz];
          } else {
            coor[i].x = globals->xcoor[kz];
            coor[i].y = globals->ycoor[kz];
            coor[i].z = globals->zcoor[kz];
          }
          scal[i] = scalar[kz];
        }
        FX_CutVolTet(val, scal, coor, &abov, &belo);
        up   += abov;
        down += belo;
      }

  } else {

    /* construct elements on-the-fly */

    for (kc = 0; kc < globals->total_dis; kc++) {
      i = kc+1;
      FXCELLGET(&i, &type, node);

      switch (type) {

      case 1:
        /* tetrahedra */
        for (i = 0; i < 4; i++) {
          kz = node[i]-1;
          if ((globals->cmask & 64) == 0) {
            coor[i] = globals->coor[kz];
          } else {
            coor[i].x = globals->xcoor[kz];
            coor[i].y = globals->ycoor[kz];
            coor[i].z = globals->zcoor[kz];
          }
          scal[i] = scalar[kz];
        }
        FX_CutVolTet(val, scal, coor, &abov, &belo);
        up   += abov;
        down += belo;
        break;

      case 2:
        /* pyramids */
        for (j = 0; j < 2; j++) {
          for (i = 0; i < 4; i++) {
            k  = pyr2tet[j][i]-1;
            kz = node[k]-1;
            if ((globals->cmask & 64) == 0) {
              coor[i] = globals->coor[kz];
            } else {
              coor[i].x = globals->xcoor[kz];
              coor[i].y = globals->ycoor[kz];
              coor[i].z = globals->zcoor[kz];
            }
            scal[i] = scalar[kz];
          }
          FX_CutVolTet(val, scal, coor, &abov, &belo);
          up   += abov;
          down += belo;
        }
        break;

      case 3:
        /* prisms */
        for (j = 0; j < 3; j++) {
          for (i = 0; i < 4; i++) {
            k  = pri2tet[j][i]-1;
            kz = node[k]-1;
            if ((globals->cmask & 64) == 0) {
              coor[i] = globals->coor[kz];
            } else {
              coor[i].x = globals->xcoor[kz];
              coor[i].y = globals->ycoor[kz];
              coor[i].z = globals->zcoor[kz];
            }
            scal[i] = scalar[kz];
          }
          FX_CutVolTet(val, scal, coor, &abov, &belo);
          up   += abov;
          down += belo;
        }
        break;

      case 4:
        /* hexahedra */
        for (j = 0; j < 6; j++) {
          for (i = 0; i < 4; i++) {
            k  = hex2tet[j][i]-1;
            kz = node[k]-1;
            if ((globals->cmask & 64) == 0) {
              coor[i] = globals->coor[kz];
            } else {
              coor[i].x = globals->xcoor[kz];
              coor[i].y = globals->ycoor[kz];
              coor[i].z = globals->zcoor[kz];
            }
            scal[i] = scalar[kz];
          }
          FX_CutVolTet(val, scal, coor, &abov, &belo);
          up   += abov;
          down += belo;
        }
        break;

      default:
        printf(" FX Warning: FXcellGet INDEX = %d -> TYPE = %d (FX_CutVol)!\n",
               kc+1, type);
      }
    }
  }

  *above = up;
  *below = down;
}


/*
 *  calculates the volumes above and below an iso-surface value from a cell list
 */

void
FX_ICutVol(FXglobal *globals, float val, float *scalar, int ncell, int *cells,
           float *above, float *below)
{
  int    kx, i, j, k, kn, n, node[8];
  float  scal[4];
  FXcoor coor[4];
  double up, down, abov, belo;
  static int pyr2tet[2][4] = { 1,2,4,5, 2,3,4,5 };
  static int pri2tet[3][4] = { 1,2,4,5, 2,3,4,5, 1,4,5,6 };
  static int hex2tet[6][4] = { 2,4,5,6, 1,2,4,5, 2,3,4,6,
                               3,4,6,8, 4,5,6,8, 3,6,7,8 };

  up = down = 0.0;

  for (kx = 0; kx < ncell; kx++) {

    FX_Cell2Nodes(globals, cells[kx], &n, node);

    if (n == 4) {

      /* tetrahedra */
      for (i = 0; i < 4; i++) {
        kn = node[i]-1;
        if ((globals->cmask & 64) == 0) {
          coor[i] = globals->coor[kn];
        } else {
          coor[i].x = globals->xcoor[kn];
          coor[i].y = globals->ycoor[kn];
          coor[i].z = globals->zcoor[kn];
        }
        scal[i] = scalar[kn];
      }
      FX_CutVolTet(val, scal, coor, &abov, &belo);
      up   += abov;
      down += belo;

    } else if (n == 5) {

      /* pyramids */
      for (j = 0; j < 2; j++) {
        for (i = 0; i < 4; i++) {
          k  = pyr2tet[j][i]-1;
          kn = node[k]-1;
          if ((globals->cmask & 64) == 0) {
            coor[i] = globals->coor[kn];
          } else {
            coor[i].x = globals->xcoor[kn];
            coor[i].y = globals->ycoor[kn];
            coor[i].z = globals->zcoor[kn];
          }
          scal[i] = scalar[kn];
        }
        FX_CutVolTet(val, scal, coor, &abov, &belo);
        up   += abov;
        down += belo;
      }

    } else if (n == 6) {

      /* prisms */
      for (j = 0; j < 3; j++) {
        for (i = 0; i < 4; i++) {
          k  = pri2tet[j][i]-1;
          kn = node[k]-1;
          if ((globals->cmask & 64) == 0) {
            coor[i] = globals->coor[kn];
          } else {
            coor[i].x = globals->xcoor[kn];
            coor[i].y = globals->ycoor[kn];
            coor[i].z = globals->zcoor[kn];
          }
          scal[i] = scalar[kn];
        }
        FX_CutVolTet(val, scal, coor, &abov, &belo);
        up   += abov;
        down += belo;
      }

    } else if (n == 8) {

      /* hexahedra */
      for (j = 0; j < 6; j++) {
        for (i = 0; i < 4; i++) {
          k  = hex2tet[j][i]-1;
          kn = node[k]-1;
          if ((globals->cmask & 64) == 0) {
            coor[i] = globals->coor[kn];
          } else {
            coor[i].x = globals->xcoor[kn];
            coor[i].y = globals->ycoor[kn];
            coor[i].z = globals->zcoor[kn];
          }
          scal[i] = scalar[kn];
        }
        FX_CutVolTet(val, scal, coor, &abov, &belo);
        up   += abov;
        down += belo;
      }

    }
  }

  *above = up;
  *below = down;
}


/*
 *  look towards the root
 */

static int
FX_GetRoot(TheTree *packet, int node)
{
  int n;

  n = packet->nodeTable[node-1];
  if (n < 0) return n;

  while (packet->tree[n].root != -1) {
    n = packet->tree[n].root;
  }
  packet->nodeTable[node-1] = n;
  return n;
}


/*
 *  Constructs the tree by looking at edges
 */

static int
FX_MakeTree(TheTree *packet, int kc, int n1, int n2)
{
  int      i, link1, link2, k1, k2;
  Tree     *tmp;
  CellList *last;

  k1 = FX_GetRoot(packet, n1);
  k2 = FX_GetRoot(packet, n2);
  if ((k1 == -2) && (k2 == -2)) return 0;

  /* at least one node is OK -- connect to the tree */

  if (packet->ntree+1 >= packet->mtree) {
    i   = packet->mtree + TREECHUNK;
    tmp = (Tree *) realloc((void *) packet->tree, i*sizeof(Tree));
    if (tmp == NULL) {
      printf(" FX Error: REALLOC for Tree (FX_MakeTree)!\n");
      return 1;
    }
    packet->mtree = i;
    packet->tree  = tmp;
  }

  if ((k1 < 0) && (k2 < 0)) {
    /* virginal nodes -- make new branch */
    if (k1 == -1) packet->nodeTable[n1-1] = packet->ntree;
    if (k2 == -1) packet->nodeTable[n2-1] = packet->ntree;
    packet->tree[packet->ntree].root = -1;
    packet->tree[packet->ntree].list = NULL;
    packet->tree[packet->ntree].last = NULL;
    link1 = packet->ntree;
    packet->ntree += 1;
  } else if ((k1 >= 0) && (k2 < 0)) {
    /* 1 virginal node -- branch connecting node and branch */
    link1 = k1;
    if (k2 == -1) packet->nodeTable[n2-1] = link1;
  } else if ((k1 < 0) && (k2 >= 0)) {
    /* 1 virginal node -- branch connecting node and branch */
    link1 = k2;
    if (k1 == -1) packet->nodeTable[n1-1] = link1;
  } else {
    /* branch connecting branch and branch -- are they the same? */
    link1 = k1;
    link2 = k2;
    if (link1 == packet->ltree) {
      link1 = link2;
      link2 = packet->ltree;
    }
    if (link1 != link2) {
      packet->tree[packet->ntree].root = -1;
      packet->tree[packet->ntree].list = packet->tree[link1].list;
      last = packet->tree[link1].last;
      if (last == NULL) {
        packet->tree[packet->ntree].list = packet->tree[link2].list;
        packet->tree[packet->ntree].last = packet->tree[link2].last;
      } else {
        last->next = packet->tree[link2].list;
        packet->tree[packet->ntree].last = packet->tree[link2].last;
      }
      if (packet->ltree == link2) packet->ltree = packet->ntree;
      packet->tree[link1].root = packet->ntree;
      packet->tree[link1].list = NULL;
      packet->tree[link2].root = packet->ntree;
      packet->tree[link2].list = NULL;
      link1 = packet->ntree;
      packet->ntree += 1;
    }
  }

  /* do we need to link up to anything else in this cell? */

  if (packet->ltree == link1) return 0;
  if (packet->ltree != -1) {
    link2 = packet->ltree;
    packet->tree[packet->ntree].root = -1;
    packet->tree[packet->ntree].list = packet->tree[link1].list;
    last = packet->tree[link1].last;
    if (last == NULL) {
      packet->tree[packet->ntree].list = packet->tree[link2].list;
      packet->tree[packet->ntree].last = packet->tree[link2].last;
    } else {
      last->next = packet->tree[link2].list;
      packet->tree[packet->ntree].last = packet->tree[link2].last;
    }
    packet->tree[link1].root = packet->ntree;
    packet->tree[link1].list = NULL;
    packet->tree[link2].root = packet->ntree;
    packet->tree[link2].list = NULL;
    link1 = packet->ntree;
    packet->ntree += 1;
  }

  /* make sure kc is in list -- should be last */

  last = packet->tree[link1].last;
  if (last == NULL) {
    packet->tree[link1].list = &(packet->cellTable[kc-1]);
    packet->tree[link1].last = &(packet->cellTable[kc-1]);
  } else {
    if (last->kcell != kc) {
      last->next = &(packet->cellTable[kc-1]);
      packet->tree[link1].last = &(packet->cellTable[kc-1]);
    }
  }
  if (packet->cellTable[kc-1].next != NULL) {
    printf(" FX Error: Thread List Cell = %d (FX_MakeTree)!\n", kc);
    return 1;
  }
  packet->ltree = link1;

  return 0;
}


/*
 *  Returns the number and cluster of cells above/below the cut value
 */

void
FX_OrderCutVol(FXglobal *globals, float val, int abel, float *scalar,
               int *ncluster, int **clusters, int **cells,
               int **cnode, int **cnodes)
{
  int      kc, n1, n2, i, j, k, kcr, kn, *clust, *cell, *enode, *nodes;
  int      ib, ni, nj, nk, nc, kns, kz, bnode, node[8], type;
  TheTree  packet;
  CellList *link;
  static int tet2edg[ 6][2] = { 1,2, 1,3, 1,4, 2,3, 2,4, 3,4 };
  static int pyr2edg[ 8][2] = { 1,2, 2,3, 3,4, 4,1, 1,5, 2,5, 3,5, 4,5 };
  static int pri2edg[ 9][2] = { 1,2, 2,3, 3,4, 4,1, 5,6, 1,6, 4,6, 2,5, 3,5 };
  static int hex2edg[12][2] = { 1,2, 2,3, 3,4, 4,1, 5,6, 6,7, 7,8, 8,5,
                                1,5, 2,6, 3,7, 4,8 };

  *ncluster = -1;
  *clusters = *cells = *cnode = *cnodes = NULL;
   clust    =  cell  =  enode =   nodes = NULL;

  packet.cellTable = (CellList *) malloc(globals->total_cells*sizeof(CellList));
  if (packet.cellTable == NULL) {
    printf(" FX Error: MALLOC for cellTable (FX_OrderCutVol)!\n");
    return;
  }
  packet.nodeTable = (int *) malloc(globals->total_nodes*sizeof(int));
  if (packet.nodeTable == NULL) {
    free(packet.cellTable);
    printf(" FX Error: MALLOC for nodeTable (FX_OrderCutVol)!\n");
    return;
  }
  packet.mtree = TREECHUNK;
  packet.ntree = 0;
  packet.tree  = (Tree *) malloc(packet.mtree*sizeof(Tree));
  if (packet.tree == NULL) {
    free(packet.cellTable);
    free(packet.nodeTable);
    printf(" FX Error: MALLOC for Tree (FX_OrderCutVol)!\n");
    return;
  }

  /* initialize tables */

  for (i = 0; i < globals->total_nodes; i++)
    if (abel == 0) {
      /* below the value */
      if (scalar[i] < val) {
        packet.nodeTable[i] = -1;
      } else {
        packet.nodeTable[i] = -2;
      }
    } else {
      /* above the value */
      if (scalar[i] >= val) {
        packet.nodeTable[i] = -1;
      } else {
        packet.nodeTable[i] = -2;
      }
    }

  for (kc = 0; kc < globals->total_cells; kc++) {
    packet.cellTable[kc].kcell = kc+1;
    packet.cellTable[kc].next  = NULL;
  }

  /* structured blocks */

  if ((globals->nblock > 0) && (globals->iblank != NULL))
    bnode = globals->block[0].last_node - 
            globals->block[0].ni*globals->block[0].nj*globals->block[0].nk;
  for (ib = 0; ib < globals->nblock; ib++) {
    ni  = globals->block[ib].ni;
    nj  = globals->block[ib].nj;
    nk  = globals->block[ib].nk;
    nc  = (ni-1)*(nj-1)*(nk-1);
    kns = globals->block[ib].last_node - ni*nj*nk + 1;
    kcr = globals->block[ib].last_cell - (ni-1)*(nj-1)*(nk-1) + 1;
    if (globals->iblank == NULL) {
      for (kc = 0; kc < nc; kc++) {
        i  =  kc%(ni-1);
        j  = (kc/(ni-1))%(nj-1);
        k  =  kc/((ni-1)*(nj-1));
        kz =  kns + i + j*ni + k*ni*nj;
        node[0] = kz;
        node[1] = kz + 1;
        node[2] = kz + 1 + ni;
        node[3] = kz     + ni;
        node[4] = kz          + ni*nj;
        node[5] = kz + 1      + ni*nj;
        node[6] = kz + 1 + ni + ni*nj;
        node[7] = kz     + ni + ni*nj;
        for (kn = i = 0; i < 8; i++) 
          if (packet.nodeTable[node[i]-1] == -2) kn++;
        if (kn == 8) continue;
        packet.ltree = -1;
        for (i = 0; i < 12; i++) {
          n1  = hex2edg[i][0]-1;
          n2  = hex2edg[i][1]-1;
          if (FX_MakeTree(&packet, kc+kcr, node[n1], node[n2]) != 0) goto err;
        }
      }
    } else {
      for (kc = 0; kc < nc; kc++) {
        i  =  kc%(ni-1);
        j  = (kc/(ni-1))%(nj-1);
        k  =  kc/((ni-1)*(nj-1));
        kz =  kns + i + j*ni + k*ni*nj;
        node[0] = kz;
        node[1] = kz + 1;
        node[2] = kz + 1 + ni;
        node[3] = kz     + ni;
        node[4] = kz          + ni*nj;
        node[5] = kz + 1      + ni*nj;
        node[6] = kz + 1 + ni + ni*nj;
        node[7] = kz     + ni + ni*nj;
        for (kn = j = i = 0; i < 8; i++) {
          if (globals->iblank[node[i]-bnode-1] ==  0) j++;
          if (packet.nodeTable[node[i]-1]      == -2) kn++;
        }
        if ((j != 0) || (kn == 8)) continue;
        packet.ltree = -1;
        for (i = 0; i < 12; i++) {
          n1  = hex2edg[i][0]-1;
          n2  = hex2edg[i][1]-1;
          if (FX_MakeTree(&packet, kc+kcr, node[n1], node[n2]) != 0) goto err;
        }
      }
    }
  }

  if ((globals->cmask & 128) == 0) {

    /* tetrahedra */
    for (kc = 0; kc < globals->ntets; kc++) {
      for (kn = i = 0; i < 4; i++) {
        n1 = globals->tets[kc].nodes[i]-1;
        if (packet.nodeTable[n1] == -2) kn++;
      }
      if (kn == 4) continue;
      packet.ltree = -1;
      for (i = 0; i < 6; i++) {
        n1 = tet2edg[i][0]-1;
        n2 = tet2edg[i][1]-1;
        if (FX_MakeTree(&packet, kc+1, globals->tets[kc].nodes[n1],
                        globals->tets[kc].nodes[n2]) != 0) goto err;
      }
    }

    /* pyramids */
    for (kc = 0; kc < globals->npyra; kc++) {
      for (kn = i = 0; i < 5; i++) {
        n1 = globals->pyra[kc].nodes[i]-1;
        if (packet.nodeTable[n1] == -2) kn++;
      }
      if (kn == 5) continue;
      packet.ltree = -1;
      for (i = 0; i < 8; i++) {
        n1  = pyr2edg[i][0]-1;
        n2  = pyr2edg[i][1]-1;
        kcr = kc + globals->ntets + 1;
        if (FX_MakeTree(&packet, kcr, globals->pyra[kc].nodes[n1],
                        globals->pyra[kc].nodes[n2]) != 0) goto err;
      }
    }

    /* prisms */
    for (kc = 0; kc < globals->nprism; kc++) {
      for (kn = i = 0; i < 6; i++) {
        n1 = globals->prism[kc].nodes[i]-1;
        if (packet.nodeTable[n1] == -2) kn++;
      }
      if (kn == 6) continue;
      packet.ltree = -1;
      for (i = 0; i < 9; i++) {
        n1  = pri2edg[i][0]-1;
        n2  = pri2edg[i][1]-1;
        kcr = kc + globals->ntets + globals->npyra + 1;
        if (FX_MakeTree(&packet, kcr, globals->prism[kc].nodes[n1],
                        globals->prism[kc].nodes[n2]) != 0) goto err;
      }
    }

    /* hexahedra */
    for (kc = 0; kc < globals->nhexa; kc++) {
      for (kn = i = 0; i < 8; i++) {
        n1 = globals->hexa[kc].nodes[i]-1;
        if (packet.nodeTable[n1] == -2) kn++;
      }
      if (kn == 8) continue;
      packet.ltree = -1;
      for (i = 0; i < 12; i++) {
        n1  = hex2edg[i][0]-1;
        n2  = hex2edg[i][1]-1;
        kcr = kc + globals->ntets + globals->npyra + globals->nprism + 1;
        if (FX_MakeTree(&packet, kcr, globals->hexa[kc].nodes[n1],
                        globals->hexa[kc].nodes[n2]) != 0) goto err;
      }
    }

  } else {

    /* construct elements on-the-fly */

    for (kc = 0; kc < globals->total_dis; kc++) {
      i = kc+1;
      FXCELLGET(&i, &type, node);

      switch (type) {

      case 1:
        /* tetrahedra */
        for (kn = i = 0; i < 4; i++) 
          if (packet.nodeTable[node[i]-1] == -2) kn++;
        if (kn == 4) break;
        packet.ltree = -1;
        for (i = 0; i < 6; i++) {
          n1 = tet2edg[i][0]-1;
          n2 = tet2edg[i][1]-1;
          if (FX_MakeTree(&packet, kc+1, node[n1], node[n2]) != 0) goto err;
        }
        break;

      case 2:
        /* pyramids */
        for (kn = i = 0; i < 5; i++) 
          if (packet.nodeTable[node[i]-1] == -2) kn++;
        if (kn == 5) break;
        packet.ltree = -1;
        for (i = 0; i < 8; i++) {
          n1  = pyr2edg[i][0]-1;
          n2  = pyr2edg[i][1]-1;
          if (FX_MakeTree(&packet, kc+1, node[n1], node[n2]) != 0) goto err;
        }
        break;

      case 3:
        /* prisms */
        for (kn = i = 0; i < 6; i++) 
          if (packet.nodeTable[node[i]-1] == -2) kn++;
        if (kn == 6) break;
        packet.ltree = -1;
        for (i = 0; i < 9; i++) {
          n1  = pri2edg[i][0]-1;
          n2  = pri2edg[i][1]-1;
          if (FX_MakeTree(&packet, kc+1, node[n1], node[n2]) != 0) goto err;
        }
        break;

      case 4:
        /* hexahedra */
        for (kn = i = 0; i < 8; i++) 
          if (packet.nodeTable[node[i]-1] == -2) kn++;
        if (kn == 8) break;
        packet.ltree = -1;
        for (i = 0; i < 12; i++) {
          n1  = hex2edg[i][0]-1;
          n2  = hex2edg[i][1]-1;
          if (FX_MakeTree(&packet, kc+1, node[n1], node[n2]) != 0) goto err;
        }
      }
    }

  }

  /* store the root number to all of the nodes */

  for (kn = 0, i = 1; i <= globals->total_nodes; i++) 
    if (FX_GetRoot(&packet, i) >= 0) kn++;

  /* count the number of branches */

  for (kc = kcr = i = 0; i < packet.ntree; i++)
    if (packet.tree[i].list != NULL) {
      kcr++;
      link = packet.tree[i].list;
      while (link != NULL) {
        kc++;
        link = link->next;
      }
    }

  if (kcr > 0) {
    clust = (int *) malloc(kcr*sizeof(int));
    if (clust == NULL) {
      printf(" FX Error: MALLOC for Clusters %d (FX_OrderCutVol)!\n", kcr);
      goto err;
    }
    enode = (int *) malloc(kcr*sizeof(int));
    if (enode == NULL) {
      free(clust);
      printf(" FX Error: MALLOC for Cluster of Nodes %d (FX_OrderCutVol)!\n", kcr);
      goto err;
    }
  }
  if (kc > 0) {
    cell = (int *) malloc(kc*sizeof(int));
    if (cell == NULL) {
      if (kcr > 0) {
        free(clust);
        free(enode);
      }
      printf(" FX Error: MALLOC for Cells %d (FX_OrderCutVol)!\n", kc);
      goto err;
    }
  }
  if (kn > 0) {
    nodes = (int *) malloc(kn*sizeof(int));
    if (nodes == NULL) {
      if (kc  > 0) free(cell);
      if (kcr > 0) {
        free(enode);
        free(clust);
      }
      printf(" FX Error: MALLOC for Nodes %d (FX_OrderCutVol)!\n", kn);
      goto err;
    }
  }

  /* allocate and store the cells in each sub-volume */

  for (kc = kcr = i = 0; i < packet.ntree; i++)
    if (packet.tree[i].list != NULL) {
      link = packet.tree[i].list;
      while (link != NULL) {
        cell[kc] = link->kcell;
        kc++;
        link = link->next;
      }
      packet.tree[i].root = kcr;
      enode[kcr] = 0;
      clust[kcr] = kc;
      kcr++;
    }

  /* take care of the nodes */

  for (i = 0; i < globals->total_nodes; i++) {
    k = packet.nodeTable[i];
    if (k < 0) continue;
    kc = packet.tree[k].root;
    enode[kc]++;
  }

  packet.cellTable[0].kcell = 0;
  for (kc = 1; kc < kcr; kc++) {
    kz = enode[kc-1];
    packet.cellTable[kc].kcell = kz;
    enode[kc] += kz;
  }
/* MGL ADDITION 29 Sept 2006
 * This line was modified to keep the calc from crashing by adding kcr > 0
 * MGL ADDITION 29 Sept 2006 */
  if (kcr > 0 && enode[kcr-1] != kn) {
    if (nodes != NULL) free(nodes);
    if (cell  != NULL) free(cell);
    if (enode != NULL) free(enode);
    if (clust != NULL) free(clust);
    printf(" FX Error: Nodes Mismatch %d != %d (FX_OrderCutVol)!\n",
           kn, enode[kcr-1]);
    goto err;
  }

  for (i = 0; i < globals->total_nodes; i++) {
    k = packet.nodeTable[i];
    if (k < 0) continue;
    kc = packet.tree[k].root;
    kz = packet.cellTable[kc].kcell;
    nodes[kz] = i+1;
    packet.cellTable[kc].kcell++;
  }
  
  *ncluster = kcr;
  *clusters = clust;
  *cells    = cell;
  *cnode    = enode;
  *cnodes   = nodes;

err:
  free(packet.tree);
  free(packet.cellTable);
  free(packet.nodeTable);
}
