/*
 * FX_VortexCore.c
 *
 * FX module that computes the Vortex Cores.
 *
 * Copyright 1999 - 2000, Massachusetts Institute of Technology.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#define COMPLEX_CONJUGATE -3

#include "FX_types.h"
#include "FX_VC.h"


#ifdef WIN32

#define PREFIX __stdcall

#else

#define PREFIX

#ifdef UNDERSCORE

#define FXVEL         fxvel_
#define FXVELPTR      fxvelptr_
#define FXVELP3D      fxvelp3d_
#define FXCELLGET     fxcellget_

#define FX_VORTEXCORE fx_vortexcore_

#else

#define FXVEL         fxvel
#define FXVELPTR      fxvelptr
#define FXVELP3D      fxvelp3d
#define FXCELLGET     fxcellget

#define FX_VORTEXCORE fx_vortexcore

#endif
#endif


/* globals */

  extern FXglobal *FX_global;


/* externals */

  extern void PREFIX FXVEL(FXcoor *,  FXcoor *);
  extern void PREFIX FXVELPTR(FXcoor **,  FXcoor *);
  extern void PREFIX FXVELP3D(float **, float **, float **,  FXcoor *);
  extern void PREFIX FXCELLGET(int *, int *, int *);

  extern void        FX_VelGrad(FXglobal *, float **);
  extern int         FX_RotAxis(int, float *, float *);
  extern int         FX_AxisXTri(FXcoor *, FXcoor *, int, int, int, float *, 
                                 float, float, float, float *, float *);
  extern int         FX_AxisXTriP3d(float *, float *, float *, float *,
                                    float *, float *, int, int, int, float *, 
                                    float, float, float, float *, float *);
  extern void        FX_SegmentVC(int, vcf *);


  /* add the face to the vortex core structures */

static int
FX_AddFaceVC(int n1, int n2, int n3, int itet, int face, float vort, 
             float *xyz, int *tab, vcf *pack)
{
  int   i, j, k1, k2, k3;
  faces *tmp;

  /* order the node numbers */

  k1 = n1;
  k2 = n2;
  k3 = n3;
  if (k1 > k2) {
    i  = k2;
    k2 = k1;
    k1 = i;
  }
  if (k2 > k3) {
    i  = k2;
    k2 = k3;
    k3 = i;
  }
  if (k1 > k2) {
    i  = k2;
    k2 = k1;
    k1 = i;
  }

  i = pack->nodes[k1-1];
  if (i == NOTFILLED) {

    /* virgin node */

    if (pack->nfaces >= pack->mfaces) {
      j = pack->mfaces + VCFHUNK;
      tmp = (faces *) realloc(pack->table, j*sizeof(faces));
      if (tmp == NULL) return -1;
      pack->table  = tmp;
      pack->mfaces = j;
    }
    pack->nodes[k1-1] = pack->nfaces;
    pack->table[pack->nfaces].node2   = k2;
    pack->table[pack->nfaces].node3   = k3;
    pack->table[pack->nfaces].itet[0] = itet;
    pack->table[pack->nfaces].itet[1] = NOTFILLED;
    pack->table[pack->nfaces].itet[2] = NOTFILLED;
    pack->table[pack->nfaces].itet[3] = NOTFILLED;
    pack->table[pack->nfaces].face[0] = face;
    pack->table[pack->nfaces].vort    = vort;
    pack->table[pack->nfaces].xyz[0]  = xyz[0];
    pack->table[pack->nfaces].xyz[1]  = xyz[1];
    pack->table[pack->nfaces].xyz[2]  = xyz[2];
    pack->table[pack->nfaces].thread  = NOTFILLED;
    *tab = pack->nfaces;
    pack->nfaces++;
    return 0;
  } else {

    /* has a thread */

again:
    if (pack->table[i].node2 == k2) {
      if (pack->table[i].node3 == k3) {
        /* hit old face! */
        xyz[0] = pack->table[i].xyz[0];
        xyz[1] = pack->table[i].xyz[1];
        xyz[2] = pack->table[i].xyz[2];
        *tab = i;
        if (pack->table[i].itet[3] != NOTFILLED) return -2;
        j = 3;
        if (pack->table[i].itet[1] == NOTFILLED) {
          j = 1;
        } else if (pack->table[i].itet[2] == NOTFILLED) {
          j = 2;
        }
        pack->table[i].itet[j] = itet;
        pack->table[i].face[j] = face;
        return j;
      }
    }

    if (pack->table[i].thread == NOTFILLED) {
      /* end of the thread */
      if (pack->nfaces >= pack->mfaces) {
        j = pack->mfaces + VCFHUNK;
        tmp = (faces *) realloc(pack->table, j*sizeof(faces));
        if (tmp == NULL) return -1;
        pack->table  = tmp;
        pack->mfaces = j;
      }
      pack->table[i].thread = pack->nfaces;
      pack->table[pack->nfaces].node2   = k2;
      pack->table[pack->nfaces].node3   = k3;
      pack->table[pack->nfaces].itet[0] = itet;
      pack->table[pack->nfaces].itet[1] = NOTFILLED;
      pack->table[pack->nfaces].itet[2] = NOTFILLED;
      pack->table[pack->nfaces].itet[3] = NOTFILLED;
      pack->table[pack->nfaces].face[0] = face;
      pack->table[pack->nfaces].vort    = vort;
      pack->table[pack->nfaces].xyz[0]  = xyz[0];
      pack->table[pack->nfaces].xyz[1]  = xyz[1];
      pack->table[pack->nfaces].xyz[2]  = xyz[2];
      pack->table[pack->nfaces].thread  = NOTFILLED;
      *tab = pack->nfaces;
      pack->nfaces++;
      return 0;
    }

    i = pack->table[i].thread;
    goto again;
  }
}


  /* intersect the tetrahedron with the axis of rotation */

static void
FX_TetInter(float *raxis, int tet[][4], int ntet, int ncell, vcf *pack)
{
  int   i, j, k, kn, nz, itet, np, n1, n2, n3, face[4], vctab[4], XTri;
  float dot, dis, vort, axis[3], point[3], vaxis[4][3], vmag[4];
  tets  *tmp;
  static int ftab[4][3] = { 1,2,3, 2,3,4, 3,4,1, 4,1,2 };

  for (i = 0; i < ntet; i++) {

    /* do we want this tet? */

   if ((FX_global->cmask & 64) == 0) {
      for (nz = j = 0; j < 4; j++) {
        kn = tet[i][j]-1;
        if (fabs(FX_global->vel[kn].x)+fabs(FX_global->vel[kn].y)+
            fabs(FX_global->vel[kn].z) == 0.0) nz++;
        if (fabs(raxis[4*kn])+fabs(raxis[4*kn+1])+fabs(raxis[4*kn+2]) == 0.0)
          nz++;
      }
    } else {
      for (nz = j = 0; j < 4; j++) {
        kn = tet[i][j]-1;
        if (fabs(FX_global->velx[kn])+fabs(FX_global->vely[kn])+
            fabs(FX_global->velz[kn]) == 0.0) nz++;
        if (fabs(raxis[4*kn])+fabs(raxis[4*kn+1])+fabs(raxis[4*kn+2]) == 0.0)
          nz++;
      }
    }
    if (nz != 0) continue;

    /* align all of the axes */

    kn = tet[i][0]-1;
    vaxis[0][0] = raxis[4*kn  ];
    vaxis[0][1] = raxis[4*kn+1];
    vaxis[0][2] = raxis[4*kn+2];
    vmag[0]     = raxis[4*kn+3];
    for (j = 1; j < 4; j++) {
      kn  = tet[i][j]-1;
      vmag[j] = raxis[4*kn+3];
      dot = vaxis[0][0]*raxis[4*kn  ] + vaxis[0][1]*raxis[4*kn+1] +
            vaxis[0][2]*raxis[4*kn+2];
      if (dot < 0.0) {
        vaxis[j][0] = -raxis[4*kn  ];
        vaxis[j][1] = -raxis[4*kn+1];
        vaxis[j][2] = -raxis[4*kn+2];
      } else {
        vaxis[j][0] =  raxis[4*kn  ];
        vaxis[j][1] =  raxis[4*kn+1];
        vaxis[j][2] =  raxis[4*kn+2];
      }
    }

    /* loop over the 4 faces and count the number of intersections */

    itet = -1;
    for (np = j = 0; j < 4; j++) {
      n1 = ftab[j][0]-1;
      n2 = ftab[j][1]-1;
      n3 = ftab[j][2]-1;
      axis[0] = vaxis[n1][0] + vaxis[n2][0] + vaxis[n3][0];
      axis[1] = vaxis[n1][1] + vaxis[n2][1] + vaxis[n3][1];
      axis[2] = vaxis[n1][2] + vaxis[n2][2] + vaxis[n3][2];
      dis     = sqrt(axis[0]*axis[0] + axis[1]*axis[1] +
                     axis[2]*axis[2]);
      if (dis !=  0.0) {
        axis[0] /= dis;
        axis[1] /= dis;
        axis[2] /= dis;

        /* check for intersection */

        if ((FX_global->cmask & 64) == 0) {
          XTri = FX_AxisXTri(FX_global->coor, FX_global->vel,
                             tet[i][n1], tet[i][n2], tet[i][n3], axis,
                             vmag[n1], vmag[n2], vmag[n3], &vort, point);
        } else {
          XTri = FX_AxisXTriP3d(FX_global->xcoor, FX_global->ycoor,
                                FX_global->zcoor, FX_global->velx,
                                FX_global->vely,  FX_global->velz,
                                tet[i][n1], tet[i][n2], tet[i][n3], axis,
                                vmag[n1], vmag[n2], vmag[n3], &vort, point);
        }
        if (XTri == 1) {
          if (np == 0) {
            if (pack->ntets >= pack->mtets) {
              nz  = pack->mtets + VCFHUNK;
              tmp = (tets *) realloc(pack->tets, nz*sizeof(tets));
              if (tmp == NULL) return;
              pack->tets  = tmp;
              pack->mtets = nz;
            }
            pack->tets[pack->ntets].cell  = ncell;
            pack->tets[pack->ntets].tet   = i+1;
            pack->tets[pack->ntets].nface = 0;
            itet = pack->ntets;
            pack->ntets += 1;
          }
          face[np] = j+1;
          k = FX_AddFaceVC(tet[i][n1], tet[i][n2], tet[i][n3], itet, j+1,
                           vort, point, &vctab[np], pack);
          if (k == -1) printf(" FX_AddFaceVC Error!\n");
          if (k  >  1) printf(" FX_AddFaceVC multiple Hit - %d %d %d - %d\n",
                              ncell, i+1, j+1, k);
          np++;
        }
      }
    }
/*
    if (np == 0) continue;
    if (np == 1) vctab[1] = face[0];
*/
    if ((np > 1) && (itet >= 0) && (itet < pack->ntets)) {
      pack->tets[itet].nface   = np;
      pack->tets[itet].face[0] = vctab[0];
      pack->tets[itet].face[1] = vctab[1];
      pack->tets[itet].face[2] = vctab[2];
      pack->tets[itet].face[3] = vctab[3];
    }

  }

}


void PREFIX
FX_VORTEXCORE(int *type, int *nseg, int **eseg, float **xyz, float **stren)
{
  int   i, j, k, kc, ib, ni, nj, nk, nc, kns, kz, bnode, tet[6][4], node[8];
  int   ncell, *pseg;
  float *vel_grad, *raxis, *pxyz, *pstren;
  vcf   pack;
  static int pyr2tet[2][4] = { 1,2,4,5, 2,3,4,5 };
  static int pri2tet[3][4] = { 1,2,4,5, 2,3,4,5, 1,4,5,6 };
  static int hex2tet[6][4] = { 2,4,5,6, 1,2,4,5, 2,3,4,6,
                               3,4,6,8, 4,5,6,8, 3,6,7,8 };
  *nseg = 0;
  *eseg = NULL;
  *xyz  = *stren = NULL;

  if (FX_global == NULL) {
    printf(" FX Error: NOT Initialized (FX_ShockFind)!\n");
    return;
  }

  if (FX_global->status != 0) {
    printf(" FX Error: Initialization Status = %d (FX_ShockFind)!\n",
           FX_global->status);
    return;
  }

  /* get velocity if required */

  if (FX_global->vload == 0) {
    if ((FX_global->cmask & 8) == 0) {
      FXVEL    (FX_global->vel, FX_global->hvel);
    } else {
      if ((FX_global->cmask & 64) == 0) {
        FXVELPTR(&FX_global->vel, FX_global->hvel);
      } else {
        FXVELP3D(&FX_global->velx, &FX_global->vely, &FX_global->velz,
                  FX_global->hvel);
      }
    }
    FX_global->vload = 1;
  }

  /* compute the velocity gradient tensor */

  FX_VelGrad(FX_global, &vel_grad);
  if (vel_grad == NULL) return;

  /* get the axis of rotation for each node */

  raxis = (float *) malloc(4*FX_global->total_nodes*sizeof(float));
  if (raxis == NULL) {
    printf(" FX Error: MALLOC for Axis storage (FX_VortexCore)!\n");
    free(vel_grad);
    return;
  }
  for (i = 0; i < FX_global->total_nodes; i++) 
    FX_RotAxis(*type, &vel_grad[9*i], &raxis[4*i]);

  free(vel_grad);

  /* intiialize temp storage */

  pack.mfaces = VCFHUNK;
  pack.table  = (faces *) malloc(VCFHUNK*sizeof(faces));
  if (pack.table == NULL) {
    printf(" FX Error: MALLOC for Face storage (FX_VortexCore)!\n");
    free(raxis);
    return;
  }
  pack.nodes  = (int *) malloc(FX_global->total_nodes*sizeof(int));
  if (pack.nodes == NULL) {
    printf(" FX Error: MALLOC for Node storage (FX_VortexCore)!\n");
    free(pack.table);
    free(raxis);
    return;
  }
  pack.nfaces = 0;
  pack.mtets  = VCFHUNK;
  pack.tets   = (tets *) malloc(VCFHUNK*sizeof(tets));
  if (pack.nodes == NULL) {
    printf(" FX Error: MALLOC for Tet storage (FX_VortexCore)!\n");
    free(pack.nodes);
    free(pack.table);
    free(raxis);
    return;
  }
  pack.ntets  = 0;
  pack.segs   = NULL;

  for (i = 0; i < FX_global->total_nodes; i++) pack.nodes[i] = NOTFILLED;

  /* look at all possible tetrahedral intersections */

  /* structured blocks */

  if ((FX_global->nblock > 0) && (FX_global->iblank != NULL))
    bnode = FX_global->block[0].last_node - (FX_global->block[0].ni*
                      FX_global->block[0].nj*FX_global->block[0].nk);
  ncell = FX_global->total_dis;
  for (ib = 0; ib < FX_global->nblock; ib++) {
    ni  = FX_global->block[ib].ni;
    nj  = FX_global->block[ib].nj;
    nk  = FX_global->block[ib].nk;
    nc  = (ni-1)*(nj-1)*(nk-1);
    kns = FX_global->block[ib].last_node - ni*nj*nk + 1;
    if (FX_global->iblank == NULL) {
      for (kc = 0; kc < nc; kc++) {
        ncell++;
        i  =  kc%(ni-1);
        j  = (kc/(ni-1))%(nj-1);
        k  =  kc/((ni-1)*(nj-1));
        kz =  kns + i + j*ni + k*ni*nj;
        node[0] = kz;
        node[1] = kz + 1;
        node[2] = kz + 1 + ni;
        node[3] = kz     + ni;
        node[4] = kz          + ni*nj;
        node[5] = kz + 1      + ni*nj;
        node[6] = kz + 1 + ni + ni*nj;
        node[7] = kz     + ni + ni*nj;
        for (j = 0; j < 6; j++)
          for (i = 0; i < 4; i++)
            tet[j][i] = node[hex2tet[j][i]-1];
        FX_TetInter(raxis, tet, 6, ncell, &pack);
      }
    } else {
      for (kc = 0; kc < nc; kc++) {
        ncell++;
        i  =  kc%(ni-1);
        j  = (kc/(ni-1))%(nj-1);
        k  =  kc/((ni-1)*(nj-1));
        kz =  kns + i + j*ni + k*ni*nj;
        node[0] = kz;
        node[1] = kz + 1;
        node[2] = kz + 1 + ni;
        node[3] = kz     + ni;
        node[4] = kz          + ni*nj;
        node[5] = kz + 1      + ni*nj;
        node[6] = kz + 1 + ni + ni*nj;
        node[7] = kz     + ni + ni*nj;
        for (j = i = 0; i < 8; i++)
          if (FX_global->iblank[node[i]-bnode-1] == 0) j++;
        if (j != 0) continue;

        for (j = 0; j < 6; j++)
          for (i = 0; i < 4; i++) 
            tet[j][i] = node[hex2tet[j][i]-1];
        FX_TetInter(raxis, tet, 6, ncell, &pack);
      }
    }
  }

  if ((FX_global->cmask & 128) == 0) {

    /* tetrahedra */
    for (ncell = kc = 0; kc < FX_global->ntets; kc++) {
      ncell++;
      for (i = 0; i < 4; i++)
        tet[0][i] = FX_global->tets[kc].nodes[i];
      FX_TetInter(raxis, tet, 1, ncell, &pack);
    }

    /* pyramids */
    for (kc = 0; kc < FX_global->npyra; kc++) {
      ncell++;
      for (j = 0; j < 2; j++)
        for (i = 0; i < 4; i++) 
          tet[j][i] = FX_global->pyra[kc].nodes[pyr2tet[j][i]-1];
      FX_TetInter(raxis, tet, 2, ncell, &pack);
    }

    /* prisms */
    for (kc = 0; kc < FX_global->nprism; kc++) {
      ncell++;
      for (j = 0; j < 3; j++)
        for (i = 0; i < 4; i++) 
          tet[j][i] = FX_global->prism[kc].nodes[pri2tet[j][i]-1];
      FX_TetInter(raxis, tet, 3, ncell, &pack);
    }

    /* hexahedra */
    for (kc = 0; kc < FX_global->nhexa; kc++) {
      ncell++;
      for (j = 0; j < 6; j++)
        for (i = 0; i < 4; i++) 
          tet[j][i] = FX_global->hexa[kc].nodes[hex2tet[j][i]-1];
      FX_TetInter(raxis, tet, 6, ncell, &pack);
    }

  } else {

    /* construct elements on-the-fly */

    for (kc = 0; kc < FX_global->total_dis; kc++) {
      i = kc+1;
      FXCELLGET(&i, &k, node);

      switch (k) {

      case 1:
        /* tetrahedra */
        for (i = 0; i < 4; i++) tet[0][i] = node[i];
        FX_TetInter(raxis, tet, 1, kc+1, &pack);
        break;

      case 2:
        /* pyramids */
        for (j = 0; j < 2; j++)
          for (i = 0; i < 4; i++)
            tet[j][i] = node[pyr2tet[j][i]-1];
        FX_TetInter(raxis, tet, 2, kc+1, &pack);
        break;

      case 3:
        /* prisms */
        for (j = 0; j < 3; j++)
          for (i = 0; i < 4; i++)
            tet[j][i] = node[pri2tet[j][i]-1];
        FX_TetInter(raxis, tet, 3, kc+1, &pack);
        break;

      case 4:
        /* hexahedra */
        for (j = 0; j < 6; j++)
          for (i = 0; i < 4; i++)
            tet[j][i] = node[hex2tet[j][i]-1];
        FX_TetInter(raxis, tet, 6, kc+1, &pack);
        break;

      default:
        printf(" FX Warning: FXcellGet(%d) -> TYPE = %d (FX_VortexCore)!\n",
               kc+1, k);
      }
    }

  }

/*
  printf(" FX_VortexCore: nfaces = %d, ntets = %d\n", pack.nfaces, pack.ntets);
*/
  free(raxis);

  /* connect the line segments */

  FX_SegmentVC(FX_global->total_nodes, &pack);
  free(pack.tets);
  free(pack.nodes);

  if (pack.nsegs != 0) {
    pseg = (int *) malloc(pack.nsegs*sizeof(int));
    if (pseg == NULL) {
      printf(" FX Error: MALLOC for Seg storage (FX_VortexCore)!\n");
      for (i = 0; i < pack.nsegs; i++) free(pack.segs[i].faces);
      free(pack.segs);
      free(pack.table);
      return;
    }
    for (kz = i = 0; i < pack.nsegs; i++) {
      for (j = 0; j < pack.segs[i].nface; j++) 
        if (pack.segs[i].faces[j] != -1) kz++;
      pseg[i] = kz;
    }
    pxyz   = (float *) malloc(3*kz*sizeof(float));
    pstren = (float *) malloc(  kz*sizeof(float));
    if ((pxyz == NULL) || (pstren == NULL)) {
      printf(" FX Error: MALLOC for Point storage (FX_VortexCore)!\n");
      if (pxyz   != NULL) free(pxyz);
      if (pstren != NULL) free(pstren);
      free(pseg);
      for (i = 0; i < pack.nsegs; i++) free(pack.segs[i].faces);
      free(pack.segs);
      free(pack.table);
      return;
    }

    for (kz = i = 0; i < pack.nsegs; i++) {
      for (j = 0; j < pack.segs[i].nface; j++) {
        k = pack.segs[i].faces[j];
        if (k == -1) continue;
	if ((k < 0) || (k >= pack.nfaces)) 
	  printf(" K problem seg = %d (%d), j = %d (%d), %d %d\n",
	         i, pack.nsegs, j, pack.segs[i].nface, k, pack.nfaces);
        pxyz[3*kz  ] = pack.table[k].xyz[0];
        pxyz[3*kz+1] = pack.table[k].xyz[1];
        pxyz[3*kz+2] = pack.table[k].xyz[2];
        pstren[kz]   = pack.table[k].vort;
        kz++;
      }
      free(pack.segs[i].faces);
    }

    *nseg  = pack.nsegs;
    *eseg  = pseg;
    *xyz   = pxyz;
    *stren = pstren;
  }

  if (pack.segs != NULL) free(pack.segs);
  free(pack.table);
}
