#include <stdio.h>
#include <stdlib.h>

/*
 *    look at and check integrity of Viz Objects file structure
 */


/* Structures */

typedef struct {
  int   index;				/* field index */
  char  title[32];			/* name */
  float flims[2];                       /* lims */
  int   slen;                           /* rank */
} Field;

typedef struct {
  int   index;                          /* field index */
  char  title[32];                      /* name */
} Surf;

typedef struct {
  char  title[32];			/* name */
} Part;

typedef struct {
  int   numpart;			/* # of partitions */
  Part  *part;                          /* partitions */
  char  title[32];                      /* name */
  int   mfield;				/* allocated size of field */
  int   nfield;				/* # of field variables */
  Field *field;                         /* Field pointer */
  char  stitle[32];                     /* state variable title */
  int   slen;                           /* state variable rank */
  int   msurf;				/* allocates size of surf */
  int   nsurf;				/* number of surfaces */
  Surf  *surf;				/* Surface pounter */
} Discipline;

typedef struct {
  char title[32];			/* name */
  int  ranki;
  int  rankr;
} SubEntity;

typedef struct {
  int       index;			/* Entity index */
  char      title[32];			/* name */
  int       nint;			/* number of ints */
  int       nreal;			/* number of reals */
  int       nsubs;			/* number of sub-entities */
  int       factor;                     /* factor */
  SubEntity *sub;			/* subentity pointer */
} Entity;

typedef struct {
  float time;                           /* time at frame */
  long  offset;                         /* pointer into file */
} Frame;


/* functions used */


int tag_equal( char* tag, char* compare )
/* compare two 8 character tags */
{
  int i;

  for (i = 0; i < 8; i++) if (tag[i] != compare[i]) return 0;
  return 1;
}

void check_tag( char* tag, char* compare )
/* error if two 8 character tags do not match */
{
  int i;

  for (i = 0; i < 8; i++) if (tag[i] != compare[i]) {
      fprintf(stderr, "\n%8s should be: %8s\n", tag, compare);
      exit(1);
    }
}

void get_info( void* buf, int size, int len, FILE* handle )
/* read data -- error out if read is incomplete */
{
#ifdef BO_DEC
  int  i, j, num;
  char *swap, save;

  num = fread(buf, size, len, handle);
  if (size == 4) {
    swap = (char *) buf;
    for (i = j = 0; i  < num; i++, j += 4) {
      save      = swap[j  ];
      swap[j  ] = swap[j+3];
      swap[j+3] = save;
      save      = swap[j+2];
      swap[j+2] = swap[j+1];
      swap[j+1] = save;
    }
  } else if (size != 1) {
    fprintf(stderr, "\nWarning: Unknown sizeof for Read: %d!\n", size);
  }

#else
  int num;

  num = fread(buf, size, len, handle);
#endif
  if (num != len) {
    fprintf(stderr, "\nERROR - Premature EOF!\n");
    exit(1);
  }
}


int main( int argc, char* argv[] )
{
  FILE       *pv_handle;
  char       tag[41];
  float      rev, version, time, dummy, xmat[6];
  int        i, j, k, ndisc, index, idisc, nentity, mentity, ient;
  int        nframe, mframe, eindex, ipart, isub, len, nhit;
  long       offset, save_pos, savedbytes;
  Frame      *frame;
  Discipline *disc;
  Field      *field;
  Surf       *surf;
  Entity     *entity;
  SubEntity  *sub;
  Part       *part;

  printf("\nVisualization Object File Dump Rev 2.00\n\n");
  if (argc > 1) {
    pv_handle = fopen(argv[1], "rb");
  } else {
    pv_handle = fopen("pV3.viz", "rb");
  }
  if (pv_handle == NULL) {
    fprintf(stderr, "\nERROR - Can not open file!\n");
    exit(1);
  }

  /* make sure the file starts correctly */

  get_info(tag, sizeof(char),   8, pv_handle);
  check_tag(tag, "S*FILE**");
  get_info(&rev, sizeof(float), 1, pv_handle);
  tag[8] = 0;
  printf("%s rev: %f\n", tag, rev);
  get_info(tag, sizeof(char),   8, pv_handle);
  check_tag(tag, "S*HEAD**");
  printf("%s\n", tag);

  /* allocate structures with a first guess on size */

  tag[40] = 0;
  ndisc   = 1;
  disc  = (Discipline *) malloc(sizeof(Discipline));
  disc->numpart = 1;
  disc->part = (Part *) malloc(disc->numpart*sizeof(Part));
  disc->mfield = 10;
  disc->nfield = 0;
  disc->field = (Field *) malloc(disc->mfield*sizeof(Field));
  disc->slen = 0;
  disc->msurf = 10;
  disc->nsurf = 0;
  disc->surf = (Surf *) malloc(disc->msurf*sizeof(Surf));

  mentity = 10;
  nentity = 0;
  entity = (Entity *) malloc(mentity*sizeof(Entity));

  mframe = 1000;
  nframe = 0;
  frame = (Frame *) malloc(mframe*sizeof(Frame));

  savedbytes = 0;

/* read through the header records */

head:
  get_info(tag, sizeof(char),   8, pv_handle);

  if (tag_equal(tag, "E*HEAD**")) {
    tag[8] = 0;
    printf("%8s\n", tag);
    goto body;
  }

  if (tag_equal(tag, "WRITER**")) {
    get_info(&tag[8],  sizeof(char), 32, pv_handle);
    get_info(&version, sizeof(float), 1, pv_handle);
    printf("%s version: %f\n", tag, version);
    goto head;
  }

  if (tag_equal(tag, "DATETIM*")) {
    get_info(&tag[8], sizeof(char), 32, pv_handle);
    printf("%40s\n", tag);
    goto head;
  }

  if (tag_equal(tag, "NDSCPLN*")) {
    get_info(&ndisc, sizeof(int), 1, pv_handle);
    if (ndisc > 1) {
      disc = (Discipline *) realloc(disc, ndisc*sizeof(Discipline));
      if (disc == NULL) {
        fprintf(stderr, "\n%8s\n", tag);
        fprintf(stderr, "ERROR in realloc!");
        exit(1);
      }
      for (idisc = 1; idisc < ndisc; idisc++) {
        disc[idisc].numpart = 1;
        disc[idisc].part = (Part *) malloc(disc->numpart*sizeof(Part));
        disc[idisc].mfield = 10;
        disc[idisc].nfield = 0;
        disc[idisc].field = (Field *) malloc(disc->mfield*sizeof(Field));
	disc[idisc].slen = 0;
        disc[idisc].msurf = 10;
        disc[idisc].nsurf = 0;
        disc[idisc].surf = (Surf *) malloc(disc->msurf*sizeof(Surf));
      }
    }
    printf("%8s  # disciplines = %d\n", tag, ndisc);
    goto head;
  }

  if (tag_equal(tag, "DSCPLNE*")) {
    get_info(&tag[8], sizeof(char), 32, pv_handle);
    get_info(&idisc, sizeof(int),    1, pv_handle);
    if ((idisc < 0) || (idisc >= ndisc)) {
      fprintf(stderr, "\n%40s\n", tag);
      fprintf(stderr, "ERROR - Discipline Index Out of Range: %d!\n", idisc);
      exit(1);
    }
    get_info(&disc[idisc].numpart, sizeof(int), 1, pv_handle);
    if (disc[idisc].numpart > 1) {
      disc[idisc].part = (Part *) realloc(disc[idisc].part,
                                          disc[idisc].numpart*sizeof(Part));
      if (disc[idisc].part == NULL) {
        fprintf(stderr, "\n%8s\n", tag);
        fprintf(stderr, "ERROR in realloc!");
        exit(1);
      }
    }
    for (j = 0; j < 32; j++) disc[idisc].title[j] = tag[8+j];
    printf("%40s  disc # = %d npart = %d\n", tag, idisc, disc[idisc].numpart);
    goto head;
  }

  if (tag_equal(tag, "PARTNUM*")) {
    get_info(&tag[8], sizeof(char), 32, pv_handle);
    get_info(&idisc, sizeof(int),    1, pv_handle);
    if ((idisc < 0) || (idisc >= ndisc)) {
      fprintf(stderr, "\n%40s\n", tag);
      fprintf(stderr, "ERROR - Discipline Index Out of Range: %d!\n", idisc);
      exit(1);
    }
    get_info(&index, sizeof(int), 1, pv_handle);
    if ((index < 0) || (index >= disc[idisc].numpart)) {
      fprintf(stderr, "\n%40s\n", tag);
      fprintf(stderr, "ERROR - Partition Index Out of Range: %d!\n", idisc);
      exit(1);
    }
    part = disc[idisc].part;
    for (j = 0; j < 32; j++) part[index].title[j] = tag[8+j];
    printf("%40s  disc # = %d ipart = %d\n", tag, idisc, index);
    get_info(&index, sizeof(int), 1, pv_handle);
    get_info(&index, sizeof(int), 1, pv_handle);
    for (j = 0; j < 16; j++)
      get_info(&xmat[j], sizeof(float), 1, pv_handle);
    goto head;
  }

  if (tag_equal(tag, "SCALAR**")) {
    get_info(&tag[8], sizeof(char), 32, pv_handle);
    get_info(&idisc, sizeof(int),    1, pv_handle);
    if ((idisc < 0) || (idisc >= ndisc)) {
      fprintf(stderr, "\n%40s\n", tag);
      fprintf(stderr, "ERROR - Discipline Index Out of Range: %d!\n", idisc);
      exit(1);
    }
    get_info(&index, sizeof(int),    1, pv_handle);
    field = disc[idisc].field;
    for (i = 0; i < disc[idisc].nfield; i++) if (field[i].index == index) {
      fprintf(stderr, "\n%40s\n", tag);
      fprintf(stderr, "ERROR - Field Index Already Defined: %d!\n", index);
      exit(1);
    }
    i = disc[idisc].nfield;
    disc[idisc].nfield++;
    if (disc[idisc].nfield >= disc[idisc].mfield) {
      disc[idisc].mfield *= 2;
      field = (Field *) realloc(disc[idisc].field, disc->mfield*sizeof(Field));
      if (field == NULL) {
        fprintf(stderr, "\n%40s\n", tag);
        fprintf(stderr, "ERROR in realloc!");
        exit(1);
      }
      disc[idisc].field = field;
    }
    field[i].index = index;
    for (j = 0; j < 32; j++) field[i].title[j] = tag[8+j];
    get_info(field[i].flims, sizeof(float), 2, pv_handle);
    field[i].slen = 1;
    printf("%40s\n        discipline: %d  index: %d  lims: %f %f \n", tag, 
            idisc, index, field[i].flims[0], field[i].flims[1]);
    goto head;
  }

  if (tag_equal(tag, "VECTOR**")) {
    get_info(&tag[8], sizeof(char), 32, pv_handle);
    get_info(&idisc, sizeof(int),    1, pv_handle);
    if ((idisc < 0) || (idisc >= ndisc)) {
      fprintf(stderr, "\n%40s\n", tag);
      fprintf(stderr, "ERROR - Discipline Index Out of Range: %d!\n", idisc);
      exit(1);
    }
    get_info(&index, sizeof(int),    1, pv_handle);
    field = disc[idisc].field;
    for (i = 0; i < disc[idisc].nfield; i++) if (field[i].index == index) {
      fprintf(stderr, "\n%40s\n", tag);
      fprintf(stderr, "ERROR - Field Index Already Defined: %d!\n", index);
      exit(1);
    }
    i = disc[idisc].nfield;
    disc[idisc].nfield++;
    if (disc[idisc].nfield >= disc[idisc].mfield) {
      disc[idisc].mfield *= 2;
      field = (Field *) realloc(disc[idisc].field, disc->mfield*sizeof(Field));
      if (field == NULL) {
        fprintf(stderr, "\n%40s\n", tag);
        fprintf(stderr, "ERROR in realloc!");
        exit(1);
      }
      disc[idisc].field = field;
    }
    field[i].index = index;
    for (j = 0; j < 32; j++) field[i].title[j] = tag[8+j];
    get_info(field[i].flims, sizeof(float), 1, pv_handle);
    field[i].slen = 3;
    printf("%40s\n        discipline: %d  index: %d  scale: %f\n", tag, 
           idisc, index, field[i].flims[0]);
    goto head;
  }

  if (tag_equal(tag, "TENSOR**")) {
    get_info(&tag[8], sizeof(char), 32, pv_handle);
    get_info(&idisc, sizeof(int),    1, pv_handle);
    if ((idisc < 0) || (idisc >= ndisc)) {
      fprintf(stderr, "\n%40s\n", tag);
      fprintf(stderr, "ERROR - Discipline Index Out of Range: %d!\n", idisc);
      exit(1);
    }
    get_info(&index, sizeof(int),    1, pv_handle);
    field = disc[idisc].field;
    for (i = 0; i < disc[idisc].nfield; i++) if (field[i].index == index) {
      fprintf(stderr, "\n%40s\n", tag);
      fprintf(stderr, "ERROR - Field Index Already Defined: %d!\n", index);
      exit(1);
    }
    i = disc[idisc].nfield;
    disc[idisc].nfield++;
    if (disc[idisc].nfield >= disc[idisc].mfield) {
      disc[idisc].mfield *= 2;
      field = (Field *) realloc(disc[idisc].field, disc->mfield*sizeof(Field));
      if (field == NULL) {
        fprintf(stderr, "\n%40s\n", tag);
        fprintf(stderr, "ERROR in realloc!");
        exit(1);
      }
      disc[idisc].field = field;
    }
    field[i].index = index;
    for (j = 0; j < 32; j++) field[i].title[j] = tag[8+j];
    field[i].slen = 9;
    printf("%40s\n        discipline: %d  index: %d \n", tag, idisc, index);
    goto head;
  }

  if (tag_equal(tag, "STATE*V*")) {
    get_info(&tag[8], sizeof(char), 32, pv_handle);
    get_info(&idisc, sizeof(int),    1, pv_handle);
    if ((idisc < 0) || (idisc >= ndisc)) {
      fprintf(stderr, "\n%40s\n", tag);
      fprintf(stderr, "ERROR - Discipline Index Out of Range: %d!\n", idisc);
      exit(1);
    }
    for (j = 0; j < 32; j++) disc[idisc].stitle[j] = tag[8+j];
    get_info(&disc[idisc].slen, sizeof(int), 1, pv_handle);
    printf("%40s\n        discipline: %d  rank: %d\n", tag, idisc,
           disc[idisc].slen);
    goto head;
  }

  if (tag_equal(tag, "SURFACE*")) {
    get_info(&tag[8], sizeof(char), 32, pv_handle);
    get_info(&idisc, sizeof(int),    1, pv_handle);
    if ((idisc < 0) || (idisc >= ndisc)) {
      fprintf(stderr, "\n%40s\n", tag);
      fprintf(stderr, "ERROR - Discipline Index Out of Range: %d!\n", idisc);
      exit(1);
    }
    get_info(&index, sizeof(int),    1, pv_handle);
    surf = disc[idisc].surf;
    for (i = 0; i < disc[idisc].nsurf; i++) if (surf[i].index == index) {
      fprintf(stderr, "\n%40s\n", tag);
      fprintf(stderr, "ERROR - Surface Index Already Defined: %d!\n", index);
      exit(1);
    }
    i = disc[idisc].nsurf;
    disc[idisc].nsurf++;
    if (disc[idisc].nsurf >= disc[idisc].msurf) {
      disc[idisc].msurf *= 2;
      surf = (Surf *) realloc(disc[idisc].surf, disc->msurf*sizeof(Surf));
      if (surf == NULL) {
        fprintf(stderr, "\n%40s\n", tag);
        fprintf(stderr, "ERROR in realloc!");
        exit(1);
      }
      disc[idisc].surf = surf;
    }
    surf[i].index = index;
    for (j = 0; j < 32; j++) surf[i].title[j] = tag[8+j];
    printf("%40s discipline: %d  index: %d\n", tag, idisc, index);
    goto head;
  }

  if ((tag_equal(tag, "ENTITY**")) || (tag_equal(tag, "SENTITY*"))) {
    get_info(&tag[8], sizeof(char), 32, pv_handle);
    get_info(&index, sizeof(int),    1, pv_handle);
    for (i = 0; i < nentity; i++) if (entity[i].index == index) {
      fprintf(stderr, "\n%40s\n", tag);
      fprintf(stderr, "ERROR - Entity Index Already Defined: %d!\n", index);
      exit(1);
    }
    i = nentity;
    nentity++;
    if (nentity >= mentity) {
      mentity *= 2;
      entity = (Entity *) realloc(entity, mentity*sizeof(Entity));
      if (entity == NULL) {
        fprintf(stderr, "\n%40s\n", tag);
        fprintf(stderr, "ERROR in realloc!");
        exit(1);
      }
    }
    entity[i].index  = index;
    entity[i].factor = 1;
    for (j = 0; j < 32; j++) entity[i].title[j] = tag[8+j];
    get_info(&entity[i].nint,  sizeof(int), 1, pv_handle);
    get_info(&entity[i].nreal, sizeof(int), 1, pv_handle);
    get_info(&entity[i].nsubs, sizeof(int), 1, pv_handle);
    entity[i].sub = (SubEntity *) malloc(entity[i].nsubs*sizeof(SubEntity));
    sub = entity[i].sub;
    for (j = 0; j < entity[i].nsubs; j++) {
      sub[j].ranki = 0;
      sub[j].rankr = 0;
    }
    if (tag_equal(tag, "SENTITY*")) {
      get_info(&entity[i].factor, sizeof(int), 1, pv_handle);
      if (entity[i].factor < 1) {
        fprintf(stderr, "\n%40s factor: %d\n", tag, entity[i].factor);
        fprintf(stderr, "ERROR - Special Entity factor too small!\n");
        exit(1);
      }
      printf("%40s index: %d nsubs: %d factor: %d\n", 
             tag, index, entity[i].nsubs, entity[i].factor);
    } else {
      printf("%40s index: %d nsubs: %d\n", tag, index, entity[i].nsubs);
    }
    goto head;
  }

  if (tag_equal(tag, "DATADEF*")) {
    get_info(&tag[8], sizeof(char), 32, pv_handle);
    get_info(&ient,   sizeof(int),   1, pv_handle);
    for (i = 0; i < nentity; i++) if (entity[i].index == ient) {
      get_info(&index, sizeof(int), 1, pv_handle);
      if ((index < 0 ) || (index >= entity[i].nsubs)) {
        fprintf(stderr, "\n%40s\n", tag);
        fprintf(stderr, "ERROR - SubEntity index out of range: %d!\n", index);
        exit(1);
      }
      sub = entity[i].sub;
      if (sub[index].ranki+sub[index].rankr != 0) {
        fprintf(stderr, "\n%40s\n", tag);
        fprintf(stderr, "ERROR - SubEntity already filled: %d!\n", index);
        exit(1);
      }
      get_info(&sub[index].ranki, sizeof(int), 1, pv_handle);
      get_info(&sub[index].rankr, sizeof(int), 1, pv_handle);
      if (sub[index].ranki+sub[index].rankr == 0) {
        fprintf(stderr, "\n%40s\n", tag);
        fprintf(stderr, "ERROR - Null SubEntity: %d!\n", index);
        exit(1);
      }
      for (j = 0; j < 32; j++) sub[index].title[j] = tag[8+j];
      printf("%40s Entity: %d  index: %d\n", tag, ient, index);
      goto head;
    }
    fprintf(stderr, "\n%40s\n", tag);
    fprintf(stderr, "ERROR - Entity Index Not Found: %d!\n", ient);
    exit(1);
  }

  tag[8] = 0;
  fprintf(stderr, "\nERROR - Illegal tag: %s!\n", tag);
  exit(1);

/* get each time frame */

body:
  offset = ftell(pv_handle);
  get_info(tag, sizeof(char), 8, pv_handle);

  if (tag_equal(tag, "E*FILE**")) {
    tag[8] = 0;
    printf("%8s # of time frames: %d  saved bytes = %d\n",
           tag, nframe, savedbytes);
    exit(0);
  }
  check_tag(tag, "S*FRAME*");
  tag[8] = 0;
  get_info(&time, sizeof(float), 1, pv_handle);
  printf("%8s time: %f\n", tag, time);
  i = nframe;
  nframe++;
  if (nframe >= mframe) {
    mframe *= 2;
    frame = (Frame *) realloc(frame, mframe*sizeof(Frame));
    if (frame == NULL) {
      fprintf(stderr, "\nERROR in realloc!");
      exit(1);
    }
  }
  frame[i].time = time;
  frame[i].offset = offset;

/* get each Entity block */

inframe:
  get_info(tag, sizeof(char), 8, pv_handle);

  if (tag_equal(tag, "E*FRAME*")) {
    get_info(&dummy, sizeof(float), 1, pv_handle);
    if (dummy != time) {
      fprintf(stderr, "\n%8s time: %f\n", tag, dummy);
      fprintf(stderr, "ERROR - Time Marker Mismatch!\n");
      exit(1);
    }
    tag[8] = 0;
    printf("%8s time: %f\n", tag, dummy);
    goto body;
  }

  if (tag_equal(tag, "S*ENTIT*")) {
    get_info(&idisc,  sizeof(int),   1, pv_handle);
    get_info(&ient,   sizeof(int),   1, pv_handle);
    get_info(&index,  sizeof(int),   1, pv_handle);
    get_info(&tag[8], sizeof(char), 32, pv_handle);
    get_info(&i,      sizeof(int),   1, pv_handle);
    if ((idisc < 0) || (idisc >= ndisc)) {
      fprintf(stderr, "\n%40s\n", tag);
      fprintf(stderr, "ERROR - Discipline Index Out of Range: %d!\n", idisc);
      exit(1);
    }
    for (eindex = 0; eindex < nentity; eindex++) 
      if (entity[eindex].index == ient) {
        for (k = 0; k < entity[eindex].nint; k++) 
          get_info(&j, sizeof(int), 1, pv_handle);
        for (k = 0; k < entity[eindex].nreal; k++) 
          get_info(&dummy, sizeof(float), 1, pv_handle);
        if (entity[eindex].factor != 1) {
          offset = entity[eindex].factor*disc[idisc].numpart*sizeof(int);
          fseek(pv_handle, offset, SEEK_CUR);
        }
        printf("%8s discipline: %d entity: %d instance: %d\n",
               tag, idisc, ient, index);
        nhit = 0;
        goto inentity;
      }
    fprintf(stderr, "\n%40s\n", tag);
    fprintf(stderr, "ERROR - Entity Index Not Found: %d!\n", ient);
    exit(1);
  }

  tag[8] = 0;
  fprintf(stderr, "\nERROR - Illegal tag: %s!\n", tag);
  exit(1);

/* get data from within the Entity */

inentity:
  get_info(tag, sizeof(char), 8, pv_handle);

  if (tag_equal(tag, "E*ENTIT*")) {
    tag[8] = 0;
    get_info(&i, sizeof(int), 1, pv_handle);
    get_info(&j, sizeof(int), 1, pv_handle);
    get_info(&k, sizeof(int), 1, pv_handle);
    if ((i != idisc) || (j != ient) || (k != index)) {
      fprintf(stderr, "\n%8s discipline: %d entity: %d instance: %d\n",
                      tag, i, j, k);
      fprintf(stderr, "ERROR - Entity Marker Mismatch!\n");
      exit(1);
    }
    printf("%8s # of field and subs: %d\n", tag, nhit);
    goto inframe;
  }

  if (tag_equal(tag, "SUBDATA*")) {
    get_info(&ipart, sizeof(int), 1, pv_handle);
    get_info(&isub,  sizeof(int), 1, pv_handle);
    get_info(&len,   sizeof(int), 1, pv_handle);
    if ((isub < 0) || (isub >= entity[eindex].nsubs)) {
      fprintf(stderr, "\n%8s sub-index: %d\n", tag, isub);
      fprintf(stderr, "ERROR - Sub Index Out of Range!\n");
      exit(1);
    }
    if ((ipart < 0) || (ipart >= entity[eindex].factor*disc[idisc].numpart)) {
      fprintf(stderr, "\n%8s part: %d\n", tag, ipart);
      fprintf(stderr, "ERROR - Partition Index Out of Range: %d!\n", idisc);
      exit(1);
    }
    sub = entity[eindex].sub;
    if (sub[isub].ranki != 0) {
      offset = len*sub[isub].ranki*sizeof(int);
    } else {
      offset = len*sub[isub].rankr*sizeof(float);
    }
    fseek(pv_handle, offset, SEEK_CUR);
    nhit++;
    goto inentity;
  }

  if (tag_equal(tag, "PREVDAT*")) {
    get_info(&ipart, sizeof(int), 1, pv_handle);
    get_info(&isub,  sizeof(int), 1, pv_handle);
    get_info(&len,   sizeof(int), 1, pv_handle);
    save_pos = ftell(pv_handle);
    offset = len;
    fseek(pv_handle, offset,   SEEK_SET);
    get_info(&tag[8], sizeof(char), 8, pv_handle);
    get_info(&j,      sizeof(int), 1, pv_handle);
    get_info(&k,      sizeof(int), 1, pv_handle);
    get_info(&len,    sizeof(int), 1, pv_handle);
    if (!tag_equal(&tag[8], "SUBDATA*")) {
      fprintf(stderr, "\nERROR - PREVDAT* points to: %8s\n", &tag[8]);
      exit(1);
    }
    if ((j != ipart) || (k != isub)) {
      fprintf(stderr, "\nPREVDAT* %d %d points to: %d %d\n",
                      ipart, isub, j, k);
      fprintf(stderr, "ERROR - PREVDAT* partition/sub-index Mismatch!\n");
      exit(1);
    }
    sub = entity[eindex].sub;
    if (sub[isub].ranki != 0) {
      offset = len*sub[isub].ranki*sizeof(int);
    } else {
      offset = len*sub[isub].rankr*sizeof(float);
    }
    savedbytes += offset;
    fseek(pv_handle, save_pos, SEEK_SET);
    nhit++;
    goto inentity;
  }

  if (tag_equal(tag, "FIELD***")) {
    get_info(&ipart, sizeof(int), 1, pv_handle);
    get_info(&isub,  sizeof(int), 1, pv_handle);
    get_info(&len,   sizeof(int), 1, pv_handle);
    if ((ipart < 0) || (ipart >= entity[eindex].factor*disc[idisc].numpart)) {
      fprintf(stderr, "\n%8s part: %d\n", tag, ipart);
      fprintf(stderr, "ERROR - Partition Index Out of Range: %d!\n", idisc);
      exit(1);
    }
    if (isub == 0) {
      if(disc[idisc].slen == 0) {
        fprintf(stderr, "\n%8s part: %d\n", tag, ipart);
        fprintf(stderr, "ERROR - No State Vector Defined: %d!\n", idisc);
        exit(1);
      }
      offset = len*disc[idisc].slen*sizeof(float);
      fseek(pv_handle, offset, SEEK_CUR);
      nhit++;
      goto inentity;
    }
    field = disc[idisc].field;
    for (i = 0; i < disc[idisc].nfield; i++) if (field[i].index == isub) {
      offset = len*field[i].slen*sizeof(float);
      fseek(pv_handle, offset, SEEK_CUR);
      nhit++;
      goto inentity;
    }
    fprintf(stderr, "\n%8s field-index: %d\n", tag, isub);
    fprintf(stderr, "ERROR - Field Index Out of Range!\n");
    exit(1);
  }

  if (tag_equal(tag, "PREVFLD*")) {
    get_info(&ipart, sizeof(int), 1, pv_handle);
    get_info(&isub,  sizeof(int), 1, pv_handle);
    get_info(&len,   sizeof(int), 1, pv_handle);
    save_pos = ftell(pv_handle);
    offset = len;
    fseek(pv_handle, offset,   SEEK_SET);
    get_info(&tag[8], sizeof(char), 8, pv_handle);
    get_info(&j,      sizeof(int), 1, pv_handle);
    get_info(&k,      sizeof(int), 1, pv_handle);
    get_info(&len,    sizeof(int), 1, pv_handle);
    if (!tag_equal(&tag[8], "FIELD***")) {
      fprintf(stderr, "\nERROR - PREVFLD* points to: %8s\n", &tag[8]);
      exit(1);
    }
    if ((j != ipart) || (k != isub)) {
      fprintf(stderr, "\nPREVFLD* %d %d points to: %d %d\n",
                      ipart, isub, j, k);
      fprintf(stderr, "ERROR - PREVFLD* partition/sub-index Mismatch!\n");
      exit(1);
    }
    if (isub == 0) {
      if(disc[idisc].slen == 0) {
        fprintf(stderr, "\n%8s part: %d\n", tag, ipart);
        fprintf(stderr, "ERROR - No State Vector Defined: %d!\n", idisc);
        exit(1);
      }
      offset = len*disc[idisc].slen*sizeof(float);
      savedbytes += offset;
      fseek(pv_handle, save_pos, SEEK_SET);
      nhit++;
      goto inentity;
    }
    field = disc[idisc].field;
    for (i = 0; i < disc[idisc].nfield; i++) if (field[i].index == isub) {
      offset = len*field[i].slen*sizeof(float);
      savedbytes += offset;
      fseek(pv_handle, save_pos, SEEK_SET);
      nhit++;
      goto inentity;
    }
    fprintf(stderr, "\n%8s field-index: %d\n", tag, isub);
    fprintf(stderr, "ERROR - Field Index Out of Range!\n");
    exit(1);
  }

  tag[8] = 0;
  fprintf(stderr, "\nERROR - Illegal tag: %s!\n", tag);
  exit(1);

}
