MODULE typedefs
  IMPLICIT NONE

!#######################################################################
! TYPE AOP
!  The aop type generates an array of pointers for use in various and
! sundry dirty activities.
!#######################################################################
  TYPE aop
    REAL,POINTER,DIMENSION(:) :: v
  END TYPE aop

!#######################################################################
! TYPE GRID
!   The grid type contains pointers for all grid related variables
! which may be allocated.  This is set up for the eventuality of moving
! mesh (transient) and varying structure.
!   The logical gridflags can be used to determine what variables are
! currently stored in memory.
!#######################################################################
  TYPE grid
    REAL,POINTER,DIMENSION(:) :: x     ! X-location
    REAL,POINTER,DIMENSION(:) :: y     ! Y-location
    REAL,POINTER,DIMENSION(:) :: z     ! Z-location
    REAL,POINTER,DIMENSION(:) :: em    ! M-location
    REAL,POINTER,DIMENSION(:) :: r     ! Radius (cylindrical/spherical)
    REAL,POINTER,DIMENSION(:) :: t     ! Theta (cylindrical/spherical)
    REAL,POINTER,DIMENSION(:) :: t0    ! Theta at time 0
    REAL,POINTER,DIMENSION(:) :: ph    ! Phi (spherical)
    REAL,POINTER,DIMENSION(:) :: i     ! Global I-location
    REAL,POINTER,DIMENSION(:) :: j     ! Global J-location
    REAL,POINTER,DIMENSION(:) :: k     ! Global K-location
    REAL,POINTER,DIMENSION(:) :: w     ! Wall distance

    INTEGER,POINTER,DIMENSION(:) :: iblank ! I-Blanking array

    INTEGER :: nblks
    INTEGER,POINTER,DIMENSION(:) :: li,lj,lk ! IJK values on the local processor
    INTEGER,POINTER,DIMENSION(:) :: gis,gjs,gks ! Beginning IJK values for the global system
    INTEGER,POINTER,DIMENSION(:) :: mvmax       ! Number of points per block
    INTEGER :: tpts ! Total number of unstructured points

    INTEGER :: mle,mte,mhb,mtp ! Blade geometry data
    INTEGER :: nbld            ! Number of blades in blade row

    INTEGER :: ntet, nhex, npris, npyr, tcells

  END TYPE grid

  TYPE dmap
    INTEGER :: id
    INTEGER :: is,ie,js,je,ks,ke
    INTEGER :: blkb
    INTEGER :: d1,d1s,d1e
    INTEGER :: d2,d2s,d2e
    INTEGER :: d3,d3s,d3e
    INTEGER :: dir1,dir2
    INTEGER :: lor1,lor2
    INTEGER :: bc
  END TYPE dmap

!#######################################################################
! TYPE FLOW
!   This type contains pointers to all of the flow variables which are
! calculated at any point in time.  As a result, variables will be
! easier to access and will not need to be recalculated unless memory
! requirements become an issue.
!   Based on the value of the logical array, flowflag, it careful
! programming will allow appropriate allocation and deallocation
! when memory is needed.
!#######################################################################
  TYPE :: flow
! Vstage variables that don't get deleted...
    REAL,POINTER,DIMENSION(:) :: p     ! Pressure
    REAL,POINTER,DIMENSION(:) :: r     ! Density
    REAL,POINTER,DIMENSION(:) :: vz    ! z-velocity
    REAL,POINTER,DIMENSION(:) :: vr    ! r-velocity
    REAL,POINTER,DIMENSION(:) :: vt    ! t-velocity
    REAL,POINTER,DIMENSION(:) :: e     ! Energy
    REAL,POINTER,DIMENSION(:) :: mut   ! Turbulent viscosity
    REAL,POINTER,DIMENSION(:) :: k     ! TKE
    REAL,POINTER,DIMENSION(:) :: eps   ! Epsilon

! Real gas
    REAL,POINTER,DIMENSION(:) :: g     ! Gamma
    REAL,POINTER,DIMENSION(:) :: rgas  ! Real gas R, single value for now
    REAL,POINTER,DIMENSION(:) :: reyn  ! Reynolds number
    REAL,POINTER,DIMENSION(:) :: dens  ! Dimensional density

! Pressures
    REAL,POINTER,DIMENSION(:) :: ptrel
    REAL,POINTER,DIMENSION(:) :: ptabs

! Angles
    REAL,POINTER,DIMENSION(:) :: alphaz
    REAL,POINTER,DIMENSION(:) :: alpham
    REAL,POINTER,DIMENSION(:) :: betaz
    REAL,POINTER,DIMENSION(:) :: betam
    REAL,POINTER,DIMENSION(:) :: phi

! Mach numbers
    REAL,POINTER,DIMENSION(:) :: machthabs
    REAL,POINTER,DIMENSION(:) :: machthrel
    REAL,POINTER,DIMENSION(:) :: machr
    REAL,POINTER,DIMENSION(:) :: machtabs
    REAL,POINTER,DIMENSION(:) :: machtrel
    REAL,POINTER,DIMENSION(:) :: machxabs
    REAL,POINTER,DIMENSION(:) :: machxrel
    REAL,POINTER,DIMENSION(:) :: machyabs
    REAL,POINTER,DIMENSION(:) :: machyrel
    REAL,POINTER,DIMENSION(:) :: machz
    REAL,POINTER,DIMENSION(:) :: machideal

! Rho-V
    REAL,POINTER,DIMENSION(:) :: rhovthabs
    REAL,POINTER,DIMENSION(:) :: rhovthrel
    REAL,POINTER,DIMENSION(:) :: rhovr
    REAL,POINTER,DIMENSION(:) :: rhovtabs
    REAL,POINTER,DIMENSION(:) :: rhovtrel
    REAL,POINTER,DIMENSION(:) :: rhovxabs
    REAL,POINTER,DIMENSION(:) :: rhovxrel
    REAL,POINTER,DIMENSION(:) :: rhovyabs
    REAL,POINTER,DIMENSION(:) :: rhovyrel
    REAL,POINTER,DIMENSION(:) :: rhovz

! Enthalpy
    REAL,POINTER,DIMENSION(:) :: h
    REAL,POINTER,DIMENSION(:) :: htabs
    REAL,POINTER,DIMENSION(:) :: htrel

! Rothalpy
    REAL,POINTER,DIMENSION(:) :: roth
    REAL,POINTER,DIMENSION(:) :: rothcp

! Entropy
    REAL,POINTER,DIMENSION(:) :: sabs ! Abs.
    REAL,POINTER,DIMENSION(:) :: srat ! Ratio

! Temperature
    REAL,POINTER,DIMENSION(:) :: temp     ! Static
    REAL,POINTER,DIMENSION(:) :: temptabs ! Tot. Abs.
    REAL,POINTER,DIMENSION(:) :: temptrel ! Tot. Rel.

! Velocities
    REAL,POINTER,DIMENSION(:) :: vx
    REAL,POINTER,DIMENSION(:) :: vxabs
    REAL,POINTER,DIMENSION(:) :: vy
    REAL,POINTER,DIMENSION(:) :: vyabs
    REAL,POINTER,DIMENSION(:) :: vtabs  ! Vtot Abs.
    REAL,POINTER,DIMENSION(:) :: vtrel  ! Vtot Rel.
    REAL,POINTER,DIMENSION(:) :: vthrel ! Vtheta rel.

! VizMath
    REAL,POINTER,DIMENSION(:,:) :: vmath     ! X,Y,Z
    REAL,POINTER,DIMENSION(:)   :: vmaths    ! For single dimensioned
    REAL,POINTER,DIMENSION(:)   :: vmathr    ! Radial
    REAL,POINTER,DIMENSION(:)   :: vmatht    ! Tangential
    REAL,POINTER,DIMENSION(:)   :: vmathsabs ! Abs. Streamwise
    REAL,POINTER,DIMENSION(:)   :: vmathsrel ! Rel. Streamwise
    REAL,POINTER,DIMENSION(:)   :: vmathtabs ! Abs. Total 
    REAL,POINTER,DIMENSION(:)   :: vmathtrel ! Rel. Total
    REAL,POINTER,DIMENSION(:)   :: vmathzabs ! Abs. Z

! Vorticities
    REAL,POINTER,DIMENSION(:) :: vortr    ! Radial-Vorticity
    REAL,POINTER,DIMENSION(:) :: vortt    ! Theta-Vorticity
    REAL,POINTER,DIMENSION(:) :: vortz    ! Rel. Z-Vorticity
    REAL,POINTER,DIMENSION(:) :: vortsabs ! Abs. Str. Vort.
    REAL,POINTER,DIMENSION(:) :: vortsrel ! Rel. Str. Vort.
    REAL,POINTER,DIMENSION(:) :: vorttabs ! Abs. Tot. Vort.
    REAL,POINTER,DIMENSION(:) :: vorttrel ! Rel. Tot. Vort.
    REAL,POINTER,DIMENSION(:) :: vortx    ! X-Vorticity
    REAL,POINTER,DIMENSION(:) :: vorty    ! Y-Vorticity
    REAL,POINTER,DIMENSION(:) :: vortzabs ! Abs. Z

! Turbulence
    REAL,POINTER,DIMENSION(:) :: turbint   ! Turb intensity
    REAL,POINTER,DIMENSION(:) :: turbrat   ! Ratio mut/mu
    REAL,POINTER,DIMENSION(:) :: turbtvisc ! Total viscosity mu+mut
    REAL,POINTER,DIMENSION(:) :: yplus

! Time-average
    REAL,POINTER,DIMENSION(:) :: tmavg    ! Time average of quantity
    REAL,POINTER,DIMENSION(:) :: tmdif    ! Time diff cur-avg
    REAL,POINTER,DIMENSION(:) :: tmmin    ! Time min, pointwise
    REAL,POINTER,DIMENSION(:) :: tmmax    ! Time max, pointwise
    REAL,POINTER,DIMENSION(:) :: tmrms    ! RMS of cur-avg
    REAL,POINTER,DIMENSION(:) :: tmfft    ! FFT

    INTEGER :: ncyc
    REAL :: tyme, fsmach, re, alpha

  END TYPE flow

!#######################################################################
! TYPE VCS
!  The vcs type is for the storage of vortex core data which is
! generated through the FX libraries
!#######################################################################
  TYPE vortexcore
    REAL,POINTER,DIMENSION(:,:) :: vcoor
    REAL,POINTER,DIMENSION(:) :: stren
    INTEGER,POINTER,DIMENSION(:) :: segs
    INTEGER nseg
  END TYPE vortexcore

  TYPE sepattline
    INTEGER, POINTER, DIMENSION(:) :: sepind
    INTEGER, POINTER, DIMENSION(:) :: attind
    REAL, POINTER, DIMENSION(:,:) :: sepxyz
    REAL, POINTER, DIMENSION(:,:) :: attxyz
    INTEGER :: calc
    INTEGER*BSIZE :: attadd, sepadd
  END TYPE sepattline

  TYPE shocksurf
    REAL, POINTER, DIMENSION(:,:) :: pts
    INTEGER :: npts
  END TYPE shocksurf

END MODULE typedefs
