
MODULE MOD_P3DRay

USE MOD_GRIDVARS

IMPLICIT NONE

LOGICAL :: DEBUGPRINT = .TRUE.

CONTAINS

!**************************************************************
SUBROUTINE CrossP(pc,p1,p2,p3,edist)
!**************************************************************
! cross product of three points, return new extruded point, p2 is the centerpoint
REAL(kind=FP), DIMENSION(3), INTENT(in) :: p1, p2, p3
REAL(kind=FP), DIMENSION(3), INTENT(out) :: pc
REAL(kind=FP), INTENT(in) :: edist
REAL(kind=FP), DIMENSION(3) :: v1, v2, vc
REAL(kind=FP) :: vlen

v1 = p2 - p1
v2 = p2 - p3
vc(1) = v1(2)*v2(3) - v1(3)*v2(2)
vc(2) = v1(3)*v2(1) - v1(1)*v2(3)
vc(3) = v1(1)*v2(2) - v1(2)*v2(1)
vlen = SQRT(SUM(vc**2))
pc = p1 + edist * vc / vlen

END SUBROUTINE CrossP

!**************************************************************
SUBROUTINE InterpGrid(x,ix,ieng,jeng,keng,ireal,jreal,kreal,ifl,jfl,kfl,ndm)
!**************************************************************
IMPLICIT NONE

LOGICAL, INTENT(in) :: ifl, jfl, kfl
INTEGER, INTENT(in) :: ndm
INTEGER, POINTER, DIMENSION(:) :: ieng, jeng, keng, ireal, jreal, kreal
REAL(kind=FP), DIMENSION(:,:), POINTER :: x
REAL(kind=DP), DIMENSION(:,:), POINTER :: ix

INTEGER :: ng = 1
INTEGER, DIMENSION(3) :: readorder = (/3,2,1/)
INTEGER :: i, j, k, n, ti1, ti2, ti3
INTEGER :: neng, ierr

REAL(kind=FP), DIMENSION(3) :: locmin, locmax, locdiff, p1, p2, p3, pc
REAL(kind=FP) :: edist

neng = ieng(ng) * jeng(ng) * keng(ng)
ALLOCATE( x(ndm, neng), stat=ierr )
CALL CheckAllocationError(ierr)

if( .NOT.ifl.AND..NOT.jfl.AND..NOT.kfl ) then
! data is not planar, roll read data into main storage and exit
!    WRITE(*,*) 'No Extrusion Needed'
    DO n=1,ndm
        x(n,:) = ix(readorder(n),:)
    ENDDO
    DEALLOCATE( ix )
    return
endif

! data is planar, sparsely write data into main storage
DO n=1,ndm
    DO k=1,kreal(ng)
        DO j=1,jreal(ng)
            DO i=1,ireal(ng)
                x(n, i + ieng(ng)*( (j-1) + jeng(ng)*(k-1) ) ) &
                    = ix(readorder(n), i + ireal(ng)*( (j-1) + jreal(ng)*(k-1) ) )
            ENDDO
        ENDDO
    ENDDO
ENDDO

! determine size of geometry, scale extrusion distance accordingly
! (.001*max_size)
locmin = MINVAL(ix,DIM=2)
locmax = MAXVAL(ix,DIM=2)
locdiff = locmax - locmin
edist = 0.0001*(SQRT(SUM(locdiff * locdiff)))
DEALLOCATE( ix )

! process and extrude along planar dimension, maintain rotation direction on
! cross product to maintain extrusion direction
if( ifl ) then
    ! I is planar, extrude normal to JxK
!    WRITE(*,*) 'Extruding On I-Plane'
    DO k=1,keng(ng)
!        WRITE(*,*) 'K=',k
        DO j=1,jeng(ng)
!            WRITE(*,*) '  J=',j
            ti2 = 1+ieng(ng)*( (j-1)+jeng(ng)*(k-1) )
            if ( (j.EQ.1) .AND. (k.EQ.keng(ng)) ) then
                ti1 = ti2 - ieng(ng)*jeng(ng) 
                ti3 = ti2 + ieng(ng)
            elseif ( (j.EQ.jeng(ng)) .AND. (k.EQ.1)) then
                ti1 = ti2 + ieng(ng)*jeng(ng)
                ti3 = ti2 - ieng(ng)
            elseif ( (j.EQ.jeng(ng)) .OR. (k.EQ.keng(ng)) ) then
                ti1 = ti2 - ieng(ng)
                ti3 = ti2 - ieng(ng)*jeng(ng)
            else
                ti1 = ti2 + ieng(ng)
                ti3 = ti2 + ieng(ng)*jeng(ng)
            endif
            CALL CrossP(x(:,ti2+1),x(:,ti1),x(:,ti2),x(:,ti3),edist)
        ENDDO
    ENDDO
elseif( jfl ) then
    ! J is planar, extrude normal to IxK
!    WRITE(*,*) 'Extruding On J-Plane'
    DO k=1,keng(ng)
!        WRITE(*,*) 'K=',k
        DO i=1,ieng(ng)
!            WRITE(*,*) '  I=',i,'    edist=',edist
            ti2 = i+ieng(ng)*( jeng(ng)*(k-1) )
            if ( (i.EQ.1) .AND. (k.EQ.keng(ng)) ) then
                ti1 = ti2 - ieng(ng)*jeng(ng) 
                ti3 = ti2 + 1
            elseif ( (i.EQ.ieng(ng)) .AND. (k.EQ.1) ) then
                ti1 = ti2 + ieng(ng)*jeng(ng)
                ti3 = ti2 - 1
            elseif ( (i.EQ.ieng(ng)) .OR. (k.EQ.keng(ng)) ) then
                ti1 = ti2 - 1
                ti3 = ti2 - ieng(ng)*jeng(ng)
            else
                ti1 = ti2 + 1
                ti3 = ti2 + ieng(ng)*jeng(ng)
            endif
            CALL CrossP(x(:,ti2+ieng(ng)),x(:,ti1),x(:,ti2),x(:,ti3),edist)
        ENDDO
    ENDDO
elseif( kfl ) then
    ! K is planar, extrude normal to IxJ
!    WRITE(*,*) 'Extruding On K-Plane'
    DO j=1,jeng(ng)
!        WRITE(*,*) 'J=',j
        DO i=1,ieng(ng)
!            WRITE(*,*) '  I=',i
            ti2 = i+ieng(ng)*( (j-1) )
            if ( (j.EQ.jeng(ng)) .AND. (i.EQ.1) ) then
                ti1 = ti2 - ieng(ng)
                ti3 = ti2 + 1
            elseif ( (j.EQ.1) .AND. (i.EQ.ieng(ng)) ) then
                ti1 = ti2 + ieng(ng)
                ti3 = ti2 - 1
            elseif ( (j.EQ.jeng(ng)) .OR. (i.EQ.ieng(ng)) ) then
                ti1 = ti2 - 1
                ti3 = ti2 - jeng(ng)
            else
                ti1 = ti2 + 1
                ti3 = ti2 + ieng(ng)
            endif
            CALL CrossP(x(:,ti2+ieng(ng)*jeng(ng)),x(:,ti1),x(:,ti2),x(:,ti3),edist)
        ENDDO
    ENDDO
endif

END SUBROUTINE InterpGrid

!**************************************************************
SUBROUTINE ReadP3DGrid(fname,grdformat,grddouble,x,ieng,jeng,keng)
!**************************************************************

IMPLICIT NONE

real (kind=FP), dimension(:,:), pointer :: x
real (kind=SP), dimension(:,:), pointer :: sx
real (kind=DP), dimension(:,:), pointer :: dx
character (len=chrlen), intent(in) :: fname

integer, dimension(:), pointer :: ieng, jeng, keng
integer, dimension(:), pointer :: ireal, jreal, kreal
integer, intent(in) :: grdformat
integer :: ierr, ng, npts, i, j, k, n, np, nreal, neng, counter

integer :: filepipe = 13
integer :: ngrd = 1
integer :: ndm = 3

logical, intent(in) :: grddouble
logical :: ifl = .FALSE., jfl = .FALSE., kfl = .FALSE.

open(filepipe, file = fname, form = 'unformatted', status = 'old')
READ(filepipe,ERR=101,END=101) !ngrd

ALLOCATE( ieng(ngrd), ireal(ngrd), &
          jeng(ngrd), jreal(ngrd), &
          keng(ngrd), kreal(ngrd), &
          stat = ierr )
CALL CheckAllocationError(ierr)

READ(filepipe,ERR=101,END=101) (ireal(ng), jreal(ng), kreal(ng), ng=1,ngrd)
ng = 1
WRITE(*,*) 'Dimensions are: I=',ireal(ng),' J=', jreal(ng),' K=', kreal(ng)
!100 format(A,i5.1,A,3i5.1)

ieng(ng) = ireal(ng)
jeng(ng) = jreal(ng)
keng(ng) = kreal(ng)

if ( ireal(ng) .EQ. 1 ) then
    ieng(ng) = 2
    ifl = .TRUE.
!    WRITE(*,*) 'I dimension is flat'
endif
if ( jreal(ng) .EQ. 1 ) then
    jeng(ng) = 2
    jfl = .TRUE.
!    WRITE(*,*) 'J dimension is flat'
endif
if ( kreal(ng) .EQ. 1 ) then
    keng(ng) = 2
    kfl = .TRUE.
!    WRITE(*,*) 'K dimension is flat'
endif

nreal = ireal(ng)*jreal(ng)*kreal(ng)
neng = ieng(ng)*jeng(ng)*keng(ng)
ALLOCATE( dx(ndm, neng), stat=ierr )
CALL CheckAllocationError(ierr)

if ( grddouble ) then
    READ(filepipe,ERR=101,END=101) ((dx(n,i), i=1,nreal), n=1,ndm)
else
    ALLOCATE( sx(ndm, neng), stat=ierr )
    CALL CheckAllocationError(ierr)
    READ(filepipe,ERR=101,END=101) ((sx(n,i), i=1,nreal), n=1,ndm)
    DO n=1,nreal
        dx(:,n) = sx(:,n)
    ENDDO
    DEALLOCATE( sx )
endif

CALL InterpGrid(x,dx,ieng,jeng,keng,ireal,jreal,kreal,ifl,jfl,kfl,ndm)
DEALLOCATE( ireal, jreal, kreal )

return

101 continue
CALL DispFileReadError()

END SUBROUTINE ReadP3DGrid

!**************************************************************
SUBROUTINE ReadP3DFlow(fname,grdformat,grddouble,q,ieng,jeng,keng,ntime)
!**************************************************************
IMPLICIT NONE

real (kind=FP), pointer, dimension(:,:,:) :: q

real (kind=SP), dimension(:,:), allocatable :: sq
real (kind=DP), dimension(:,:), allocatable :: dq
character (len=chrlen), intent(in) :: fname

integer, dimension(:), pointer :: ieng, jeng, keng
integer, dimension(:), allocatable :: ireal, jreal, kreal
integer, intent(in) :: grdformat, ntime
integer :: ierr, neng, nreal, i, j, k, n, m, np

integer :: filepipe = 13
integer :: ngrd = 1
integer :: nvar = 5
integer :: ng = 1

logical, intent(in) :: grddouble
LOGICAL :: ifl = .FALSE., jfl = .FALSE., kfl = .FALSE.

! Assume that if length is 2, flow was extruded
neng = ieng(ngrd) * jeng(ngrd) * keng(ngrd)
ALLOCATE( ireal(ngrd), jreal(ngrd), kreal(ngrd), stat=ierr )
CALL CheckAllocationError(ierr)
DO n=1,ng
    ireal(n) = ieng(n)
    jreal(n) = jeng(n)
    kreal(n) = keng(n)
ENDDO
if( ieng(ng) .EQ. 2 ) then
    ireal(ng) = 1
    ifl = .TRUE.
elseif( jeng(ng) .EQ. 2 ) then
    jreal(ng) = 1
    jfl = .TRUE.
elseif( keng(ng) .EQ. 2 ) then
    kreal(ng) = 1
    kfl = .TRUE.
endif
nreal = ireal(ngrd) * jreal(ngrd) * kreal(ngrd)

allocate( q(ntime,nvar,neng), stat=ierr)
CALL CheckAllocationError(ierr)

open(filepipe, file = fname, form = 'unformatted', status = 'old')
do m=1,ntime
    if ( grddouble ) then
        ALLOCATE( dq(nvar, nreal), stat=ierr )
        CALL CheckAllocationError(ierr)
        READ(filepipe,ERR=101,END=101) ((dq(n,i), i=1,nreal), n=1,nvar)
        DO n=1,nvar
            DO k=1,kreal(ng)
                DO j=1,jreal(ng)
                    DO i=1,ireal(ng)
                        q( m, n, i + ieng(ng)*( (j-1) + jeng(ng)*(k-1) ) ) &
                            = dq(n, i + ireal(ng)*( (j-1) + jreal(ng)*(k-1) ) )
                    ENDDO
                ENDDO
            ENDDO
        ENDDO
        DEALLOCATE( dq )
    else
        ALLOCATE( sq(nvar, nreal), stat=ierr )
        CALL CheckAllocationError(ierr)
        READ(filepipe,ERR=101,END=101) ((sq(n,i), i=1,nreal), n=1,nvar)
        DO n=1,nvar
            DO k=1,kreal(ng)
                DO j=1,jreal(ng)
                    DO i=1,ireal(ng)
                        q( m, n, i + ieng(ng)*( (j-1) + jeng(ng)*(k-1) ) ) &
                            = sq(n, i + ireal(ng)*( (j-1) + jreal(ng)*(k-1) ) )
                    ENDDO
                ENDDO
            ENDDO
        ENDDO
        DEALLOCATE( sq )
    endif
enddo
101 continue
close(filepipe)
deallocate( ireal, jreal, kreal )
do m=1,ntime
    if( ifl ) then
        DO j=1,jeng(ng)
            DO k=1,keng(ng)
                q(m, :, 2 + ieng(ng)*( (j-1) + jeng(ng)*(k-1) ) ) &
                     = q(m, :, 1 + ieng(ng)*( (j-1) + jeng(ng)*(k-1) ) )
            ENDDO
        ENDDO
    elseif( jfl ) then
        DO i=1,ieng(ng)
            DO k=1,keng(ng)
                q(m, :, i + ieng(ng)*( 1 + jeng(ng)*(k-1) ) ) &
                     = q(m, :, i + ieng(ng)*( jeng(ng)*(k-1) ) )
            ENDDO
        ENDDO
    elseif( kfl ) then
        DO i=1,ieng(ng)
            DO j=1,jeng(ng)
                q(m, :, i + ieng(ng)*( (j-1) + jeng(ng) ) ) &
                     = q(m, :, i + ieng(ng)*( j-1 ) )
            ENDDO
        ENDDO
    endif
enddo


END SUBROUTINE ReadP3DFlow

!**************************************************************
SUBROUTINE ReadP3DTime(fname,grdformat,grddouble,rntime)
!**************************************************************
IMPLICIT NONE

real (kind=SP) :: sx
real (kind=DP) :: dx
character (len=chrlen), intent(in) :: fname

integer, intent(in) :: grdformat
integer :: filepipe = 13
integer, intent(out) :: rntime

logical, intent(in) :: grddouble

rntime = 0
open(filepipe, file = fname, form = 'unformatted', status = 'old')
READ(filepipe,ERR=101,END=101) !ngrd

do while(1)
  if ( grddouble ) then
    READ(filepipe,ERR=101,END=101) dx
  else
    READ(filepipe,ERR=101,END=101) sx
  endif
  rntime = rntime+1
enddo
101 continue
close(filepipe)

END SUBROUTINE ReadP3DTime



END MODULE MOD_P3DRay
