MODULE avusconvvar

  USE MOD_GRIDVARS

  IMPLICIT NONE

  integer :: NBCS
  integer,dimension(:),allocatable :: fvboundbc,wall_type

! Numeric tags (codes) for FIELDVIEW binary file format. 

! Bit pattern used to detect byte ordering.
      integer FV_MAGIC
      parameter(FV_MAGIC = 66051)

! Data section id codes
      integer FV_NODES
      parameter(FV_NODES = 1001)

      integer FV_FACES
      parameter(FV_FACES = 1002)

      integer FV_ELEMENTS
      parameter(FV_ELEMENTS = 1003)

      integer FV_VARIABLES
      parameter(FV_VARIABLES = 1004)

      integer FV_BNDRY_VARS
      parameter(FV_BNDRY_VARS = 1006)

! Element (cell) type id codes
  integer,parameter :: FV_TET = 1, FV_HEX = 2, FV_PRISM = 3, FV_PYRAMID = 4

! Values for "wall_info" array (see comments in ftn_encode_header)
      integer A_WALL
      parameter (A_WALL = 7)

      integer NOT_A_WALL
      parameter (NOT_A_WALL = 0)


  type cell_type
    integer(selected_int_kind(2)) :: nf,np
    integer :: faces(6),nodes(8),wall(6),hw
  end type cell_type
  type face_type
    integer(selected_int_kind(2)) :: np
    integer :: bc,nodes(4)
  end type face_type

  type(face_type) :: tmpfac(6)
  type(face_type),dimension(:),allocatable :: face
  type(cell_type),dimension(:),allocatable :: cell

END MODULE avusconvvar


MODULE avusconv

IMPLICIT NONE


CONTAINS

!#######################################################################
SUBROUTINE CleanAvusConv()
!#######################################################################
USE avusconvvar

IF(ALLOCATED(cell)) DEALLOCATE(cell)
IF(ALLOCATED(face)) DEALLOCATE(face)

IF(ALLOCATED(fvboundbc)) DEALLOCATE(fvboundbc)
IF(ALLOCATED(wall_type)) DEALLOCATE(wall_type)

END SUBROUTINE CleanAvusConv

!#######################################################################
  SUBROUTINE Set_AVUS_Cells(npatch,sumfaces,ipnt,icell,sumcells,numppf,bndcnt,bndloc,bcname)
!#######################################################################
    USE avusconvvar
    
    INTEGER, INTENT(IN) :: npatch
    INTEGER, POINTER, DIMENSION(:,:) :: ipnt, icell
    INTEGER, POINTER, DIMENSION(:) :: numppf
    INTEGER, INTENT(IN) :: sumfaces, sumcells
    INTEGER, DIMENSION(npatch), INTENT(OUT) :: bndcnt,bndloc
    CHARACTER*80, DIMENSION(npatch), INTENT(OUT) :: bcname
    
    INTEGER :: k,i,lcell,rcell,bndind
    INTEGER :: istat
    CHARACTER*80 :: intchr


    bndind = 0
    bndcnt = 0
    bndloc = 0
    allocate(face(sumfaces),cell(sumcells),stat=istat)
     do i=1,sumfaces
        face(i)%bc = 0
        face(i)%nodes = 0
     end do
     do i=1,sumcells
        cell(i)%nf = 0
        cell(i)%faces = 0
        cell(i)%nodes = 0
        cell(i)%wall = 0
     end do

     do i=1,sumfaces

        do k=1,numppf(i)
           face(i)%nodes(k) = ipnt(k,i)
        enddo
        face(i)%np = numppf(i)
        lcell = icell(1,i)
        rcell = icell(2,i)
 
        if(lcell > 0) then
           cell(lcell)%nf = cell(lcell)%nf + 1
           cell(lcell)%faces(cell(lcell)%nf) = i
        else
           IF(allocated(fvboundbc)) THEN
              face(i)%bc = fvboundbc(-lcell)
              bndcnt(fvboundbc(-lcell)) = bndcnt(fvboundbc(-lcell)) + 1
           ELSE
              IF(.not.ANY(bndloc.eq.-lcell)) THEN
                 bndind = bndind + 1
                 bndloc(bndind) = -lcell
                 IF(TRIM(bcname(bndind)).eq.'') THEN
                   CALL Int2Chr(-lcell,intchr)
                   bcname(bndind) = 'BC [' // TRIM(intchr) // ']'
                 ENDIF
              ENDIF
              WHERE(bndloc.eq.-lcell) bndcnt = bndcnt + 1
              k = 1
              DO WHILE(bndloc(k).ne.-lcell)
                k = k + 1
              ENDDO
              face(i)%bc = k
           ENDIF
        endif

        if(rcell > 0) then
           cell(rcell)%nf = cell(rcell)%nf + 1
           cell(rcell)%faces(cell(rcell)%nf) = i
        else
           IF(allocated(fvboundbc)) THEN
              face(i)%bc = fvboundbc(-rcell)
              bndcnt(fvboundbc(-rcell)) = bndcnt(fvboundbc(-rcell)) + 1
           ELSE
              IF(.not.ANY(bndloc.eq.-rcell)) THEN
                 bndind = bndind + 1
                 bndloc(bndind) = -rcell
                 IF(TRIM(bcname(bndind)).eq.'') THEN
                   CALL Int2Chr(-rcell,intchr)
                   bcname(bndind) = 'BC [' // TRIM(intchr) // ']'
                 ENDIF
              ENDIF
              WHERE(bndloc.eq.-rcell) bndcnt = bndcnt + 1
              k = 1
              DO WHILE(bndloc(k).ne.-rcell)
                k = k + 1
              ENDDO
              face(i)%bc = k
           ENDIF
        endif
     end do

! Order face points
     call fvptorder(sumfaces,sumcells,cell,face)

  END SUBROUTINE Set_AVUS_Cells

!***********************************************************************************************************************
  SUBROUTINE Int2Chr(int,chr)
  USE avusconvvar, ONLY : chrlen

  INTEGER, INTENT(IN) :: int
  CHARACTER(len=chrlen), INTENT(OUT) :: chr

  WRITE(chr,"(i6)") int
  chr = ADJUSTL(chr)


  END SUBROUTINE Int2Chr
!***********************************************************************************************************************
! Makes sure face points ordered, as rest of reader assumes this
! Matt Grismer 11Feb99
subroutine fvptorder(nfaces,ncells,cell,face)
  USE avusconvvar, ONLY : cell_type, face_type, chrlen, CheckAllocationError

  implicit none

integer,intent(in) :: nfaces,ncells
type(cell_type),dimension(ncells),intent(in) :: cell
type(face_type),dimension(nfaces),intent(inout) :: face

logical(kind=1),dimension(:),allocatable :: ordered
integer :: nf,nc,nfo,k,i,j,iff,jf,icnt,nfopcnt,fintr(2,4),ierr

  allocate(ordered(nfaces),stat=ierr)
  Call CheckAllocationError(ierr)

! Since we don't have an array giving the cells associated with a given face
! we'll just go through the cell-face array we do have.
! This will repeat faces, but we'll track the faces that are already ordered
  write(*,*) 'Ordering FieldView face points...'

  ordered = .false.

  do nc=1,ncells
    do iff=1,cell(nc)%nf
        nf = cell(nc)%faces(iff)

        if (ordered(nf) .or. face(nf)%nodes(4) == 0) then  !3 or less ordered
         ordered(nf) = .true.
         goto 300
        end if
    
        ordered(nf) = .true.
        do i=1,4
         fintr(1,i) = 0
         fintr(2,i) = 0
        end do
        icnt = 1
        do jf=1,cell(nc)%nf		!Loop over all other faces of cell
         nfo = cell(nc)%faces(jf)

         if (nfo /= nf) then		!Compare only to faces different then NF
          if (face(nfo)%nodes(4) == 0) then
           nfopcnt = 3
          else 
           nfopcnt = 4
          endif

          k = 1
          do i=1,4			!Loop over points of NF
           do j=1,nfopcnt		!Loop over other face points (NFO)

            if (face(nfo)%nodes(j) == face(nf)%nodes(i)) then	!Find matching points
             fintr(k,icnt) = face(nfo)%nodes(j)
             k = k + 1
            end if
      
            if (k > 2) then		!Exit loops when 2 matches found
             icnt = icnt + 1
             goto 400
            end if	

           end do
          end do

400      end if
        end do				!End loop over other faces

        face(nf)%nodes(1) = fintr(1,1)	!Pick two points to start face
        face(nf)%nodes(2) = fintr(2,1)

        i = 3
        k = 2
					!Determine order for rest of points
500     if (fintr(1,k) == face(nf)%nodes(i-1)) then
         face(nf)%nodes(i) = fintr(2,k)
         i = i + 1
        elseif (fintr(2,k) == face(nf)%nodes(i-1)) then
         face(nf)%nodes(i) = fintr(1,k)
         i = i + 1
        end if

        k = k + 1

        if (i > 4) goto 300	!Stop when all points determined
        if (k > 4) k = 2	!Keep looping through FINTR until finished
        goto 500

300    end do		!End loop over faces of cell
      end do		!End loop over cells

  deallocate(ordered)

end subroutine FVPTORDER


!***********************************************************************************************************************
subroutine readavusbc(bcfilename,maxpax,bcname)
USE avusconvvar

implicit none

character (len=*),intent(in) :: bcfilename
integer,intent(in) :: maxpax
character (len=chrlen), dimension(maxpax), intent(out) :: bcname

character (len=80) :: temp1,temp2,temp3
character (len=132) :: string
integer :: i,k,l,ll,idpx(maxpax),deca,ichr,npatch,istat
integer,dimension(maxpax) :: bcnumpt,namept
real :: bcvalt(13)

  NPATCH = 0
  NBCS = 0
    
  BCNUMPT = 0
  NAMEPT = 0

  bcname = ''

  allocate(WALL_TYPE(0:MAXPAX))
  WALL_TYPE = -1
  WALL_TYPE(0) = 0

  IF (TRIM(bcfilename) .eq. '') RETURN

  open(71,file=bcfilename,form='formatted',status='old',iostat=ISTAT)
  if(ISTAT /= 0) then
    write(*,*) 'AVUS bc file does not exist.'
    write(*,*)
    return
  end if

  read (71,*,iostat=istat) 
  read (71,*,iostat=istat) 
  read (71,*,iostat=istat) 

! while loop to read boundary condition specification blocks

  do while (NPATCH < MAXPAX)        
    IDPX = 0

    read (71,*,iostat=ISTAT) 
      if (istat < 0) goto 10
    read (71,'(A132)',iostat=istat) STRING
      if (istat < 0) goto 10

    L = 0
    LL = 0
    do I=132,1,-1
      ICHR = iachar(STRING(I:I))  
      if ((ICHR >= 48) .and. (ICHR <= 57)) then
        if (L == 0) LL = LL + 1
        DECA = 10**L
        K = ICHR - 48
        IDPX(LL) = IDPX(LL) + DECA*K
        L = L + 1
      else
        L = 0
      end if
    end do

    if (LL == 0) goto 10

    NBCS = NBCS + 1

    do I = 0,ll-1
      BCNUMPT(NPATCH + 1 + i) = IDPX(i+1)
      NAMEPT(NPATCH + 1 + i) = NBCS
    end do

    read (71,'(A80)',iostat=istat) BCNAME(NBCS)
    read (71,'(A80)',iostat=istat) TEMP1

! --- REMOVE ALL NON-ALPHABETIC CHARACTERS 
!     AND CONVERT TO UPPERCASE             ---

    K = 0
    do I=1,80
      ICHR = iachar(TEMP1(I:I))  
      if ((ICHR.ge.65) .and. (ICHR.le.90)) then
        K = K + 1 
        TEMP1(K:K) = TEMP1(I:I)
      elseif ((ICHR.ge.97) .and. (ICHR.le.122)) then
        K = K + 1
        TEMP1(K:K) = achar(ICHR - 32)
      end if
    end do

!-----------------------------------------------------------------------
! 'FAR' = Farfield
!         - 'FIX' = Fixed
!         - 'MOD' = Modified Riemann Invariant
!         - 'ORG' = Original Riemann Invariant
! 'SNK' = Sink
!         - 'PRS' = Static Pressure
!         - 'MAS' = Mass Flow
!         - 'CMF' = Corrected Mass Flow
!         - 'BLD' = Bleed Plenum Pressure
!         - 'ARR' = Area Ratio
! 'SRC' = Source
! 'WAL' = Solid Wall
!         - 'SLP' = Slip
!         - 'ISO' = Isothermal No Slip
!         - 'ADI' = Adiabatic No Slip Wall
!         - 'GEN' = General No Slip Wall
! 'PER' = Periodic
!         - 'LIN' = Linear
!         - 'CIR' = Circumferential
! 'USR' = User Specified
!-----------------------------------------------------------------------

! --- READ FARFIELD BC INFO ---

     if (TEMP1(1:3) == 'FAR') then
          WALL_TYPE(NBCS) = NOT_A_WALL

          read (71,'(A80)',iostat=istat) TEMP2
          read (71,   *   ,iostat=istat)
          read (71,   *   ,iostat=istat) (BCVALT(I),I=5,10)

! --- READ SINK BC INFO ---

     elseif (TEMP1(1:3) == 'SIN') then
          WALL_TYPE(NBCS) = NOT_A_WALL

          TEMP1(1:3) = 'SNK'

          read (71,'(A80)',iostat=istat) TEMP2
          read (71,   *   ,iostat=istat)
          read (71,   *   ,iostat=istat) (BCVALT(I),I=1,2)
          read (71,'(A80)',iostat=istat) TEMP3

          K = 0
          do I=1,80
           ICHR = iachar(TEMP2(I:I))  
           if ((ICHR.ge.65) .and. (ICHR.le.90)) then
            K = K + 1 
            TEMP2(K:K) = TEMP2(I:I)
           elseif ((ICHR.ge.97) .and. (ICHR.le.122)) then
            K = K + 1
            TEMP2(K:K) = achar(ICHR - 32)
           end if
          end do

          if (TEMP2(1:3).eq.'STA')  then
           TEMP2(1:3) = 'PRS'
          elseif (TEMP2(1:3).eq.'COR') then
           TEMP2(1:3) = 'CMF'
          elseif (TEMP2(1:3).eq.'BLE') then
           TEMP2(1:3) = 'BLD'
          elseif (TEMP2(1:3).eq.'ARE') then
           TEMP2(1:3) = 'ARR'
          end if

          K = 0
          do I=1,80
           ICHR = iachar(TEMP3(I:I))  
           if ((ICHR.ge.65) .and. (ICHR.le.90)) then
            K = K + 1 
            TEMP3(K:K) = TEMP3(I:I)
           elseif ((ICHR.ge.97) .and. (ICHR.le.122)) then
            K = K + 1
            TEMP3(K:K) = achar(ICHR - 32)
           end if
          end do

          if (TEMP2(1:3).ne.'BLD') then
           if ((TEMP3(1:1).eq.'Y') .or. (TEMP3(1:1).eq.'D') .or. &
               (TEMP3(1:1).eq.'S') ) then
            read (71,   *   ,iostat=istat)
            read (71,   *   ,iostat=istat) (BCVALT(I),I=3,9)
           end if
          end if

! --- READ SOURCE BC INFO ---

     elseif (TEMP1(1:3) == 'SOU') then
          WALL_TYPE(NBCS) = NOT_A_WALL

          TEMP1(1:3) = 'SRC'

          read (71,   *   ,iostat=istat)
          read (71,   *   ,iostat=istat)
          read (71,   *   ,iostat=istat) (BCVALT(I),I=1,13)
          read (71,'(A80)',iostat=istat) TEMP3

          K = 0
          do I=1,80
           ICHR = iachar(TEMP3(I:I))  
           if (ICHR.ge.65 .and. ICHR.le.90) then
            K = K + 1 
            TEMP3(K:K) = TEMP3(I:I)
           elseif (ICHR.ge.97 .and. ICHR.le.122) then
            K = K + 1
            TEMP3(K:K) = achar(ICHR - 32)
           end if
          end do

! --- READ WALL BC INFO ---

     elseif (TEMP1(1:3) == 'SOL') then
          WALL_TYPE(NBCS) = A_WALL

          TEMP1(1:3) = 'WAL'

          read (71,'(A80)',iostat=istat) TEMP2
          K = 0
          do I=1,80
           ICHR = iachar(TEMP2(I:I))  
           if (ICHR.ge.65 .and. ICHR.le.90) then
            K = K + 1 
            TEMP2(K:K) = TEMP2(I:I)
           elseif (ICHR.ge.97 .and. ICHR.le.122) then
            K = K + 1
            TEMP2(K:K) = achar(ICHR - 32)
           end if
          end do

          if (TEMP2(1:3).eq.'SLI')  then
           TEMP2(1:3) = 'SLP'
          elseif (TEMP2(1:3).eq.'ISO') then
           BCVALT(1) = 1.E+10
           read (71,   *   ,iostat=istat)
           read (71,   *   ,iostat=istat)  BCVALT(2)
          elseif (TEMP2(1:3).eq.'GEN') then
           read (71,   *   ,iostat=istat)
           read (71,   *   ,iostat=istat) (BCVALT(I),I=1,10)
          end if

          read (71,'(A80)',iostat=istat) TEMP3

! --- READ ROTOR/PROPELLOR BC INFO ---

    elseif ((TEMP1(1:3) == 'ROT').or.(TEMP1(1:3) == 'PRO')) then
         WALL_TYPE(NBCS) = NOT_A_WALL

         TEMP1(1:3) = 'ROT'
         read (71,   *   ,iostat=istat) TEMP2

         K = 0
         do I=1,80
          ICHR = iachar(TEMP2(I:I))  
          if (ICHR.ge.65 .and. ICHR.le.90) then
           K = K + 1
           TEMP2(K:K) = TEMP2(I:I)
          elseif (ICHR.ge.97 .and. ICHR.le.122) then
           K = K + 1
           TEMP2(K:K) = achar(ICHR - 32)
          end if
         end do

         read (71,   *   ,iostat=istat)
         read (71,   *   ,iostat=istat) (BCVALT(I),I=1,5)
         read (71,'(A80)',iostat=istat) TEMP3

         K = 0
         do I=1,80
          ICHR = iachar(TEMP3(I:I))  
          if (ICHR.ge.65 .and. ICHR.le.90) then
           K = K + 1
           TEMP3(K:K) = TEMP3(I:I)
          elseif (ICHR.ge.97 .and. ICHR.le.122) then
           K = K + 1
           TEMP3(K:K) = achar(ICHR - 32)
          end if
         end do

! --- READ PERIODIC BC INFO ---

     elseif (TEMP1(1:3) == 'PER') then
          WALL_TYPE(NBCS) = NOT_A_WALL

          read (71,'(A80)',iostat=istat) TEMP2

          K = 0
          do I=1,80
           ICHR = iachar(TEMP2(I:I))  
           if (ICHR.ge.65 .and. ICHR.le.90) then
            K = K + 1 
            TEMP2(K:K) = TEMP2(I:I)
           elseif (ICHR.ge.97 .and. ICHR.le.122) then
            K = K + 1
            TEMP2(K:K) = achar(ICHR - 32)
           end if
          end do

          read (71,   *   ,iostat=istat)
          read (71,   *   ,iostat=istat) (BCVALT(I),I=1,6)
          BCVALT(7) = BCVALT(7)

! --- READ USER-SPECIFIED BC INFO ---

     elseif (TEMP1(1:3) == 'USE') then
          WALL_TYPE(NBCS) = NOT_A_WALL

          TEMP1(1:3) = 'USR'
 
          read (71,'(A80)',iostat=istat) TEMP2

          K = 0
          do I=1,80
           ICHR = iachar(TEMP2(I:I))
           if (ICHR.ge.65 .and. ICHR.le.90) then
            K = K + 1
            TEMP2(K:K) = TEMP2(I:I)
           elseif (ICHR.ge.97 .and. ICHR.le.122) then
            K = K + 1
            TEMP2(K:K) = achar(ICHR - 32)
           end if
          end do


          if (TEMP2(1:3).eq.'SPE') then
           read (71,   *   ,iostat=istat)
           read (71,   *   ,iostat=istat) (BCVALT(I),I=1,6)
          end if

     end if

     NPATCH = NPATCH + LL

  end do

10 close(71)
   if (NPATCH /= MAXPAX) then
     write(*,*) 'There is a mismatch between the number of patches in the'
     write(*,*) 'grid and the number listed in the BC file.'
     stop
   end if

   close(71)

! Create FVBOUNDBC array, relates patch number to fv bc number below
  allocate( fvboundbc(minval(bcnumpt):maxval(bcnumpt)) )
  fvboundbc = 0
  fvboundbc(bcnumpt) = namept

end subroutine readavusbc




!***********************************************************************************************************************
! Arranges hex nodes in proper order from Cobalt grid
! Matt Grismer 9Oct96

subroutine user_hex(face,node_ids,face_types)
USE avusconvvar, ONLY : face_type, wall_type

implicit none

type(face_type),intent(inout) :: face(6)
integer,intent(out) :: node_ids(8),face_types(6)

type(face_type) :: tmpfac
integer :: fid,nppf,k,nid,nidb,nida,degen(4),four,spcnid
integer :: loop4(-3:8)
data loop4 /1,2,3,4,1,2,3,4,1,2,3,4/

! check for degenerate 3 pt faces
      k = 0
      do fid=1,6
       if(face(fid)%nodes(4) .eq. 0) then
        k = k + 1
        degen(k) = fid
       else
        four = fid
       end if
      end do
      if (k .ne. 0 .and. k .ne. 2) then
       write(*,*) 'Unknown 6 faced cell type...'
       return
      end if

! if degenerate, find shared point and repeat it where appropriate
      if (k .eq. 2) then
       fid = degen(1)
       nid = face(fid)%nodes(1)
       nida = face(fid)%nodes(2)
       nidb = face(fid)%nodes(3)

       fid = degen(2)
       if(face(fid)%nodes(1).eq.nid .or. face(fid)%nodes(1).eq.nida &
          .or. face(fid)%nodes(1).eq.nidb) then
         spcnid = face(fid)%nodes(1)
         face(fid)%nodes(4) = spcnid
       elseif(face(fid)%nodes(2).eq.nid .or. face(fid)%nodes(2).eq.nida &
              .or. face(fid)%nodes(2).eq.nidb) then
         spcnid = face(fid)%nodes(2)
         face(fid)%nodes(4) = face(fid)%nodes(3)
         face(fid)%nodes(3) = spcnid
       elseif(face(fid)%nodes(3).eq.nid .or. face(fid)%nodes(3).eq.nida &
              .or. face(fid)%nodes(3).eq.nidb) then
         spcnid = face(fid)%nodes(3)
         face(fid)%nodes(4) = spcnid
       else
        write(*,*) 'No matching degenerate pt found for degen hex...'
        return
       end if

       fid = degen(1)
       if (spcnid .eq. nid) then
         face(fid)%nodes(4) = nid
       elseif (spcnid .eq. nida) then
         face(fid)%nodes(4) = face(fid)%nodes(3)
         face(fid)%nodes(3) = nida
       elseif (spcnid .eq. nidb) then
         face(fid)%nodes(4) = nidb
       end if

! make sure face(1) is not degenerate, otherwise logic below will fail
       if (degen(1).eq.1 .or. degen(2).eq.1) then
        tmpfac = face(1)
        face(1) = face(four)
        face(four) = tmpfac
       end if  
      end if

! any face will work to begin, follow funky ordering required by fieldview
! first 4 points actually make face 3 
      fid = 1
      node_ids(1) = face(fid)%nodes(1)
      node_ids(2) = face(fid)%nodes(2)
      node_ids(4) = face(fid)%nodes(3)
      node_ids(3) = face(fid)%nodes(4)
      face_types(3) = wall_type(face(fid)%bc)
! now get other nodes and face conditions
        do fid=2,6
         nppf = 0
         do k=1,3
          nid = face(fid)%nodes(loop4(k))
          nidb = face(fid)%nodes(loop4(k-1))
          nida = face(fid)%nodes(loop4(k+1))

          if(nid.eq.node_ids(1)) then
! face 5
           if(nidb.eq.node_ids(2)) then
            node_ids(5) = nida
            node_ids(6) = face(fid)%nodes(loop4(k+2))
            face_types(5) = wall_type(face(fid)%bc)
           elseif(nida.eq.node_ids(2)) then 
            node_ids(5) = nidb  
            node_ids(6) = face(fid)%nodes(loop4(k-2))
            face_types(5) = wall_type(face(fid)%bc)
! face 1
           elseif(nidb.eq.node_ids(3)) then
            node_ids(5) = nida
            node_ids(7) = face(fid)%nodes(loop4(k+2))
            face_types(1) = wall_type(face(fid)%bc)            
           elseif(nida.eq.node_ids(3)) then
            node_ids(5) = nidb  
            node_ids(7) = face(fid)%nodes(loop4(k-2))
            face_types(1) = wall_type(face(fid)%bc)
           endif
           goto 899

          elseif(nid.eq.node_ids(2)) then
! face 2
           if(nidb.eq.node_ids(4)) then
            node_ids(6) = nida
            node_ids(8) = face(fid)%nodes(loop4(k+2))
            face_types(2) = wall_type(face(fid)%bc)
           elseif(nida.eq.node_ids(4)) then 
            node_ids(6) = nidb  
            node_ids(8) = face(fid)%nodes(loop4(k-2))
            face_types(2) = wall_type(face(fid)%bc)
! face 5
           elseif(nidb.eq.node_ids(1)) then
            node_ids(6) = nida
            node_ids(5) = face(fid)%nodes(loop4(k+2))
            face_types(5) = wall_type(face(fid)%bc)
           elseif(nida.eq.node_ids(1)) then 
            node_ids(6) = nidb  
            node_ids(5) = face(fid)%nodes(loop4(k-2))
            face_types(5) = wall_type(face(fid)%bc)
           endif
           goto 899

          elseif(nid.eq.node_ids(3)) then
! face 1
           if(nidb.eq.node_ids(1)) then
            node_ids(7) = nida
            node_ids(5) = face(fid)%nodes(loop4(k+2))
            face_types(1) = wall_type(face(fid)%bc)
           elseif(nida.eq.node_ids(1)) then 
            node_ids(7) = nidb  
            node_ids(5) = face(fid)%nodes(loop4(k-2))
            face_types(1) = wall_type(face(fid)%bc)
! face 6
           elseif(nidb.eq.node_ids(4)) then
            node_ids(7) = nida
            node_ids(8) = face(fid)%nodes(loop4(k+2))
            face_types(6) = wall_type(face(fid)%bc)
           elseif(nida.eq.node_ids(4)) then 
            node_ids(7) = nidb  
            node_ids(8) = face(fid)%nodes(loop4(k-2))
            face_types(6) = wall_type(face(fid)%bc)
           endif
           goto 899

          elseif(nid.eq.node_ids(4)) then
! face 2
           if(nidb.eq.node_ids(2)) then
            node_ids(8) = nida
            node_ids(6) = face(fid)%nodes(loop4(k+2))
            face_types(2) = wall_type(face(fid)%bc)
           elseif(nida.eq.node_ids(2)) then 
            node_ids(8) = nidb  
            node_ids(6) = face(fid)%nodes(loop4(k-2))
            face_types(2) = wall_type(face(fid)%bc)
! face 6
           elseif(nidb.eq.node_ids(3)) then
            node_ids(8) = nida
            node_ids(7) = face(fid)%nodes(loop4(k+2))
            face_types(6) = wall_type(face(fid)%bc)
           elseif(nida.eq.node_ids(3)) then 
            node_ids(8) = nidb  
            node_ids(7) = face(fid)%nodes(loop4(k-2))
            face_types(6) = wall_type(face(fid)%bc)
           endif
           goto 899
          else
           nppf = nppf + 1
           if(nppf.eq.3) face_types(4) = wall_type(face(fid)%bc)
          endif
         end do
899     end do

end subroutine user_hex


!***********************************************************************************************************************
! Arranges prism nodes in proper order from Cobalt grid
! Assumes face(1) has four points
! Matt Grismer 9Oct96
subroutine user_prism(face,node_ids,face_types)
USE avusconvvar, ONLY : face_type, wall_type

implicit none

type(face_type),intent(in) :: face(6)
integer,intent(out) :: node_ids(8),face_types(6)

integer :: fid,nid,nida,nidb,k
integer :: loop4(-3:8),loop3(-2:6)
data loop4 /1,2,3,4,1,2,3,4,1,2,3,4/
data loop3 /1,2,3,1,2,3,1,2,3/

! four sided face is face 1 for prism
        node_ids(1:4) = face(1)%nodes(1:4)
        face_types(1) = wall_type(face(1)%bc)

! if nodes 1 & 2 are on the same triangular face, need to rotate nodes 1-4 once
         nid = node_ids(1)
         nida = node_ids(2)
         do fid=1,5
          if(face(fid)%nodes(4) == 0) then
           if(nid == face(fid)%nodes(1).or.nid == face(fid)%nodes(2) &
              .or.nid == face(fid)%nodes(3)) then
            if(nida == face(fid)%nodes(1).or.nida == face(fid)%nodes(2) &
               .or.nida == face(fid)%nodes(3)) then
              node_ids(5) = node_ids(1)
              node_ids(1) = node_ids(2)
              node_ids(2) = node_ids(3)
              node_ids(3) = node_ids(4)
              node_ids(4) = node_ids(5)
              goto 688
             endif
            endif
           endif
          end do 
! get other points and labels faces
688      do fid=2,5
! triangular face, either face 4 or 5
          if(face(fid)%nodes(4) == 0) then
           do k=1,2
            nid = face(fid)%nodes(loop3(k))
            nidb = face(fid)%nodes(loop3(k-1))
            nida = face(fid)%nodes(loop3(k+1))
! face 4
            if(nid == node_ids(1)) then
             if(nidb == node_ids(4)) then
              node_ids(6) = nida
             elseif(nida == node_ids(4)) then
              node_ids(6) = nidb
             endif
             face_types(4) = wall_type(face(fid)%bc)
             goto 699
 
            elseif(nid == node_ids(4)) then
             if(nidb == node_ids(1)) then
              node_ids(6) = nida
             elseif(nida == node_ids(1)) then
              node_ids(6) = nidb
             endif
             face_types(4) = wall_type(face(fid)%bc)
             goto 699
! face 5
            elseif(nid == node_ids(2)) then
             if(nidb == node_ids(3)) then
              node_ids(5) = nida
             elseif(nida == node_ids(3)) then
              node_ids(5) = nidb
             endif
             face_types(5) = wall_type(face(fid)%bc)
             goto 699
 
            elseif(nid == node_ids(3)) then
             if(nidb == node_ids(2)) then
              node_ids(5) = nida
             elseif(nida == node_ids(2)) then
              node_ids(5) = nidb
             endif
             face_types(5) = wall_type(face(fid)%bc)
             goto 699
            endif
           end do
! rectangular face, either face 2 or 3
          else
           do k=1,3
            nid = face(fid)%nodes(loop4(k))
            nidb = face(fid)%nodes(loop4(k-1))
            nida = face(fid)%nodes(loop4(k+1))
! face 2
            if(nid == node_ids(1)) then
             if(nidb == node_ids(2)) then
              node_ids(6) = nida
              node_ids(5) = face(fid)%nodes(loop4(k+2))
             elseif(nida == node_ids(2)) then 
              node_ids(6) = nidb  
              node_ids(5) = face(fid)%nodes(loop4(k-2))
             endif
             face_types(2) = wall_type(face(fid)%bc)
             goto 699

            elseif(nid == node_ids(2)) then
             if(nidb == node_ids(1)) then
              node_ids(5) = nida
              node_ids(6) = face(fid)%nodes(loop4(k+2))
             elseif(nida == node_ids(1)) then
              node_ids(5) = nidb  
              node_ids(6) = face(fid)%nodes(loop4(k-2))
             endif
             face_types(2) = wall_type(face(fid)%bc)
             goto 699
! face 3 
            elseif(nid == node_ids(3)) then
             if(nidb == node_ids(4)) then
              node_ids(5) = nida
              node_ids(6) = face(fid)%nodes(loop4(k+2))
             elseif(nida == node_ids(4)) then 
              node_ids(5) = nidb  
              node_ids(6) = face(fid)%nodes(loop4(k-2))
             endif
             face_types(3) = wall_type(face(fid)%bc)
             goto 699

            elseif(nid == node_ids(4)) then
             if(nidb == node_ids(3)) then
              node_ids(6) = nida
              node_ids(5) = face(fid)%nodes(loop4(k+2))
             elseif(nida == node_ids(3)) then
              node_ids(6) = nidb  
              node_ids(5) = face(fid)%nodes(loop4(k-2))
             endif
             face_types(3) = wall_type(face(fid)%bc)
             goto 699
            endif
           end do
          endif
699      end do

end subroutine user_prism


!***********************************************************************************************************************
! Arranges pyramid nodes in proper order from Cobalt grid
! Assumes face(1) has four points
! Matt Grismer 9Oct96

subroutine user_pyr(face,node_ids,face_types)
USE avusconvvar, ONLY : face_type, wall_type

implicit none

type(face_type),intent(in) :: face(6)
integer,intent(out) :: node_ids(8),face_types(6)

integer :: fid,nid,nida,nidb,k
integer :: loop3(-2:6)
data loop3 /1,2,3,1,2,3,1,2,3/

! four sided face is face 1 for pyramid
        node_ids(1:4) = face(1)%nodes(1:4)
        face_types(1) = wall_type(face(1)%bc)

! get last point and label other faces
         do fid=2,5

          do k=1,2
           nid = face(fid)%nodes(loop3(k))
           nidb = face(fid)%nodes(loop3(k-1))
           nida = face(fid)%nodes(loop3(k+1))

           if(nid == node_ids(1)) then
! face 5
            if(nidb == node_ids(2)) then
             node_ids(5) = nida
             face_types(5) = wall_type(face(fid)%bc)
            elseif(nida == node_ids(2)) then
             node_ids(5) = nidb
             face_types(5) = wall_type(face(fid)%bc)
! face 4
            elseif(nidb == node_ids(4)) then
             node_ids(5) = nida
             face_types(4) = wall_type(face(fid)%bc)
            elseif(nida == node_ids(4)) then
             node_ids(5) = nidb
             face_types(4) = wall_type(face(fid)%bc)
            endif
            goto 799

           elseif(nid == node_ids(2)) then
! face 5
            if(nidb == node_ids(1)) then
             node_ids(5) = nida
             face_types(5) = wall_type(face(fid)%bc)
           elseif(nida == node_ids(1)) then
             node_ids(5) = nidb
             face_types(5) = wall_type(face(fid)%bc)
! face 2
            elseif(nidb == node_ids(3)) then
             node_ids(5) = nida
             face_types(2) = wall_type(face(fid)%bc)
            elseif(nida == node_ids(3)) then
             node_ids(5) = nidb
             face_types(2) = wall_type(face(fid)%bc)
            endif
            goto 799

           elseif(nid == node_ids(3)) then
! face 3
            if(nidb == node_ids(4)) then
             node_ids(5) = nida
             face_types(3) = wall_type(face(fid)%bc)
            elseif(nida == node_ids(4)) then
             node_ids(5) = nidb
             face_types(3) = wall_type(face(fid)%bc)
! face 2
            elseif(nidb == node_ids(2)) then
             node_ids(5) = nida
             face_types(2) = wall_type(face(fid)%bc)
            elseif(nida == node_ids(2)) then
             node_ids(5) = nidb
             face_types(2) = wall_type(face(fid)%bc)
            endif
            goto 799

           elseif(nid == node_ids(4)) then

! face 3
            if(nidb == node_ids(3)) then
             node_ids(5) = nida
             face_types(3) = wall_type(face(fid)%bc)
            elseif(nida == node_ids(3)) then
             node_ids(5) = nidb
             face_types(3) = wall_type(face(fid)%bc)
! face 4
            elseif(nidb == node_ids(1)) then
             node_ids(5) = nida
             face_types(4) = wall_type(face(fid)%bc)
            elseif(nida == node_ids(1)) then
             node_ids(5) = nidb
             face_types(4) = wall_type(face(fid)%bc)
            endif
            goto 799
           endif
          end do
799      end do

end subroutine user_pyr




!***********************************************************************************************************************
! Extrudes quadrilateral faces to hexahedrals (2D to 3D)
! Bob Tomaro - January 1998

subroutine user_quad(face,node_ids,face_types,nodecnt)
USE avusconvvar, ONLY : face_type, wall_type, NOT_A_WALL

implicit none

type(face_type),intent(in) :: face(6)
integer,intent(in) :: nodecnt
integer,intent(out) :: node_ids(8),face_types(6)

integer :: ns,np,nf(4)

! order the four quadrilateral faces that make up the hexahedral 

      node_ids(1) = face(1)%nodes(1)
      node_ids(2) = face(1)%nodes(2)

      do ns=2,4
       np = 0
       if (node_ids(2) == face(ns)%nodes(1)) np = 2 
       if (node_ids(2) == face(ns)%nodes(2)) np = 1
       if (np > 0) then
        node_ids(3) = face(ns)%nodes(np)
        nf(3) = ns
       end if
      end do

      do ns=2,4
       np = 0
       if (node_ids(3) == face(ns)%nodes(1)) np = 2 
       if (node_ids(3) == face(ns)%nodes(2)) np = 1
       if (np > 0 .and. ns .ne. nf(3)) then
        node_ids(4) = face(ns)%nodes(np)
        nf(4) = ns
       end if
      end do

! find 'other' face

      do ns=2,4
       if ((nf(3) /= ns).and.(nf(4) /= ns)) nf(2) = ns
      end do 

! extrude nodes 1-4

      node_ids(5) = node_ids(1) + nodecnt
      node_ids(6) = node_ids(2) + nodecnt
      node_ids(7) = node_ids(3) + nodecnt
      node_ids(8) = node_ids(4) + nodecnt

      face_types(1) = wall_type(face(nf(3))%bc)
      face_types(2) = wall_type(face(nf(4))%bc)
      face_types(5) = wall_type(face(    1)%bc)
      face_types(6) = wall_type(face(nf(2))%bc)

! faces 3 and 4 are the 'cap' faces

      face_types(3) = NOT_A_WALL
      face_types(4) = NOT_A_WALL

end subroutine user_quad



!***********************************************************************************************************************
! Arranges tet nodes in proper order from Cobalt grid
! Matt Grismer 9Oct96
subroutine user_tet(face,node_ids,face_types)
USE avusconvvar, ONLY : face_type, wall_type

implicit none

type(face_type),intent(in) :: face(6)
integer,intent(out) :: node_ids(8),face_types(6)

integer :: fid,ifcnt2,ifcnt3,ifcntu

! Only face(1-4) for tet, any face will work as face 1
        fid = 1
        node_ids(1:3) = face(fid)%nodes(1:3)
        face_types(1) = wall_type(face(fid)%bc)

! get other node and label other faces
        fid = 2
	if(face(fid)%nodes(1) /= node_ids(1).and.face(fid)%nodes(1) /= &
           node_ids(2).and.face(fid)%nodes(1) /= node_ids(3)) then
	 node_ids(4) = face(fid)%nodes(1)
        elseif(face(fid)%nodes(2) /= node_ids(1).and.face(fid)%nodes(2) /= &
               node_ids(2).and.face(fid)%nodes(2) /= node_ids(3)) then
	 node_ids(4) = face(fid)%nodes(2)
        else
	 node_ids(4) = face(fid)%nodes(3)
	endif

        ifcnt2 = node_ids(2) + node_ids(3) + node_ids(4)
        ifcnt3 = node_ids(1) + node_ids(3) + node_ids(4)
        do fid=2,4
         ifcntu = face(fid)%nodes(1) + face(fid)%nodes(2) + face(fid)%nodes(3)
         if(ifcntu.eq.ifcnt2) then
          face_types(2) = wall_type(face(fid)%bc)
         elseif(ifcntu.eq.ifcnt3) then
          face_types(3) = wall_type(face(fid)%bc)
         else
          face_types(4) = wall_type(face(fid)%bc)
         endif
        end do

end subroutine user_tet



!***********************************************************************************************************************
! Extrudes quadrilateral faces to hexahedrals (2D to 3D)
! Bob Tomaro - January 1998

subroutine user_tri(face,node_ids,face_types,nodecnt)
USE avusconvvar, ONLY : face_type, wall_type

implicit none

type(face_type),intent(in) :: face(6)
integer,intent(in) :: nodecnt
integer,intent(out) :: node_ids(8),face_types(6)

         node_ids(1) = face(1)%nodes(1)
         node_ids(4) = face(1)%nodes(2)
         face_types(1) = wall_type(face(1)%bc)

         node_ids(6) = face(2)%nodes(1)

         if (node_ids(6) == node_ids(4)) then
           node_ids(6) = face(2)%nodes(2)
           face_types(3) = wall_type(face(2)%bc)
           face_types(2) = wall_type(face(3)%bc)
         elseif (node_ids(6) == node_ids(1)) then
           node_ids(6) = face(2)%nodes(2)
           face_types(2) = wall_type(face(2)%bc)
           face_types(3) = wall_type(face(3)%bc)
         elseif (face(2)%nodes(2) == node_ids(4)) then
           face_types(3) = wall_type(face(2)%bc)
           face_types(2) = wall_type(face(3)%bc)
         else
           face_types(2) = wall_type(face(2)%bc)
           face_types(3) = wall_type(face(3)%bc)
         end if

         node_ids(2) = node_ids(1) + nodecnt
         node_ids(3) = node_ids(4) + nodecnt
         node_ids(5) = node_ids(6) + nodecnt

end subroutine user_tri
END MODULE avusconv
